c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Compute_grid_energy.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	8/17/07
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	Compute the grid energy, using:
c
c			E = 1/2 Sum_{i,j,k} phi(i,j,k) q(i,j,k)
c
c			where phi is the electrostatic potential
c			      q is the charge
c
c	This version just loops over all grid points
c
c	Input:
c
c		u : dimensioneless potential
c
c	Output:
c
c		ener: grid energy
c
c	Units consideration:
c
c	u is the dimensionless electrostatic potential, i.e.:
c
c	u = ec * phi
c	    --------
c            kB T
c
c	hence phi = kB*T*u/ec, and phi*q = kB * T * u *z
c	where z is the partial charge in electron unit.
c
c	The product u*z gives the energy in kB * T units.
c
c	Note that the charge on the grid are stored in crg_map, and have been
c	multiplied by C_charge; we need to remove this factor
c
	subroutine compute_grid_energy_vol(u,crg_map,ener,hx,hy,hz,
     1                          nx,ny,nz)
c
c ==========================================================================================
c = poisson.h defines array sizes             ==============================================
c ==========================================================================================
c
	include 'poisson.h'
c
c ==========================================================================================
c = Declares all local/global arrays          ==============================================
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
c
c	Parameters
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
c	Information on grid
c ==========================================================================================
c
	real*8	dx,dy,dz
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Information on maps over the grid:
c ==========================================================================================
c
c	helmoltz_map:		region in which helmoltz term is non zero
c
	real*8	crg_map(nx,ny,nz)
c
c       solution of the equation and energy values
c ==========================================================================================
c
	real*8	ener
c
	real*8	u(nx,ny,nz)
c
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
c
c ==========================================================================================
c = Compute energy             =============================================================
c ==========================================================================================
c
	ener = 0
	do 300 k = 1,nx
		dz = hz(k+1)-hz(k)
		do 200 j = 1,ny
			dy = hy(j+1)-hy(j)
			do 100 i = 1,nz
				dx = hx(i+1)-hx(i)
				if(crg_map(i,j,k).ne.0) then
					ener = ener + u(i,j,k)*
     1					crg_map(i,j,k)*dx*dy*dz
				endif
100			continue
200		continue
300	continue
c
	ener = ener/(2.d0*C_charge)
c
        write(6,*) ' '
	write(6,*) 'Grid energy =',ener,' kT'
        write(6,*) ' '
c
c	ener above is in KB*T unit: converts in kJ:
c
	ener = KB*Temp*Avogadro*(1.d-3)*ener
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	compute_grid_energy_atom.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	Purpose:	Compute the grid energy, using:
c
c			E = 1/2 Sum_{i,j,k} phi(i,j,k) q(i,j,k)
c
c			where phi is the electrostatic potential
c			      q is the charge
c
c	This version loops over each atom, and computes its contribution
c	to the energy
c
c	Input:
c
c		u : dimensioneless potential
c
c	Output:
c
c		ener: grid energy
c
c	For unit consideration, see compute_grid_energy_vol
c
	subroutine compute_grid_energy_atom(u,ener,ene_atom,hx,hy,hz,
     1                          nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c	Define all variables
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
c
	integer	ipoint,istepx,istepy,istepz
	integer	ix,iy,iz,kx,ky,kz
	integer	Nvert_in,n_in
c
	real*8	val,valx,valy,valz,rad,rad2,crg,r_max
	real*8	diffx,diffy,diffz
c
	integer	posx(*),posy(*),posz(*)
c
c	Information on grid
c ==========================================================================================
c
c
	real*8	dx,dy,dz
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Input parameters
c ==========================================================================================
c
	integer	surf_type,flag_pymol,flag_box,flag_distrib,flag_eps
	integer	grid_type,bnd_type
	integer	ion_zone
c
	real*8	R_probe,R_ion
c
c	Electrostatics potential
c ==========================================================================================
c
	real*8	u(nx,ny,nz)
c
c	Energy
c
	real*8	ener,ener2
        real*8  ene_atom(*)
c
c	Constants
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
c	Pointers
c ==========================================================================================
c
        pointer (ptr_posx,posx)
        pointer (ptr_posy,posy)
        pointer (ptr_posz,posz)
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        common /charges/   ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                     ptr_label2,iform,natom,ncharge
	common /constants/ l_Bjerrum,pi,Temp,KB,electron,Ang,Avogadro,
     1			   eps0,beta,C_charge,C_prop
	common /input_par/ R_probe,R_ion,flag_pymol,surf_type,
     1			   grid_type,ion_zone,flag_box,
     2			   flag_distrib,flag_eps,bnd_type
c
	ener = 0
c
	if(flag_distrib.eq.0) then
c
c ==========================================================================================
c	Method 1: trilinear
c ==========================================================================================
c
	     do 100 ipoint = 1,natom
c
c
c	   	Find grid cell to which it belongs (coordinate of the LFD vertex,
c		see below for definition of vertices)
c
		call locate_cell(atom_coord(3*(ipoint-1)+1),hx,ix,
     1				 nx)
		call locate_cell(atom_coord(3*(ipoint-1)+2),hy,iy,
     1				 ny)
		call locate_cell(atom_coord(3*(ipoint-1)+3),hz,iz,
     1				 nz)
c
		dx = hx(ix+1)-hx(ix)
		dy = hy(iy+1)-hy(iy)
		dz = hz(iz+1)-hz(iz)
c
c		Get distance from point to YZ, XZ and XY planes of the grid cell
c
		diffx = (atom_coord(3*(ipoint-1)+1)-hx(ix))/dx
		diffy = (atom_coord(3*(ipoint-1)+2)-hy(iy))/dy
		diffz = (atom_coord(3*(ipoint-1)+3)-hz(iz))/dz
c
c
		crg = atom_charge(ipoint)
c
c		Now interpolate charge and get corresponding energy contribution
c
		ener2 = 0
		call trilinear_ener(ener2,u,ix,iy,iz,diffx,diffy,
     1		diffz,crg,nx,ny,nz)
		ener = ener + ener2
                ene_atom(ipoint) = ener2
c
100	    continue
c
	elseif(flag_distrib.eq.1) then
c
c ==========================================================================================
c	Method 2: Uniform spherical
c		  (see crg_distrib for details)
c ==========================================================================================
c
            r_max = 0.d0
            do 450 ipoint = 1,natom
                if(atom_radius(ipoint).gt.r_max) then
                        r_max = atom_radius(ipoint)
                endif
450         continue
c
            dx = hx(2)-hx(1)
            n_in = nint(2*r_max/dx) + 1
            n_in = n_in**3
            ptr_posx = malloc(4*n_in)
            ptr_posy = malloc(4*n_in)
            ptr_posz = malloc(4*n_in)
c
	    do 900 ipoint = 1,natom
c
c	   	Find grid cell to which it belongs (coordinate of the LFD vertex,
c		see below for definition of vertices)
c
		call locate_cell(atom_coord(3*(ipoint-1)+1),hx,ix,
     1				 nx)
		call locate_cell(atom_coord(3*(ipoint-1)+2),hy,iy,
     1				 ny)
		call locate_cell(atom_coord(3*(ipoint-1)+3),hz,iz,
     1				 nz)
c
		rad = atom_radius(ipoint)
		crg = atom_charge(ipoint)
		rad2 = rad*rad
c
c	   	Number of vertices to check in each dimension:
c
		call defstep(rad,hx,ix,istepx,nx)
		call defstep(rad,hy,iy,istepy,ny)
		call defstep(rad,hz,iz,istepz,nz)
c
c		Now loop over all these vertices
c
		Nvert_in = 0
c
		do 400 k = 1,2*istepz + 1
c
		    kz = iz - istepz + k -1
		    valz=(atom_coord(3*ipoint)-hz(kz))**2
c
		    do 300 j = 1,2*istepy + 1
c
			ky = iy - istepy + j -1 
		   	valy=(atom_coord(3*ipoint-1)-hy(ky))**2
c
			do 200 i = 1,2*istepx + 1
c
				kx = ix - istepx + i -1
				valx=(atom_coord(3*ipoint-2)-hx(kx))**2
c
c				if vertex inside ball ipoint, store
c
				if(valx+valy+valz.lt.rad2) then
					Nvert_in = Nvert_in + 1
					posx(Nvert_in)=kx
					posy(Nvert_in)=ky
					posz(Nvert_in)=kz
				endif
c
200			continue
300		    continue
400		continue
c
		if(Nvert_in.gt.8) then
c
c			Distribute charge uniformly
c
			val = crg/dble(Nvert_in)
                        ene_atom(ipoint) = 0
			do 500 i = 1,Nvert_in
			   ener = ener + val*u(posx(i),posy(i),posz(i))
                           ene_atom(ipoint) = ene_atom(ipoint)+
     1                          val*u(posx(i),posy(i),posz(i))
500			continue
c
		else
c
c			Do trilinear interpolation
c
c			Get distance from point to YZ, XZ and XY planes of the grid cell
c
			dx = hx(ix+1)-hx(ix)
			dy = hy(iy+1)-hy(iy)
			dz = hz(iz+1)-hz(iz)
c
			diffx = (atom_coord(3*(ipoint-1)+1)-hx(ix))/dx
			diffy = (atom_coord(3*(ipoint-1)+2)-hy(iy))/dy
			diffz = (atom_coord(3*(ipoint-1)+3)-hz(iz))/dz
c
                        ener2 = 0
			call trilinear_ener(ener2,u,ix,iy,iz,diffx,
     1			diffy,diffz,crg,nx,ny,nz)
                        ener = ener + ener2
                        ene_atom(ipoint) = ener2
c
		endif
c
900	    continue
c
            call free(ptr_posx)
            call free(ptr_posy)
            call free(ptr_posz)
c
	endif
c
c	Scale by 0.5, as this is a reaction field energy
c
	ener = ener/2.d0
        write(6,*) ' '
	write(6,*) 'Grid energy =',ener,' kT'
        write(6,*) ' '
c
c	ener above is in KB*T unit: converts in kJ:
c
	ener = KB*Temp*Avogadro*(1.d-3)*ener
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	trilinear_ener.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine distributes a charge on the 8 corners of a grid cell
c	and computes its contribution to the grid energy
c
c	The grid cell has 6 faces: L(eft) R(ight), F(ront) B(ack),
c	and U(p) D(own).
c	The eight vertices can be defined :
c	LFD, LFU, RFD, RFU, LBD, LBU, RBD, RBU.
c
c	Input:
c
c		ener:		current energy value
c		ix,iy,iz:	position of the LFB corner of the cube
c
c		diffx,diffy,diffz : distance of point to the YZ, XZ and XY plane of the
c				    grid cell, respectively
c
c		crg		  : charge to be distributed
c		u      		  : electrostatics potential
c
	subroutine trilinear_ener(ener,u,ix,iy,iz,diffx,diffy,
     1			diffz,crg,nx,ny,nz)
c
c ==========================================================================================
c = Define variables                          
c ==========================================================================================
c
	integer	ix,iy,iz
	integer	nx,ny,nz
c
	real*8	diffx,diffy,diffz,One_diffx,One_diffy,One_diffz
	real*8	crg
	real*8	ener
c
	real*8	u(nx,ny,nz)
c
	One_diffx = 1.d0-diffx
	One_diffy = 1.d0-diffy
	One_diffz = 1.d0-diffz
c
c ==========================================================================================
c = Interpolate value                         
c ==========================================================================================
c
c	Charge on LFD (point (ix,iy,iz))
c
	ener = ener + u(ix,iy,iz)*
     1		One_diffx*One_diffy*One_diffz*crg
c
c	Charge on RFD (point (ix+1,iy,iz)
c
	ener = ener + u(ix+1,iy,iz)*
     1		diffx*One_diffy*One_diffz*crg
c
c	Charge on LBD (point (ix,iy+1,iz)
c
	ener = ener + u(ix,iy+1,iz)*
     1		One_diffx*diffy*One_diffz*crg
c
c	Charge on RBD (point (ix+1,iy+1,iz)
c
	ener = ener + u(ix+1,iy+1,iz)*
     1		diffx*diffy*One_diffz*crg
c
c	Charge on LFU (point (ix,iy,iz+1)
c
	ener = ener + u(ix,iy,iz+1)*
     1		One_diffx*One_diffy*diffz*crg
c
c	Charge on RFU (point (ix+1,iy,iz+1)
c
	ener = ener + u(ix+1,iy,iz+1)*
     1		diffx*One_diffy*diffz*crg
c
c	Charge on LBU (point (ix,iy+1,iz+1)
c
	ener = ener + u(ix,iy+1,iz+1)*
     1		One_diffx*diffy*diffz*crg
c
c	Charge on RBU (point (ix+1,iy+1,iz+1)
c
	ener = ener + u(ix+1,iy+1,iz+1)*
     1		diffx*diffy*diffz*crg
c
	return
	end
