c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Direct_Solver_gpbl.f
c       (Direct, iterative PDE solver for PB equations, based on Michael Holst MG package)
c
c	Possible direct solvers:
c		- Successive Over Relaxation (SOR)
c		- Red-Black Gauss Seidel
c		- Jacobi
c
c	Version 1:      7/4/07
c
c	Author: Patrice Koehl (in collaboration with Marc Delarue)
c
c	It is inspired form Michael Holst MG code, which is available at:
c	http://www.fetk.org/codes/pmg/index.html
c
c ==========================================================================================
c ==========================================================================================
c 	Direct_Solver_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c	Direct_Solver implements direct, iterative solvers to the GPBL equation
c
	subroutine Direct_Solver_gpbl(iparam,rparam,iwork,rwork,u,
     1				xpos,ypos,zpos,bndx,bndy,bndz,
     2				eps_X,eps_Y,eps_Z,helm,rhs,
     3                          grid_info)
c
c	Input:
c		iparam:		integer array: gives flags to Solver
c		rparam:		real array: constants for solver
c		iwork		integer work array
c		rwork		real work array
c
c	One dimensional:
c		xpos		positions of grid points along X axis
c		ypos		positions of grid points along Y axis
c		zpos		positions of grid points along Z axis
c
c	Two dimensional:
c		bndx		boundary conditions (X=1 and X=Nx planes)
c		bndy		boundary conditions (Y=1 and Y=Ny planes)
c		bndz		boundary conditions (Z=1 and Z=Nz planes)
c
c	Three dimensional:
c		eps_X		epsilon map, defined at positions (i+0.5,j,k)
c		eps_Y		epsilon map, defined at positions (i,j+0.5,k)
c		eps_Z		epsilon map, defined at positions (i,j,k+0.5)
c
c		helm		Helmholtz term in the equation
c		rhs		right hand side of the equation
c
c	Output:
c		u		the dimensionless electrostatic potential map
c
c	Dimensions of all arrays have been set in the calling program
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer		iparam(*),iwork(*)
	real*8		rparam(*),rwork(*),u(*)
	real*8		xpos(*),ypos(*),zpos(*),bndx(*),bndy(*),bndz(*)
	real*8		eps_X(*),eps_Y(*),eps_Z(*),helm(*),rhs(*)
c
        integer*1       grid_info(*)
c
c	Local variables:
c ==========================================================================================
c
	integer	i,n,nx,ny,nz,niwk,nrwk,nlevel,ierror
	integer k_iz,k_w0
	integer k_ipc,k_rpc,k_cc,k_fc,k_vect
c
	integer	n_rpc,n_ipc,n_iz,ireal_tot,iint_tot
c
c ==========================================================================================
c 	Decode input array iparam, and do some sanity check
c ==========================================================================================
c
	nrwk	= iparam(1)
	niwk	= iparam(2)
	nx	= iparam(3)
	ny	= iparam(4)
	nz	= iparam(5)
	nlevel	= iparam(6)
c
	if(nlevel.le.0.or.nx.le.0.or.ny.le.0.or.nz.le.0) then
		write(6,*) 
     1          'nx,ny,nz and nlevel must be strictly positive!!'
		ierror = -1
		iparam(51) = ierror
		return
	endif
c
        n = nx*ny*nz
        n_iz = 10*(nlevel+1)
        n_ipc = 100*(nlevel+1)
        n_rpc = n_ipc
c
        iint_tot = n_iz + n_ipc
        ireal_tot = n_rpc + 3*n
c
c	Check that work arrays (both real and integer) have enough space
c ==========================================================================================
c
	if(nrwk.lt.ireal_tot) then
		write(6,*)'Real work space must be at least: ',ireal_tot
		ierror = -3
		iparam(51) = ierror
		return
	endif
c
	if(niwk.lt.iint_tot) then
		write(6,*) 'Integer work space must be at least: ',
     1			iint_tot
		ierror = -3
		iparam(51) = ierror
		return
	endif
c
c       Initialize work array
c ==========================================================================================
c
	do 100 i = 1,ireal_tot
		rwork(i) = 0
100	continue
c
c	Split the integer work array
c ==========================================================================================
c
	k_iz = 1
	k_ipc = k_iz + n_iz
c
c	Split the real work array
c ==========================================================================================
c
	k_rpc = 1
	k_cc  = k_rpc + n_rpc
	k_fc  = k_cc + n
	k_w0  = k_fc + n
	k_vect  = k_w0 + n
c
c ==========================================================================================
c	Start the multigrid solver
c ==========================================================================================
c
	call direct_gpbl(iparam,rparam,nx,ny,nz,u,iwork(k_iz),
     1		rwork(k_w0),iwork(k_ipc),rwork(k_rpc),
     2		rwork(k_vect),rwork(k_cc),rwork(k_fc),
     3		xpos,ypos,zpos,bndx,bndy,bndz,eps_X,eps_Y,eps_Z,helm,
     4		rhs,grid_info)
c
c ==========================================================================================
c	The program should have found the solution...return to the
c	main program
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	direct_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine uses a Newton method, combined with a linear multigrid
c	solver, to solve modified Poisson-Boltzmann equations.
c
	subroutine direct_gpbl(iparam,rparam,nx,ny,nz,u,iz,w0,
     1		ipc,rpc,vect,cc,fc,xpos,ypos,zpos,gxcf,gycf,gzcf,
     2		a1cf,a2cf,a3cf,ccf,fcf,grid_info)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer		iparam(*),ipc(*),iz(*)
	real*8		rparam(*),rpc(*),vect(*),cc(*),fc(*)
        integer*1       grid_info(*)
	real*8		u(*),w0(*)
	real*8		xpos(*),ypos(*),zpos(*),gxcf(*),gycf(*),gzcf(*)
	real*8		a1cf(*),a2cf(*),a3cf(*),ccf(*),fcf(*)
c
c	Local variables:
c ==========================================================================================
c
	integer		mgkey,nlevel,itmax,iok,iinfo,istop,ipkey
	integer		nx,ny,nz,ilev,ido,iters,ierror,mgsmoo
c
c	real*8		epsmach
	real*8		epsilon,errtol,omegan
c
c ==========================================================================================
c 	Decode the iparam and rparam arrays
c ==========================================================================================
c
	nlevel = iparam(6)	! Number of mesh level
	mgkey  = iparam(9)	! multigrid method: 0=variable, v-cycle, 1=nest operation
	itmax  = iparam(10)	! maximum number of iterations
	istop  = iparam(11)	! stopping criteria
	iinfo  = iparam(12)	! runtime status messages :
				! 0=none, 1=some, 2=many, 3=more
	ipkey  = iparam(14)	! Defines type of equation considered
				! 0: Laplace, 1: linear PB, 2: non linear PB
				! 3: Size modified PB, 4: Generalized PB Langevin
	mgsmoo = iparam(20)	! type of smoothing (0=wjac, 1=gsrb, 2=sor, 3=rich)
c
	errtol = rparam(1)	! tolerance factor
	omegan = rparam(10)	! non linear relaxation parameter
c
c ==========================================================================================
c 	Build the multigrid data structure in array iz
c ==========================================================================================
c
	call build_structure(nx,ny,nz,nlevel,iz)
c
c ==========================================================================================
c 	Build operator and right-hand-size on fine grid
c ==========================================================================================
c
	ido = 0
	call build_gpbl(nx,ny,nz,nlevel,ipkey,ido,iz,
     1	ipc,rpc,cc,fc,xpos,ypos,zpos,
     2	gxcf,gycf,gzcf,ccf,fcf,grid_info)
c
c ==========================================================================================
c 	Prepare for SOR solver and solve
c ==========================================================================================
c
c	Determine machine epsilon (hard coded)
c ==========================================================================================
c
c	epsilon = epsmach(0)
	epsilon = 1.d-9
c
c	Impose zero Dirichlet boundary conditions
c ==========================================================================================
c
	call set_bound_zero(nx,ny,nz,u)
c
c	Call nonlinear iterative solver
c ==========================================================================================
c
	iok = 1
c
        call direct_driver_gpbl(nx,ny,nz,u,w0,a1cf,a2cf,istop,itmax,
     1          iters,ierror,iok,iinfo,epsilon,errtol,omegan,mgsmoo,
     2          ipc,rpc,vect,cc,fc,xpos,ypos,zpos,grid_info)
c
c	Restore boundary conditions
c ==========================================================================================
c
	call set_bound(nx,ny,nz,u,gxcf,gycf,gzcf)
c
c ==========================================================================================
c 	Return and end
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	direct_driver_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c	Newton_vcycle is the inexact Newton multilevel solver
c
	subroutine direct_driver_gpbl(nx,ny,nz,x,r,w1,w2,istop,itmax,
     1			iters,ierror,
     2			iok,iinfo,epsilon,errtol,omega,mgsmoo,
     3			ipc,rpc,vect,cc,fc,xpos,ypos,zpos,grid_info)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer	nx,ny,nz
	integer	istop,itmax,iters,ierror,mgsmoo
	integer	mgsolv,iok,iinfo
c
	integer	ipc(*)
c
        integer*1 grid_info(*)
c
	real*8	epsilon,errtol,omega
	real*8	x(*),r(*),w1(*),w2(*)
	real*8	rpc(*),vect(*),cc(*),fc(*)
        real*8  xpos(*),ypos(*),zpos(*)
c
c	Local variables
c ==========================================================================================
c
        integer iresid,iadjoint,itmax_s,iters_s
c
	real*8	xnorm1
	real*8	rsden,rsnrm,orsnrm
	real*8	errtol_s,omega_s
c
c ==========================================================================================
c	Initialization
c ==========================================================================================
c
c	Start clock and printout
c ==========================================================================================
c	
	if(iok.ne.0) then
		call print_ini
c		call print_residual(iok,-1,0.0d0,0.0d0,0.0d0)
	endif
c
c	Compute denominator for stopping criteria
c ==========================================================================================
c	
c	We only implement the relative residual
c
	if(istop.eq.1) then
c
c		Compute initial residual with zero initial guess
c		This is analogous to simply taking the norm of the rhs of the
c		equation
c
		call set_all_zeros(nx,ny,nz,w1)
		call residue_gpbl(x,nx,ny,nz,
     1          xpos,ypos,zpos,grid_info,vect,fc,cc,r)
		rsden = xnorm1(nx,ny,nz,r)
                write(6,*) 'Initial residual :',rsden
	else
		write(6,*) 'Bad istop value ....'
		stop
	endif
c
	if(rsden.eq.0.0d0) then
		rsden = 1.d0
		write(6,*) 'Warning: rhs is zero ....'
	endif
c
	rsnrm = rsden
	orsnrm = rsnrm
c
	if(iok.ne.0) then
		call print_residual(iok,0,rsnrm,rsden,orsnrm)
	endif
c
c ==========================================================================================
c	iterations
c ==========================================================================================
c
        iters=0
c
100     continue
c
                iters = iters + 10
c
                iresid = 1
                iadjoint = 0
                errtol_s = 0.0d0
                itmax_s = 10
                omega_s = omega
c
                call nonlin_smoothing_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,
     1		x,w1,w2,r,itmax_s,iters_s,errtol_s,omega_s,iresid,
     2		iadjoint,grid_info,xpos,ypos,zpos,mgsmoo)
c
c		Check stopping criteria
c
		orsnrm = rsnrm
		if(istop.eq.1) then
			rsnrm = xnorm1(nx,ny,nz,r)
		else
			write(6,*) 'Bad istop value ....'
		endif
		call print_residual(iok,iters,rsnrm,rsden,
     1			orsnrm)
		if((rsnrm/rsden).le.errtol) goto 400
c
		if(iters.ge.itmax) goto 400
c
	goto 100
c
400	continue
c
	return
	end
