c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Elec_Map.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This file contains a series of subroutines that
c			generate maps based on the electrostatic potential
c			computed from the PDE, and then write these maps
c			in ASCII files, under a variety of possible format.
c
c			types of maps considered:
c				- electrostatic potential
c				- energy density
c				- ion number density
c				- ion charge density
c
c			The formats currently implemented are:
c
c				OpenDX (preferred format by APBS)
c				UHBD
c				XPLOR / CNS map (can be read by pymol)
c				EZD map (in fact, new EZD O format; can be read by pymol)
c				binary
c				Ascii
c
c			All these formats (except binary!) are ASCII
c
c ==========================================================================================
c ==========================================================================================
c	Generate_Map.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This file generates a map based on the current electrostatic
c			potential computed on a grid
c
c	Input:
c		u		electrostatic potential
c		nx,ny,nz	size of the grid
c		map_type	flag defining the map to be computed
c
c	Output:
c		map		computed map
c		title		header for file in which map will be written
c
	subroutine Generate_Map(u,hx,hy,hz,nx,ny,nz,grid_info,map_type,
     1                  map,title,Space)
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks and formats    ==========================
c ==========================================================================================
c
        include 'poisson.h'
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer	map_type
c
	real*8	arg,coef,eps,norm
	real*8	Space(*)
c
	character*72 title
c
c	Information on equation and solver
c ==========================================================================================
c
	real*8	tol
	integer method,nlevel,nlin,equation,ndiscrete,ijac,nsmooth
c
c
c	Potential and map
c ==========================================================================================
c
	real*8 nwater
	real*8	u(nx,ny,nz),map(nx,ny,nz)
c
c
c	Information on ions and dielectric constants
c ==========================================================================================
c
	integer	i_ion,n_ion,n_zz
c
	real*8	eps1,eps2
	real*8	ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8	kappa(max_ion)
c
c	Information on grid
c ==========================================================================================
c
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Partition of the grid
c
	integer*1	grid_info(nx,ny,nz)
c
c	Constants
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
c	Common blocks
c ==========================================================================================
c
	common /elec_const/ 	eps1,eps2
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
c
	common /pdesolver/	tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1			   	Avogadro,eps0,beta,C_charge,C_prop
	common /numwater/ nwater
c
c	
	nwater = 0.d0
c	
c ==========================================================================================
c =    Compute maps
c ==========================================================================================
c
	if(map_type.eq.1) then
c
c		This is the simplest case: directly outputs the electrostatic potential
c
		do 300 k = 1,nz
		    do 200 j = 1,ny
			do 100 i = 1,nx
				map(i,j,k) = u(i,j,k)
100			continue
200		    continue
300		continue
c
		title='Electrostatic potential (kT/e)'
c
	elseif(map_type.eq.2) then
c
c		Here we compute the energy density as ED = 0.5 * eps * E^2
c		where eps is the dielectric constant, and E is the electric field
c
                call energy_density(u,hx,hy,hz,nx,ny,nz,grid_info,map,
     1          title,Space)
c
		title='Electrostatics energy density ((kT/eA)^2)'
c
	elseif(map_type.eq.3) then
c
c		Compute the ion number density (in mol/l)
c
                if(n_ion.eq.0) then
                        write(6,*) 'No ion map generated !!'
                        return
                endif
c
                if(equation.eq.1.or.equation.eq.2) then
                        call ion_density_pbe(u,nx,ny,nz,grid_info,map,
     1                  title)
                elseif(equation.eq.3) then
                        call ion_density_smpb(u,nx,ny,nz,grid_info,map,
     1                  title)
                elseif(equation.eq.4) then
                        call ion_density_gpbl(u,hx,hy,hz,nx,ny,nz,
     1                  grid_info,map,title,Space)
                endif
c
	elseif(map_type.eq.4) then
c
c		Compute the ion charge density (in e.mol/l)
c
                if(n_ion.eq.0) then
                        write(6,*) 'No ion map generated !!'
                        return
                endif
c
                if(equation.eq.1.or.equation.eq.2) then
                        call charge_density_pbe(u,nx,ny,nz,grid_info,
     1                  map,title)
                elseif(equation.eq.3) then
                        call charge_density_smpb(u,nx,ny,nz,grid_info,
     1                  map,title)
                elseif(equation.eq.4) then
                        call charge_density_gpbl(u,hx,hy,hz,nx,ny,nz,
     1                  grid_info,map,title,Space)
                endif
c
        elseif(map_type.eq.5) then
c
c               Compute the water dipole density (in mol/l)
c
                if(equation.lt.4) then
                        write(6,*) 'No water density map generated !!'
                        return
                endif
c
                call water_density_gpbl(u,hx,hy,hz,nx,ny,nz,grid_info,
     1          map,title,Space,nwater)
c
	endif
c
	return
	end
c 
c ==========================================================================================
c ==========================================================================================
c	WriteMap_bin.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine writes a map in binary format
c       (internal usage)
c
c
	subroutine WriteMap_bin(l_map,map,hx,hy,hz,nx,ny,nz)
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks and formats    ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer	l_map
c
c	Information on grid
c ==========================================================================================
c
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Map to be written out
c ==========================================================================================
c
	real*8	map(nx,ny,nz)
c
        write(l_map) nx,ny,nz
        write(l_map) (hx(i),i=1,nx)
        write(l_map) (hy(i),i=1,ny)
        write(l_map) (hz(i),i=1,nz)
c
        write(l_map) (((map(i,j,k),i=1,nx),j=1,ny),k=1,nz)
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	WriteMap_ascii.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine writes a map in Ascii format
c       (internal usage)
c
c
	subroutine WriteMap_ascii(l_map,map,hx,hy,hz,nx,ny,nz)
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks and formats    ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer	l_map
c
c	Information on grid
c ==========================================================================================
c
	real*8	hx(nx),hy(ny),hz(nz)
c
1	format(10(f8.3,2x))
c
c	Map to be written out
c ==========================================================================================
c
	real*8	map(nx,ny,nz)
c
        write(l_map,*) nx,ny,nz
c
        write(l_map,1) (hx(i),i=1,nx)
        write(l_map,1) (hy(i),i=1,ny)
        write(l_map,1) (hz(i),i=1,nz)
c
	do 300 i = 1,nx
		do 200 j = 1,ny
			do 100 k = 1,nx
				write(l_map,*) map(i,j,k)
100			continue
200		continue
300	continue
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	WriteMap_OpenDX.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine writes a map in OpenDX format
c
c
	subroutine WriteMap_OpenDx(l_map,map,hx,hy,hz,nx,ny,nz,
     1          origin,title)
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks and formats    ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer	l_map
c
	character title*72
c
c	Map to be written out
c ==========================================================================================
c
	real*8	origin(3)
	real*8	map(nx,ny,nz)
c
c	Information on grid
c ==========================================================================================
c
	real*8	sizex,sizey,sizez
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Partition of the grid
c
c	Formats (most information) is OpenDX specific
c ==========================================================================================
c
1	format('# Data from GenSolv V 0.1',/,'#')
2	format('#')
3	format('object 1 class gridpositions counts',1x,i3,1x,i3,1x,i3)
4	format('origin ',1pe15.6,1x,1pe15.6,1x,1pe15.6)
5	format('delta ',1pe15.6,1x,1pe15.6,1x,1pe15.6)
6	format('object 2 class gridpositions counts',1x,i3,1x,i3,1x,i3)
7	format(
     1  'object 3 class array type double rank 0 items',i8,
     2  ' data follows')
8	format(1x,1pe15.6,1x,1pe15.6,1x,1pe15.6)
9	format('attribute "dep" string "positions"')
10	format(
     1	'object "regular positions regular connections" class field')
11	format('component "positions" value 1')
12	format('component "connections" value 2')
13	format('component "data" value 3')
14	format('# ',a72)
c
        sizex = hx(2) - hx(1)
        sizey = hy(2) - hy(1)
        sizez = hz(2) - hz(1)
c
c ==========================================================================================
c =    Write file in OpenDX format                                ==========================
c ==========================================================================================
c
c	Write headers
c ==========================================================================================
c
	write(l_map,1)
c
	write(l_map,14) title
	write(l_map,2)
c
	write(l_map,3) nx,ny,nz
	write(l_map,4) (origin(i),i=1,3)
	write(l_map,5) sizex,0.d0,0.d0
	write(l_map,5) 0.d0,sizey,0.d0
	write(l_map,5) 0.d0,0.d0,sizez
	write(l_map,6) nx,ny,nz
	write(l_map,7) nx*ny*nz
c
c	Write map itself
c ==========================================================================================
c
	write(l_map,8) (((map(i,j,k),k=1,nz),j=1,ny),i=1,nx)
c
c	Write trailer
c ==========================================================================================
c
	write(l_map,9)
	write(l_map,10)
	write(l_map,11)
	write(l_map,12)
	write(l_map,13)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	WriteMap_UHBD.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine writes a map in UHBD format
c
c
	subroutine WriteMap_UHBD(l_map,map,hx,hy,hz,nx,ny,nz,
     1                  origin,title)
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks and formats    ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer	l_map
c
	real*8	one,zero
c
	character	title*72
c
c	Map to be written out
c ==========================================================================================
c
	real*8	origin(3)
	real*8	map(nx,ny,nz)
c
c	Information on grid
c ==========================================================================================
c
	real*8	sizex,sizey,sizez
	real*8	hx(nx),hy(ny),hz(nz)
c
c
c	Formats (most information) is UHBD specific
c ==========================================================================================
c
1	format(a72)
2	format(1pe12.5,1pe12.5,5(i7))
3	format(3(i7),1pe12.5,1pe12.5,1pe12.5,1pe12.5)
4	format(1pe12.5,1pe12.5,1pe12.5,1pe12.5)
5	format(1pe12.5,1pe12.5,i7,i7)
6	format(3(i7))
7	format(6(1pe12.5))
c
        sizex = hx(2) - hx(1)
        sizey = hy(2) - hy(1)
        sizez = hz(2) - hz(1)
c
c ==========================================================================================
c =    Write file in UHBD format                                ==========================
c ==========================================================================================
c
	one = 1.d0
	zero = 0.d0
c
c	UBBD was written for symmetric meshes only...
c ==========================================================================================
c
	if(sizex.ne.sizey.or.sizey.ne.sizez) then
		write(6,*) ' '
		write(6,*) 'Warning: could not generate UHBD file' 
		write(6,*)  'UHBD format is only for uniform grid'
		write(6,*) ' '
		return
	endif
c
c	Write headers
c ==========================================================================================
c
	write(l_map,1) title
	write(l_map,2) one,zero,-1,0,nz,1,nz
	write(l_map,3) nx,ny,nz,sizex,origin(1)-sizex,origin(2)-sizex,
     1	origin(3)-sizex
	write(l_map,4) zero,zero,zero,zero
	write(l_map,5) zero,zero,0,0
c
c	Write map
c ==========================================================================================
c
	do 300 k = 1,nz
		write(l_map,6) k,nx,ny
		write(l_map,7) ((map(i,j,k),i=1,nx),j=1,ny)
300	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	WriteMap_Xplor.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine writes a map in Xplor/CNS format
c
c
	subroutine WriteMap_Xplor(l_map,map,hx,hy,hz,nx,ny,nz,
     1                  origin,title)
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks and formats    ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer nvalxmin,nvalxmax,nvalymin,nvalymax,nvalzmin,nvalzmax
	integer	l_map,nval
c
	real*8	a,b,c,alpha,beta,gamma
	real*8	x,map_mean,map_var,map_std
c
	character title*72
c
c	Map to be written out
c ==========================================================================================
c
	real*8	origin(3)
	real*8	map(nx,ny,nz)
c
c	Information on grid
c ==========================================================================================
c
	real*8	sizex,sizey,sizez
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Formats (most information) is Xplor specific
c ==========================================================================================
c
2	format(/,'      1')
4	format('REMARKS:',a72)
5	format(9(i8))
6	format(6(1pe12.5))
7	format('ZYX')
8	format(i8)
9	format(6(1pe12.5))
10	format(2(1pe12.4,1x))
c
        sizex = hx(2) - hx(1)
        sizey = hy(2) - hy(1)
        sizez = hz(2) - hz(1)
c
c ==========================================================================================
c =    Generate variables for X-plor format                       ==========================
c ==========================================================================================
c
c	Write the whole grid
c
	nvalxmin = -(nx-1)/2
	nvalxmax = (nx-1)/2
	nvalymin = -(ny-1)/2
	nvalymax = (ny-1)/2
	nvalzmin = -(nz-1)/2
	nvalzmax = (nz-1)/2
c
	a = sizex*nx
	b = sizey*ny
	c = sizez*nz
c
	alpha = 90.d0
	beta  = 90.d0
	gamma = 90.d0
c
c ==========================================================================================
c =    Write file in XPLOR format                                 ==========================
c ==========================================================================================
c
c	Write headers
c ==========================================================================================
c
	write(l_map,2)
	write(l_map,4) title
	write(l_map,5) nx,nvalxmin,nvalxmax,ny,nvalymin,nvalymax,
     1		nz,nvalzmin,nvalzmax
	write(l_map,6) a,b,c,alpha,beta,gamma
	write(l_map,7)
c
c	Write map itself
c ==========================================================================================
c
	do 100 k = 1,nz
		write(l_map,8) k-1 -(nz-1)/2
		write(l_map,9) ((map(i,j,k),i=1,nx),j=1,ny)
100	continue
c
c	Compute mean and std values of the map
c ==========================================================================================
c
	map_mean = 0
	map_var  = 0
	do 400 k = 1,nz
		do 300 j = 1,ny
			do 200 i = 1,nx
				x = map(i,j,k)
				map_mean = map_mean + x
				map_var  = map_var + x*x
200			continue
300		continue
400	continue
c
        nval = nx*ny*nz
	map_mean = map_mean/nval
	map_var = map_var/nval - map_mean*map_mean
	map_std = sqrt(map_var)
c
c	Write trailer
c ==========================================================================================
c
	write(l_map,8) -9999
	write(l_map,10) map_mean,map_std
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	WriteMap_EZD.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine writes a map in EZD format (in fact, new EZD format
c			of O)
c
c
	subroutine WriteMap_EZD(l_map,map,hx,hy,hz,nx,ny,nz,
     1                  origin,title)
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks and formats    ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer nvalxmin,nvalxmax,nvalymin,nvalymax,nvalzmin,nvalzmax
	integer	l_map,nchar,nval,nline
c
	real*8	a,b,c,alpha,beta,gamma
	real*8	scale
c
	character title*72
	character line*100,newline*100
c
c	Map to be written out
c ==========================================================================================
c
	real*8	origin(3)
	real*8	map(*)
c
c	Information on grid
c ==========================================================================================
c
	real*8	sizex,sizey,sizez
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Formats (most information) is EZD specific
c ==========================================================================================
c
1	format('EZD_MAP')
2	format('! ',a72)
3	format('CELL',6(f10.3))
4	format('ORIGIN',3(i4))
5	format('EXTENT',3(i4))
6	format('GRID',3(i4))
7	format('SCALE',f8.3)
8	format('MAP')
9	format(7(f8.0))
10	format('END')
11	format(a)
c
        sizex = hx(2) -hx(1)
        sizey = hy(2) -hy(1)
        sizez = hz(2) -hz(1)
c
c ==========================================================================================
c =    Generate variables for EZD format                          ==========================
c ==========================================================================================
c
c	Write the whole grid
c
	nvalxmin = -(nx-1)/2
	nvalxmax = nx
	nvalymin = -(ny-1)/2
	nvalymax = ny
	nvalzmin = -(nz-1)/2
	nvalzmax = nz
c
	a = sizex*nx
	b = sizey*ny
	c = sizez*nz
	alpha = 90.d0
	beta  = 90.d0
	gamma = 90.d0
c
	scale = 1.d0
c
c ==========================================================================================
c =    Write file in new EZD format                               ==========================
c ==========================================================================================
c
c	Write headers
c ==========================================================================================
c
	write(l_map,1)
	write(l_map,3) a,b,c,alpha,beta,gamma
	write(l_map,4) nvalxmin,nvalymin,nvalzmin
	write(l_map,5) nvalxmax,nvalymax,nvalzmax
	write(l_map,6) nx,ny,nz
	write(l_map,7) scale
	write(l_map,8) 
c
c	Write map itself
c ==========================================================================================
c
	nval = nx*ny*nz
	nline = nval/7 + 1
	if(mod(nval,7).eq.0) nline = nline -1
c
c	New EZD format assumes that the map is written in Fortran, with
c	first index running first, then second index, then third index
c	
	do 100 i = 1,nline-1
		write(l_map,9) (map(k),k=7*(i-1)+1,7*i)
100	continue
	write(l_map,9) (map(k),k=7*(nline-1)+1,nval)
c
c	Write trailer
c ==========================================================================================
c
	write(l_map,10) 
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	pretty_line.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine makes a line "pretty": it removes extra white spaces that are
c	not needed
c
c	Example: "This     is           a     white           line"
c	Becomes: "This is a white line"
c
	subroutine pretty_line(line,newline,ipos)
c
	integer	i,j
	integer	ipos,istart
c
	character	line*100, newline*100
c
	ipos   = 0
	istart = 0
c
	do 100 i = 1,80
c
		if(line(i:i).ne.' ') then
			ipos = ipos + 1
			newline(ipos:ipos) = line(i:i)
			istart = 0
		else
			if(istart.eq.0) then
				ipos = ipos + 1
				newline(ipos:ipos) = line(i:i)
				istart = 1
			endif
		endif
c
100	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Elec_Map.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine generates a map, and writes it in a file
c
c	Input:
c		u		: electrostatics potential
c		map		: array to store map to be constructed
c		nx,ny,nz 	: grid size
c		map_type	: type of map:
c					1: electrostatics potential
c					2: energy density
c					3: ion density (number)
c					4: ion density (charge)
c                                       5: Water dipole density (GPBL only)
c		l_map		: logical unit for file in which map is written
c		flag_format	: format for file:
c					1: OpenDx
c					2: UHBD
c					3: Xplor
c					4: EZD
c		Origin		: position of front, left bottom corner of grid
c
	subroutine Elec_Map(u,hx,hy,hz,nx,ny,nz,map_type,map,l_map,
     1          flag_format,Origin,grid_info,Space)
c
c ==========================================================================================
c =    Setup: declare variables, set formats                      ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	nx,ny,nz,map_type,l_map,flag_format
c
	real*8	Origin(3)
	real*8	u(*),map(*),Space(*)
        real*8  hx(nx),hy(ny),hz(nz)
	real*8 nwater
c
	integer*1	grid_info(*)
c
	character	title*72
c	
	common /numwater/ nwater
c
c ==========================================================================================
c =    Generate the map                                           ==========================
c ==========================================================================================
c
	call Generate_Map(u,hx,hy,hz,nx,ny,nz,grid_info,
     1                  map_type,map,title,Space)
c
c ==========================================================================================
c =    Print out the map                                          ==========================
c ==========================================================================================
c
	if(flag_format.eq.1) then
c
c	OpenDx format
c
		call WriteMap_OpenDx(l_map,map,hx,hy,hz,nx,ny,nz,
     1                                  Origin,title)
c
	elseif(flag_format.eq.2) then
c
c	UHBD format
c
		call WriteMap_UHBD(l_map,map,hx,hy,hz,nx,ny,nz,
     1                                  Origin,title)
c
	elseif(flag_format.eq.3) then
c
c	Xplor/CNS format
c
		call WriteMap_Xplor(l_map,map,hx,hy,hz,nx,ny,nz,
     1                                  Origin,title)
c
	elseif(flag_format.eq.4) then
c
c	EZD (new) format for O (pymol can also read this format)
c
		call WriteMap_EZD(l_map,map,hx,hy,hz,nx,ny,nz,
     1                                  Origin,title)
c
        elseif(flag_format.eq.5) then
c
c       Simple binary format
c
		call WriteMap_bin(l_map,map,hx,hy,hz,nx,ny,nz)
c
        elseif(flag_format.eq.6) then
c
c       Simple ascii format
c
		call WriteMap_ascii(l_map,map,hx,hy,hz,nx,ny,nz)
c
	endif
c
	return
	end	
c ==========================================================================================
c ==========================================================================================
c	Write_Mol_Center.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine generates a new PDB file, in which the molecule
c			has been centered.
c
c	In XPLOR and EZD, the grid is centered on 0,0,0 and can only be translated
c	by integer units of the grid size. It is therefore not possible to match exactly
c	the grid center to the molecule center...unless we keep the grid centered,
c	and re-center the molecule!
c	This is what is done in this subroutine
c
c	Input:
c		lout:	logical unit for new PDB file
c
	subroutine Write_Mol_Center(l_out)
c
	include 'poisson.h'
c
c ==========================================================================================
c =    Setup: declare variables, set formats                      ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,natom,ncharge,l_out
	integer iform
c
	real*8	cg(3),c_grid(3)
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30,atom_label2(*)*12
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        common /charges/ ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                   ptr_label2,iform,natom,ncharge
c
	common /centers/ cg,c_grid
c
c ==========================================================================================
c = Output format (to write in log file)  ==================================================
c ==========================================================================================
c
1	format(a30,3f8.3,a12)
2	format('REMARK: File generated by GenSolv v 0.1')
3	format('REMARK: PDB file for molecule centered on grid center')
4	format('REMARK')
5	format('END')
c
c ==========================================================================================
c = Generate PDB file                     ==================================================
c ==========================================================================================
c
c	Write header
c ==========================================================================================
c
	write(l_out,2)
	write(l_out,3)
	write(l_out,4)
c
c	Write coordinates
c ==========================================================================================
c
	do 100 i = 1,natom
		write(l_out,1) atom_label(i),
     1		(atom_coord(3*(i-1)+j),j=1,3),atom_label2(i)
100	continue
c
c	Write trailer    
c ==========================================================================================
c
	write(l_out,5)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Write_Ene_Atom.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/04/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine write the energy for each atom
c
c	Input:
c		lout:	logical unit for new PDB file
c               ene_atom: energy for each atom
c
	subroutine Write_Ene_Atom(l_out,ene_atom)
c
	include 'poisson.h'
c
c ==========================================================================================
c =    Setup: declare variables, set formats                      ==========================
c ==========================================================================================
c
c       Local variables
c ==========================================================================================
c
	integer	i,j,natom,ncharge,l_out
	integer iform
c
	real*8	cg(3),c_grid(3)
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
        real*8  ene_atom(*)
c
	character atom_label(*)*30,atom_label2(*)*12
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        common /charges/ ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                   ptr_label2,iform,natom,ncharge
c
	common /centers/ cg,c_grid
c
c ==========================================================================================
c = Output format (to write in log file)  ==================================================
c ==========================================================================================
c
1	format(a30,3f8.3,2x,f10.3)
2	format('REMARK: File generated by AquaSol v 1.0')
3	format('REMARK: 0.5*Grid energy per atom (kT)')
4	format('REMARK')
5	format('END')
c
c ==========================================================================================
c = Generate PDB file                     ==================================================
c ==========================================================================================
c
c	Write header
c ==========================================================================================
c
	write(l_out,2)
	write(l_out,3)
	write(l_out,4)
c
c	Write coordinates
c ==========================================================================================
c
	do 100 i = 1,natom
		write(l_out,1) atom_label(i),
     1		(atom_coord(3*(i-1)+j),j=1,3),ene_atom(i)
100	continue
c
c	Write trailer    
c ==========================================================================================
c
	write(l_out,5)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Get_efield.f
c ==========================================================================================
c ==========================================================================================
c
        subroutine get_efield(u,nx,ny,nz,Ex,Ey,Ez)
c
        integer i,j,k,nx,ny,nz
c
        real*8  u(nx,ny,nz)
        real*8  Ex(nx,ny,nz)
        real*8  Ey(nx,ny,nz)
        real*8  Ez(nx,ny,nz)
c
        do 300 k = 2,nz-1
c
                do 200 j = 2,ny-1
c
                        do 100 i = 2,nx-1
c
                                Ex(i,j,k)=u(i+1,j,k)-u(i-1,j,k)
                                Ey(i,j,k)=u(i,j+1,k)-u(i,j-1,k)
                                Ez(i,j,k)=u(i,j,k+1)-u(i,j,k-1)
c
100                     continue
c
200             continue
c
300     continue
c
        return
        end
