c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	FAS_Solver_pbe.f
c       (FAS-based PDE solver for PB equations, based on Michael Holst MG package)
c
c	Version 1:      7/4/07
c
c	Author: Patrice Koehl (in collaboration with Marc Delarue)
c
c	This file contains a series of subroutines for solving modified Poisson-Boltzmann
c	equations. It is inspired form Michael Holst MG code, which is available at:
c	http://www.fetk.org/codes/pmg/index.html
c
c ==========================================================================================
c ==========================================================================================
c 	FAS_Solver_pbe.f
c ==========================================================================================
c ==========================================================================================
c
c	FAS_Solver (where FAS stands for "Full Approximation Scheme" 
c	implements a nonlinear Multilevel solver  to the problem
c	of solving the Poisson-Boltzmann equation.
c
	subroutine FAS_Solver_pbe(iparam,rparam,iwork,rwork,u,
     1				xpos,ypos,zpos,bndx,bndy,bndz,
     2				eps_X,eps_Y,eps_Z,helm,rhs)
c
c	Input:
c		iparam:		integer array: gives flags to Solver
c		rparam:		real array: constants for solver
c		iwork		integer work array
c		rwork		real work array
c
c	One dimensional:
c		xpos		positions of grid points along X axis
c		ypos		positions of grid points along Y axis
c		zpos		positions of grid points along Z axis
c
c	Two dimensional:
c		bndx		boundary conditions (X=1 and X=Nx planes)
c		bndy		boundary conditions (Y=1 and Y=Ny planes)
c		bndz		boundary conditions (Z=1 and Z=Nz planes)
c
c	Three dimensional:
c		eps_X		epsilon map, defined at positions (i+0.5,j,k)
c		eps_Y		epsilon map, defined at positions (i,j+0.5,k)
c		eps_Z		epsilon map, defined at positions (i,j,k+0.5)
c
c		helm		Helmholtz term in the equation
c		rhs		right hand side of the equation
c
c	Output:
c		u		the dimensionless electrostatic potential map
c
c	Dimensions of all arrays have been set in the calling program
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer		iparam(*),iwork(*)
	real*8		rparam(*),rwork(*),u(*)
	real*8		xpos(*),ypos(*),zpos(*),bndx(*),bndy(*),bndz(*)
	real*8		eps_X(*),eps_Y(*),eps_Z(*),helm(*),rhs(*)
c
c	Local variables:
c ==========================================================================================
c
	integer	i,nx,ny,nz,niwk,nrwk,nlevel,ierror,maxlevel,mxlevel
	integer	mgcoar,mgdisc,mgsolv,mgsten
	integer k_iz
	integer k_ipc,k_rpc,k_ac,k_cc,k_fc
c
c	Variables from multigrid size
c ==========================================================================================
c
	integer	nxc,nyc,nzc,nf,nc,narr,narrc
	integer	n_rpc,n_ipc,n_iz,ireal_tot,iint_tot
c
c ==========================================================================================
c 	Decode input array iparam, and do some sanity check
c ==========================================================================================
c
	nrwk	= iparam(1)
	niwk	= iparam(2)
	nx	= iparam(3)
	ny	= iparam(4)
	nz	= iparam(5)
	nlevel	= iparam(6)
	mgcoar	= iparam(18)
	mgdisc	= iparam(19)
	mgsolv	= iparam(21)
c
	if(nlevel.le.0.or.nx.le.0.or.ny.le.0.or.nz.le.0) then
		write(6,*) 
     1          'nx,ny,nz and nlevel must be strictly positive!!'
		ierror = -1
		iparam(51) = ierror
		return
	endif
c
c	Check that the number of levels proposed is compatible with the grid size
c
	mxlevel = maxlevel(nx,ny,nz)
	if(nlevel.gt.mxlevel) then
		write(6,*) 'The max. # of levels for your grid is: ',
     1			mxlevel
        nlevel = mxlevel
c		ierror = -2
c		iparam(51) = ierror
c		return
	endif
c
c ==========================================================================================
c	Get grid size info for all grid levels: this will be used to partition the arrays
c	iwork and rwork so that they can store the whole hierarchy of grids
c ==========================================================================================
c
        mgsten = 0
	call multigrid_size(mgsten,mgcoar,mgdisc,mgsolv,nx,ny,nz,nlevel,
     1		nxc,nyc,nzc,nf,nc,narr,narrc,n_rpc,n_iz,n_ipc,
     2		ireal_tot,iint_tot)
c
c ==========================================================================================
c       multigrid_size assumes we need 2 storage arrays of size nf
c       and two storage arrays of size narr: in fact, we do not need
c       the arrays of size nf for FAS: apply correction
c ==========================================================================================
c
        ireal_tot = ireal_tot - 2*nf
c
c	Check that work arrays (both real and integer) have enough space
c ==========================================================================================
c
	if(nrwk.lt.ireal_tot) then
		write(6,*)'Real work space must be at least: ',ireal_tot
		ierror = -3
		iparam(51) = ierror
		return
	endif
c
	if(niwk.lt.iint_tot) then
		write(6,*) 'Integer work space must be at least: ',
     1			iint_tot
		ierror = -3
		iparam(51) = ierror
		return
	endif
c
c       Initialize work array
c ==========================================================================================
c
	do 100 i = 1,ireal_tot
		rwork(i) = 0
100	continue
c
c	Split the integer work array
c ==========================================================================================
c
	k_iz = 1
	k_ipc = k_iz + n_iz
c
c	Split the real work array
c ==========================================================================================
c
	k_rpc = 1
	k_cc  = k_rpc + n_rpc
	k_fc  = k_cc + narr
	k_ac  = k_fc + narr
c
c ==========================================================================================
c	Start the multigrid solver
c ==========================================================================================
c
	call fas_mg_pbe(iparam,rparam,nx,ny,nz,u,iwork(k_iz),
     1		iwork(k_ipc),rwork(k_rpc),
     2		rwork(k_ac),rwork(k_cc),rwork(k_fc),
     3		xpos,ypos,zpos,bndx,bndy,bndz,eps_X,eps_Y,eps_Z,helm,
     4		rhs)
c
c ==========================================================================================
c	The program should have found the solution...return to the
c	main program
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	fas_mg_pbe.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine uses the full approximation scheme, i.e. a non linear
c	multilevel solver, to solve modified Poisson-Boltzmann equations.
c
	subroutine fas_mg_pbe(iparam,rparam,nx,ny,nz,u,iz,
     1		ipc,rpc,ac,cc,fc,xpos,ypos,zpos,gxcf,gycf,gzcf,
     2		a1cf,a2cf,a3cf,ccf,fcf)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer		iparam(*),ipc(*),iz(*)
	real*8		rparam(*),rpc(*),ac(*),cc(*),fc(*)
	real*8		u(*)
	real*8		xpos(*),ypos(*),zpos(*),gxcf(*),gycf(*),gzcf(*)
	real*8		a1cf(*),a2cf(*),a3cf(*),ccf(*),fcf(*)
c
c	Local variables:
c ==========================================================================================
c
	integer		mgkey,nlevel,itmax,iok,iinfo,istop,ipkey,nu1,nu2
	integer		nx,ny,nz,ilev,ido,iters,ierror,nlevel_real
	integer		mgprol,mgcoar,mgsolv,mgdisc,mgsmoo
c
c	real*8		epsmach
	real*8		epsilon,errtol,omegal,omegan
c
c ==========================================================================================
c 	Decode the iparam and rparam arrays
c ==========================================================================================
c
	nlevel = iparam(6)	! Number of mesh level
	nu1    = iparam(7)	! number of pre-smoothings
	nu2    = iparam(8)	! number of post-smoothings
	mgkey  = iparam(9)	! multigrid method: 0=variable, v-cycle, 1=nest operation
	itmax  = iparam(10)	! maximum number of iterations
	istop  = iparam(11)	! stopping criteria
	iinfo  = iparam(12)	! runtime status messages :
				! 0=none, 1=some, 2=many, 3=more
	ipkey  = iparam(14)	! Defines type of equation considered
				! 0: Laplace, 1: linear PB, 2: non linear PB
				! 3: Size modified PB, 4: Generalized PB Langevin
	mgprol = iparam(17)	! type of prolongation (0=trilinear)
	mgcoar = iparam(18)	! coarse equations (0=stan, 1=harm, 2=Galerkin)
	mgdisc = iparam(19)	! discretization (0=fv, 1=fe)
	mgsmoo = iparam(20)	! type of smoothing (0=wjac, 1=gsrb, 2=sor, 3=rich, 4=cghs)
	mgsolv = iparam(21)	! coarse equation solver (0=cghs, 1=nanded Linpack)
c
	errtol = rparam(1)	! tolerance factor
	omegal = rparam(9)	! linear relaxation parameter
	omegan = rparam(10)	! non linear relaxation parameter
c
c ==========================================================================================
c 	Build the multigrid data structure in array iz
c ==========================================================================================
c
	call build_structure(nx,ny,nz,nlevel,iz)
c
c ==========================================================================================
c 	Build operator and right-hand-size on fine grid
c ==========================================================================================
c
	ido = 0
	call build_operator(nx,ny,nz,nlevel,ipkey,iinfo,ido,iz,mgprol,
     1	mgcoar,mgsolv,mgdisc,ipc,rpc,ac,cc,fc,xpos,ypos,zpos,
     2	gxcf,gycf,gzcf,a1cf,a2cf,a3cf,ccf,fcf)
c
c ==========================================================================================
c 	Build operator and right-hand-size on all coarse grids
c ==========================================================================================
c
	ido = 1
	call build_operator(nx,ny,nz,nlevel,ipkey,iinfo,ido,iz,mgprol,
     1	mgcoar,mgsolv,mgdisc,ipc,rpc,ac,cc,fc,xpos,ypos,zpos,
     2	gxcf,gycf,gzcf,a1cf,a2cf,a3cf,ccf,fcf)
c
c ==========================================================================================
c 	Prepare for FAS solver and solve
c ==========================================================================================
c
c	Determine machine epsilon (hard coded)
c ==========================================================================================
c
c	epsilon = epsmach(0)
	epsilon = 1.d-9
c
c	Impose zero Dirichlet boundary conditions
c ==========================================================================================
c
	call set_bound_zero(nx,ny,nz,u)
c
c	Call multigrid solver
c ==========================================================================================
c
	nlevel_real = nlevel
	iok = 1
	ilev = 1
c
	if(mgkey.eq.0) then
		call fas_vcycle_pbe(nx,ny,nz,u,iz,a1cf,a2cf,a3cf,ccf,
     1          fcf,istop,itmax,iters,ierror,nlevel,ilev,nlevel_real,
     2		mgsolv,iok,iinfo,epsilon,errtol,omegan,nu1,nu2,
     3		mgsmoo,ipc,rpc,ac,cc,fc)
	elseif(mgkey.eq.1) then
		call fas_nested_pbe(nx,ny,nz,u,iz,a1cf,a2cf,a3cf,ccf,
     1          fcf,istop,itmax,iters,ierror,nlevel,ilev,nlevel_real,
     2		mgsolv,iok,iinfo,epsilon,errtol,omegan,nu1,nu2,
     3		mgsmoo,ipc,rpc,ac,cc,fc)
	else
		write(6,*) 'Bad mgkey!'
	endif
c
c	Restore boundary conditions
c ==========================================================================================
c
	call set_bound(nx,ny,nz,u,gxcf,gycf,gzcf)
c
c ==========================================================================================
c 	Return and end
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	fas_nested_pbe.f
c ==========================================================================================
c ==========================================================================================
c
c	Newton_nested uses nested iteration for an inexact Newton multilevel method
c
	subroutine fas_nested_pbe(nx,ny,nz,u,iz,w0,w1,w2,w3,w4,
     1			istop,itmax,iters,ierror,nlevel,ilev,
     2			nlevel_real,mgsolv,
     2			iok,iinfo,epsilon,errtol,omega,nu1,nu2,mgsmoo,
     3			ipc,rpc,ac,cc,fc)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer	nx,ny,nz
	integer	istop,itmax,iters,ierror,level,nlevel,ilev,nlevel_real
	integer	mgsolv,mgsmoo,iok,iinfo,nu1,nu2
c
	integer	ipc(*),iz(50,*)
c
	real*8	epsilon,errtol,omega
	real*8	u(*),w0(*),w1(*),w2(*),w3(*),w4(*)
	real*8	rpc(*),ac(*),cc(*),fc(*)
c
c	Local variables:
c ==========================================================================================
c
	integer	nxf,nyf,nzf,nxc,nyc,nzc
	integer	itmxd,nleveld,iterd,iokd,istpd
c
	real*8	errd
c
c	Recover grid size
c
	nxf = nx
	nyf = ny
	nzf = nz
c
	call make_coarse(nlevel-1,nxf,nyf,nzf,nxc,nyc,nzc)
c
c	Move up grids: interpolate solution, and use Newton_vcycle
c
	do 100 level=nlevel_real,ilev+1,-1
c
		errd = errtol
		itmxd = 1
		nleveld = nlevel_real - level + 1
		iterd = 0
		iokd = iok
		istpd = istop
		call fas_vcycle_pbe(nxc,nyc,nzc,u,iz,w0,w1,w2,w3,w4,
     1		istpd,itmxd,iterd,ierror,nleveld,level,nlevel_real,
     2		mgsolv,iokd,iinfo,epsilon,errd,omega,nu1,nu2,mgsmoo,
     3		ipc,rpc,ac,cc,fc)
c
		call make_fine(1,nxc,nyc,nzc,nxf,nyf,nzf)
c
		call interpolate(nxc,nyc,nzc,nxf,nyf,nzf,u(iz(1,level)),
     1		u(iz(1,level-1)))
c
		nxc = nxf
		nyc = nyf
		nzc = nzf
c
100	continue
c
	level = ilev
	call fas_vcycle_pbe(nxc,nyc,nzc,u,iz,w0,w1,w2,w3,w4,
     1	istop,itmax,iters,ierror,nlevel,level,nlevel_real,mgsolv,
     2	iok,iinfo,epsilon,errtol,omega,nu1,nu2,mgsmoo,
     3		ipc,rpc,ac,cc,fc)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	fas_vcycle_pbe.f
c ==========================================================================================
c ==========================================================================================
c
c	Non Linear multilevel method.
c
c	Input:
c		fine and coarse grid discrete linear operators L_h, L_H
c		fine grid rhs : f_h
c		fine grid approximate solution: u_h
c
c	Output:
c		fine grid improved solution: u_h
c
c	The two grid algorithm is:
c		(1) pre-smoothing	    u1_h = smooth(L_h,f_h,u_h)
c		(2) restrict defect	    d_H = r(L_h(u1_h) - f_h)
c		(3) Solve for correction    c_H = L_H^{-1} d_H
c		(4) prolongate and correct  u2_h = u1_h - p(c_H)
c		(5) post-smoothing	    u_h = smooth(L_h,f_h,u2_h)
c
c	Notes:
c		(1) u1_h must be kept on each level until "c_H" is computed
c		(2) u_h on each level is stored in the "x" array
c		(3) d_H is f_h on the next coarser grid
c		(4) c_h is u_h on the next coarser grid
c		(5) d_H is stored in the r array (must be kept for post-smooth)
c		(6) f_h is stored in the fc array
c		(7) L_h on all levels is stored in the ac array
c		(8) signs maybe reversed; i.e., residual is used in place of the
c		    defect, ...
c
c ==========================================================================================
c
	subroutine fas_vcycle_pbe(nx,ny,nz,x,iz,w0,w1,w2,w3,w4,istop,
     1			itmax,iters,ierror,nlev,ilev,nlevel_real,mgsolv,
     2			iok,iinfo,epsilon,errtol,omega,nu1,nu2,mgsmoo,
     3			ipc,rpc,ac,cc,fc)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer	nx,ny,nz
        integer nxf,nyf,nzf,nxc,nyc,nzc
	integer	istop,itmax,iters,ierror,nlev,ilev,nlevel_real
	integer	mgsolv,iok,iinfo,nu1,nu2,mgsmoo
	integer	nuuu,iadjoint,ivariable_v
c
	integer	ipc(*),iz(50,*)
c
	real*8	epsilon,errtol,omega
	real*8	x(*),w0(*),w1(*),w2(*),w3(*),w4(*)
	real*8	rpc(*),ac(*),cc(*),fc(*)
c
c	Local variables
c ==========================================================================================
c
	integer	level,lev,i
	integer	itmax_s, iters_s, ierror_s, iok_s, iinfo_s,istop_s
	integer	mgsmoo_s,iresid
c
	real*8	xnorm1,dotvect,xnorm_inf
	real*8	rsden,rsnrm,orsnrm
	real*8	xdamp
	real*8	errtol_s
c
1       format(1x,'Warning: problem on the coarse grid: iters_s =',i4)
c ==========================================================================================
c	Initialization
c ==========================================================================================
c
c	Recover level information
c ==========================================================================================
c
	level = 1
	lev  = (ilev-1) + level
c
c	Recover gridsizes
c ==========================================================================================
c
	nxf = nx
	nyf = ny
	nzf = nz
c
	call make_coarse(nlev-1,nxf,nyf,nzf,nxc,nyc,nzc)
c
	if(iok.eq.1) then
                call print_ini
		call print_residual(iok,-1,0.d0,0.d0,0.d0)
	endif
c
c	Compute denominator for stopping criterion
c ==========================================================================================
c
	if(iok.ne.0) then
		if(istop.eq.0) then
			rsden = 1.d0
		elseif(istop.eq.1) then
			rsden = xnorm1(nxf,nyf,nzf,fc(iz(1,lev)))
		else
			write(6,*) 'nlin_mg: bad istop value'
			stop
		endif
		if(rsden.eq.0.d0) then
			rsden = 1.0d0
			write(6,*) 'Warning...in nlin_mg...rhs is 0'
		endif
		rsnrm = rsden
		orsnrm = rsnrm
		if(iok.eq.1) call print_residual(iok,0,rsnrm,rsden,
     1			orsnrm)
	endif
c
c ==========================================================================================
c 	Solve directly if nlev = 1
c ==========================================================================================
c
	if(nlev.eq.1) then
c
c		Solve using ncghs, mgsmoo_s=4 (no residual)
c
		iresid = 0
		iadjoint = 0
		itmax_s = 100
		iters_s = 0
		errtol_s = epsilon
		mgsmoo_s = mgsmoo
		call set_all_zeros(nxf,nyf,nzf,x(iz(1,lev)))
		call nonlin_smoothing(nxf,nyf,nzf,
     1		ipc(iz(5,lev)),rpc(iz(6,lev)),
     2		ac(iz(7,lev)),cc(iz(1,lev)),fc(iz(1,lev)),
     3		x(iz(1,lev)),w1,w2,w3,
     4		itmax_s,iters_s,errtol_s,omega,iresid,iadjoint,
     5		mgsmoo_s)
c
		if(iters_s .ge.itmax_s) then
			write(6,1) iters_s
		endif
c
c		Compute stopping criteria
c
		iters = 1
		if(iok.eq.0) then
			orsnrm = rsnrm
			if(istop.eq.1) then
				call residual(nxf,nyf,nzf,
     1				ipc(iz(5,lev)),rpc(iz(6,lev)),
     2				ac(iz(7,lev)),cc(iz(1,lev)),
     3				fc(iz(1,lev)),x(iz(1,lev)),w1,w2)
				rsnrm = xnorm1(nxf,nyf,nzf,w1)
			else
				write(6,*) 'In fas_vcycle, bad istop..'
				stop
			endif
			if(iok.eq.1) call print_residual(iok,iters,
     1					rsnrm,rsden,orsnrm)
		endif
c
c		Return
c
		goto 500
c
	endif
c
c ==========================================================================================
c 	Begin mg iteration
c ==========================================================================================
c
c	Remark: nxf,nyf,nzf change during loop
c
	iters = 0
100	continue
c
c		fine level initialization
c		==========================
c
		level = 1
		lev = (ilev-1) + level
c
c		nu1 pre-smoothing on fine grid
c		==============================
c
		iresid   = 1
		iadjoint = 0
		iters_s  = 0
		errtol_s = 0.d0
		nuuu     = ivariable_v(nu1,lev)
c
		call nonlin_smoothing(nxf,nyf,nzf,
     1		ipc(iz(5,lev)),rpc(iz(6,lev)),
     2		ac(iz(7,lev)),cc(iz(1,lev)),fc(iz(1,lev)),
     3		x(iz(1,lev)),w2,w3,w1,
     4		nuuu,iters_s,errtol_s,omega,iresid,iadjoint,
     5		mgsmoo)
c
		call vectcopy(nxf,nyf,nzf,w1,w0(iz(1,lev)))
c
c ==========================================================================================
c 	Begin cycling down to coarse grid
c ==========================================================================================
c
		do 200 level = 2,nlev
c
			lev = (ilev-1)+level
c
c			Find new grid size
c			===================
c
			call make_coarse(1,nxf,nyf,nzf,nxc,nyc,nzc)
c
c			Restrict residual on coarser grid
c			=====================================
c
			call restrict(nxf,nyf,nzf,nxc,nyc,nzc,
     1				w1,w0(iz(1,lev)))
c
			call extract(nxf,nyf,nzf,nxc,nyc,nzc,
     2				x(iz(1,lev-1)),w4(iz(1,lev)))
c
c			New grid size
c
			nxf = nxc
			nyf = nyc
			nzf = nzc
c
			call nmatvec_band(nxf,nyf,nzf,
     1			ipc(iz(5,lev)),rpc(iz(6,lev)),
     2			ac(iz(7,lev)),cc(iz(1,lev)),
     3			w4(iz(1,lev)),fc(iz(1,lev)),w3)
c
			call vectsmult(nxf,nyf,nzf,(-1.d0),
     1				w0(iz(1,lev)),fc(iz(1,lev)))
c
c			If not yet on coarsest level ....
c			=====================================
c
			if(level.ne.nlev) then
c
c				nu1 pre-smoothing, with computation of residual
c				===============================================
c
				call vectcopy(nxf,nyf,nzf,
     1                          w4(iz(1,lev)),x(iz(1,lev)))
c
				iresid   = 1
				iadjoint = 0
				iters_s  = 0
				errtol_s = 0.d0
				nuuu     = ivariable_v(nu1,lev)
c
				call nonlin_smoothing(nxf,nyf,nzf,
     1				ipc(iz(5,lev)),rpc(iz(6,lev)),
     2				ac(iz(7,lev)),cc(iz(1,lev)),
     3				fc(iz(1,lev)),x(iz(1,lev)),w2,w3,w1,
     4				nuuu,iters_s,errtol_s,omega,iresid,
     5				iadjoint,mgsmoo)
c
			endif
c
200		continue
c
c ==========================================================================================
c 	Solve on coarse grid
c ==========================================================================================
c
		level = nlev
		lev = (ilev-1)+level
c
c
c		Use iterative method
c		=====================
c
		iresid = 0
		iadjoint = 0
		itmax_s = 100
		iters_s = 0
		errtol_s = epsilon
		mgsmoo_s = mgsmoo
c
		call set_all_zeros(nxf,nyf,nzf,x(iz(1,lev)))
		call nonlin_smoothing(nxf,nyf,nzf,
     1		ipc(iz(5,lev)),rpc(iz(6,lev)),
     2		ac(iz(7,lev)),cc(iz(1,lev)),fc(iz(1,lev)),
     3		x(iz(1,lev)),w1,w2,w3,
     4		itmax_s,iters_s,errtol_s,omega,iresid,iadjoint,
     5		mgsmoo_s)
c
c		if(iters_s .ge.itmax_s) then
c		write(6,1) iters_s
c		endif
c
c ==========================================================================================
c 	Cycle back to fine grid
c ==========================================================================================
c
		do 300 level = nlev-1,1,-1
c
			lev = (ilev-1)+level
c
c			Find new grid size
c			=====================
c
			call make_fine(1,nxf,nyf,nzf,nxc,nyc,nzc)
c
c			Form difference of new approx. at the coarse grid
c			=================================================
c
			call vectsmult(nxf,nyf,nzf,(-1.d0),
     1			w4(iz(1,lev+1)),x(iz(1,lev+1)))
c
c			Call linesearch on the coarser grid
c			to compute damping parameter
c			===================================
c
			call linesearch(nxf,nyf,nzf,xdamp,
     1			ipc(iz(5,lev+1)),rpc(iz(6,lev+1)),
     2			ac(iz(7,lev+1)),cc(iz(1,lev+1)),
     3			fc(iz(1,lev+1)),x(iz(1,lev+1)),
     4			w4(iz(1,lev+1)),w0(iz(1,lev+1)),
     5			w1,w2,w3)
c
c			Interpolate to next finer grid
c			================================
c
			call interpolate(nxf,nyf,nzf,nxc,nyc,nzc,
     1			x(iz(1,lev+1)),w1)
c
c			New grid size
c			=============
c
			nxf = nxc
			nyf = nyc
			nzf = nzc
c
c			coarse grid correction
c			======================
c
c			xdamp = -0.8d0
			call vectsmult(nxf,nyf,nzf,xdamp,w1,
     1			x(iz(1,lev)))
c
c			Safeguard: if "correction" leads to overflow, ignore
c			====================================================
c
			call residual(nxf,nyf,nzf,ipc(iz(5,lev)),
     1			rpc(iz(6,lev)),ac(iz(7,lev)),cc(iz(1,lev)),
     2			fc(iz(1,lev)),x(iz(1,lev)),w3,w0)
			if(xnorm1(nxf,nyf,nzf,w3).gt.1.d10) then
				xdamp = -xdamp
				call vectsmult(nxf,nyf,nzf,xdamp,w1,
     1				x(iz(1,lev)))
			endif
c
c			nu2 post-smoothing for correction (no residual)
c			==============================================
c
			iresid   = 0
			iadjoint = 1
			iters_s  = 0
			errtol_s = 0.d0
			nuuu     = ivariable_v(nu2,lev)
c
			call nonlin_smoothing(nxf,nyf,nzf,
     1			ipc(iz(5,lev)),rpc(iz(6,lev)),
     2			ac(iz(7,lev)),cc(iz(1,lev)),
     3			fc(iz(1,lev)),x(iz(1,lev)),w1,w2,w3,
     4			nuuu,iters_s,errtol_s,omega,iresid,
     5			iadjoint,mgsmoo)
c
300		continue
c
c ==========================================================================================
c 	Iteration complete: do some i/o
c ==========================================================================================
c
		iters = iters + 1
c
		if(iok.ne.0) then
c
			orsnrm = rsnrm
			if(istop.eq.0) then
				call residual(nxf,nyf,nzf,
     1				ipc(iz(5,lev)),rpc(iz(6,lev)),
     2				ac(iz(7,lev)),cc(iz(1,lev)),
     3				fc(iz(1,lev)),x(iz(1,lev)),w1,w2)
				rsnrm = xnorm1(nxf,nyf,nzf,w1)
			elseif(istop.eq.1) then
				call residual(nxf,nyf,nzf,
     1				ipc(iz(5,lev)),rpc(iz(6,lev)),
     2				ac(iz(7,lev)),cc(iz(1,lev)),
     3				fc(iz(1,lev)),x(iz(1,lev)),w1,w2)
				rsnrm = xnorm1(nxf,nyf,nzf,w1)
			else
				write(6,*) 'Bad istop in lin_mg'
				stop
			endif
c
			if(iok.eq.1) call print_residual(iok,iters,
     1					rsnrm,rsden,orsnrm)
c
			if(rsnrm/rsden .le. errtol) goto 500
c
		endif
c
		if(iters.ge.itmax) goto 400
c
	goto 100
c
400	continue
	ierror = 1
c
500	continue
c
	return
	end
