c ==========================================================================================
c ==========================================================================================
c 	GridSize.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine computes the sizes of the real and integer work arrays for
c	the multigrid solver
c	
c	Input:
c
c               equation:       PDE equation solved:
c                               0: Laplace
c                               1: Linearized PBE
c                               2: Nonlinear PBE
c                               3: Size Modified PBE
c                               4: DIpolar Poisson Boltzmann Langevin
c
c		method  :	Method used to solve the PDE:
c				0= CGMG
c				1= Inexact Newton
c				2= FAS
c				3= Direct method
c
c		ijac	:	Stencil type for Jacobian (Inexact Newton)
c				0= Stencil of size 7
c				1= Stencil of size 27
c
c		mgcoar	:	coarsening technique
c				0= standard discretization
c				1= averaged coefficient + standard discretization
c				2=algebraic Galerkin coarsening
c
c		mgdisc	:	discretization technique
c				0=box method
c				1=fem method
c
c		mgsolv	:	coarse grid solver:
c				0 = conjugate gradient
c				1 = symmetric banded Linpack solver
c
c		nx,ny,nz:	grid dimensions
c
c		nlevel:		number of multigrid levels
c
c	Output:
c		ireal_tot	minimum size required for work array rwork
c		iint_tot	minimum size required for work array iwork
c
c
c	The work arrays rwork and iwork are chopped into smaller pieces according to:
c
c	real*8	ac(STORE)		: operators (i.e. A matrix) on all levels
c	real*8	helm(narr),fc(narr)	: Helmholtz terms and rhs, all levels
c	real*8	rpc(100*(nlevel+1))	: real info; all levels
c	real*8	w1(nf),w2(nf),...	: Storage arrays (number depends on method)
c
c	integer ipc(100*(nlevel+1))	: integer info; all levels
c	integer	iz(50,nlevel+1)		: pointers to delimit rwork and iwork for each level
c
c	STORE depends on the discretization and coarsening of grid solver:
c
c	STORE = 4*nf + 4*narrc + NBAND*nc 	(mgdisc=box, mgcoar=stan/harm)
c	STORE = 4*nf +14*narrc + NBAND*nc 	(mgdisc=box, mgcoar=gal)
c	STORE =14*nf +14*narrc + NBAND*nc 	(mgdisc=fem, mgcoar=stan/harm/gal)
c
c	and NBAND =
c		0			  	(mgsolv=iterative)
c		1+(nxc-2)*(nyc-2)	  	(mgsolv = 7-pt banded Linpack)
c		1+(nxc-2)*(nyc-2)+(nxc-2)+1 	(mgsolv = 27-pt banded Linpack)
c
	subroutine GridSize(equation,method,ijac,mgcoar,mgdisc,mgsolv,
     1          nx,ny,nz,nlevel,ireal_tot,iint_tot)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables
c ==========================================================================================
c
	integer	mgcoar,mgdisc,mgsolv,method,ijac,equation
	integer	nx,ny,nz,nlevel
c
c	parameters
c ==========================================================================================
c
	integer	num_nf,num_narr,num_narrc
c
c	Output variables
c ==========================================================================================
c
	integer iint_tot
c
        integer*8       ireal_tot
c
c	Miscellaneous:
c ==========================================================================================
c
	integer	i
	integer	nc_band,num_band,n_band,nxf,nyf,nzf,level
	integer	num_nf_oper,num_narrc_oper,n_vect
	integer	nxc,nyc,nzc,nf,nc,narr,narrc
	integer	n_rpc,n_ipc,n_iz
c
c ==========================================================================================
c 	Get array sizes      
c ==========================================================================================
c
	nf	= nx * ny * nz
	narr	= nf
	nxf	= nx
	nyf	= ny
	nzf	= nz
	nxc	= nx
	nyc	= ny
	nzc	= nz
c
c	Loop over all levels, where level 1 is the finest
c ==========================================================================================
c
	do 100 i = 2,nlevel
c
c		find new grid size
c
		call make_coarse(1,nxf,nyf,nzf,nxc,nyc,nzc)
c
c		New grid size
c
		nxf = nxc
		nyf = nyc
		nzf = nzc
c
c		Add unknowns on this level to total
c
		narr = narr + nxf*nyf*nzf
c
100	continue
c
	nc = nxc * nyc *nzc
	narrc = narr - nf
c
c	Flag, based on grid type: box (i.e. structured mesh), or fem (i.e. unstructured) mesh
c ==========================================================================================
c
	if(mgdisc.eq.0) then
		if(method.eq.1.and.ijac.eq.1) then
			num_nf_oper = 14
		else
			num_nf_oper = 4
		endif
	elseif(mgdisc.eq.1) then
		num_nf_oper = 14
	else
		write(6,*) 
     1		'Problem with solver: invalid discretization choice!'
	endif
c
c	Flag, based on coarsening technique: Galerkin, or standard?
c ==========================================================================================
c
	if (( (mgcoar.eq.0).or.(mgcoar.eq.1)).and.(mgdisc.eq.0)) then
		num_narrc_oper = 4
	elseif(mgcoar.eq.2) then
		num_narrc_oper = 14
	else
		write(6,*) 
     1		'Problem with solver: invalid coarsening choice!'
	endif
        if(method.eq.1) num_narrc_oper = 14
c
c	Storage for system solver on coarser grid
c ==========================================================================================
c
	if(mgsolv.eq.0) then
		n_band = 0
	elseif (mgsolv.eq.1) then
		if (( (mgcoar.eq.0).or.(mgcoar.eq.1)).and.
     1		(mgdisc.eq.0)) then
			num_band = 1+(nxc-2)*(nyc-2)
		else
			num_band = 1+(nxc-2)*(nyc-2)+(nxc-2)+1
		endif
		nc_band = (nxc-2)*(nyc-2)*(nzc-2)
		n_band = nc_band *num_band
	else
		write(6,*) 
     1		'Problem with solver: invalid choice for system solver!'
	endif
c
c	Number of auxiliary arrays
c ==========================================================================================
c
        num_narr = 2
c
	if(method.le.1) then
		num_nf = 2
        elseif(method.eq.2) then
                num_nf = 1
                if(equation.eq.4) then
                        num_nf_oper = 0
                        num_narrc_oper = 0
                endif
        elseif(method.eq.3) then
                num_nf = 0
                if(equation.eq.4) then
                        num_nf_oper = 0
                        num_narrc_oper = 0
                endif
        endif
c
c       Additional vector for DPBL equation
c ==========================================================================================
c
        if(equation.le.3) then
                n_vect = 0
        else
                if(method.eq.1) then
                        n_vect = nx*ny + 10*ijac*nx*ny
                else
                        n_vect = nx*ny
                endif
        endif
c
c	Size of work and int arrays:
c ==========================================================================================
c
	n_rpc = 100*(nlevel+1)
c
c	Resulting total (real storage) 
c
	ireal_tot = num_narr*narr
c
        ireal_tot = ireal_tot
     1			+(num_nf + num_nf_oper)*nf
     2			+ num_narrc_oper*narrc
     4			+ n_band + n_rpc + n_vect
c
c	Resulting total (integer storage)
c
	n_iz = 50*(nlevel+1)
	n_ipc = 100*(nlevel+1)
c
	iint_tot = n_iz + n_ipc
c
	return
	end
