c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Langevin7.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This file contains a library of routines that are
c			needed for computing the inexact Jacobian the Generalized PBL equation
c
c ==========================================================================================
c ==========================================================================================
c	BuildJ_langevin7.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the subroutine that computes the 
c			Jacobian of the GPBL equation
c
c	This version computes an approximation of the Jacobian: it truncates the stencil
c	of each points to its 6 directly connected neighbour (box mesh)
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		hx,hy,hz : coordinates on the grid
c		grid_info: Information on the grid that specifies the position
c			   of the molecules as well as the regions in which
c			   solvent and/or ions are present
c		fc	 : source term (fixed charges)
c		cc	 : Helmholtz term (mobile ions)
c		iok	 : flag for computing free energy (iok=1)
c               iflag    : flag (not used in this routine)
c		Space	 : storage space (size nx*ny)
c
c	Output:
c		jac	 : Jacobian (4 values per row)
c		resid	 : residual of the PDE for the given phi
c		free_energy: value of the functional free energy (if iok set to 1)
c	Output:
c
	subroutine buildJ_langevin7(phi,nx,ny,nz,jac,
     1			hx,hy,hz,grid_info,fc,cc,resid,
     2                  free_energy,iok,iflag,Space)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	integer	nx,ny,nz,iok,iflag
c
        integer*1 grid_info(*)
c
        real*8  free_energy
	real*8	phi(*)
	real*8	fc(*),resid(*),cc(*)
	real*8	jac(nx*ny*nz,*)
	real*8	hx(*),hy(*),hz(*)
c
	real*8	Space(*)
c
c ==========================================================================================
c	Call routine
c ==========================================================================================
c
	call compute_J_langevin7(phi,nx,ny,nz,
     1			hx,hy,hz,grid_info,
     2			jac(1,1),jac(1,2),jac(1,3),jac(1,4),
     5			fc,cc,resid,free_energy,iok,Space)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Compute_J_langevin7.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the position specific epsilon for
c			the GPBL equation
c
c	This is the standard version, that deals with any number of ions (up to 10)
c	and any valences for these ions
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: info on shifted grids
c
c	Output:
c		jac	: the Jacobian of the GPBL equation
c		resid	: residual vector
c
	subroutine compute_J_langevin7(phi,nx,ny,nz,
     1			x,y,z,grid_info,
     2			JoC,JoE,JoN,JuC,
     4			fc,cc,resid,free_energy,iok,Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
	integer	info1,iok,itype
c
        real*8  free_energy
        real*8  free_edens,free_charge,free_iondip
	real*8	eps_X,eps_Y,eps_Z
	real*8	coef_oE,coef_oN,coef_uC
	real*8	coef_oEh,coef_oNh,coef_uCh
	real*8	hxm1,hym1,hzm1
	real*8	hx,hy,hz
c
	real*8	phi(nx,ny,nz)
	real*8	fc(nx,ny,nz),resid(nx,ny,nz)
	real*8	cc(nx,ny,nz)
c
c	Jacobian for GPBL equation
c ==========================================================================================
c
	real*8	JoC(nx,ny,nz),JoE(nx,ny,nz),JoN(nx,ny,nz)
	real*8	JuC(nx,ny,nz)
c
	real*8	Enorm(nx,ny)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8	vol,p0_e,coef_l
	real*8	coef_Z0,phi0,ratio
	real*8  l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8  C_charge,C_prop,fac_edens
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion
c
	real*8	u_val,F1_u,sinhc_u,val,exp_mu
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,dm1,der_d
	real*8	der_d2,g,der_g
	real*8	arg,exp_max,exp_min
	real*8	l_dip,dvol
        real*8  eps1,eps2
c
	parameter       (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1                     	Avogadro,eps0,beta,C_charge,C_prop
        common /elec_const/     eps1,eps2
 
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
c	Make sure all arrays are initialized to 0
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,resid)
c
	call set_all_zeros(nx,ny,nz,JoC)
	call set_all_zeros(nx,ny,nz,JoE)
	call set_all_zeros(nx,ny,nz,JoN)
	call set_all_zeros(nx,ny,nz,JuC)
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
		k_ion(i) = kappa(i)/(1-phi0)
100	continue
c
	fac_edens = -(1d-10/(beta*electron*electron))*eps0/2
c
c ==========================================================================================
c	Compute eps for the X-shifted map
c ==========================================================================================
c
c	Compute electric field on x-shifted grid
c
	do 600 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_X(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 500 j = 2,ny-1
c
		hym1 = y(j) - y(j-1)
		hy = y(j+1)-y(j)
c
		do 400 i = 1,nx-1
c
			hx = x(i+1)-x(i)
c
                       	coef_oE  = (hym1+hy)*(hzm1+hz)/(4.d0*hx)
			coef_oEh = coef_oE*(phi(i,j,k)-phi(i+1,j,k))
c
			info1 = ibits(grid_info(i,j,k),2,2)
c
			if(info1.eq.0) then
c
				eps_X = eps1
c
				if(i.ne.1) then
					JoC(i,j,k) = JoC(i,j,k)
     1					+ coef_oE*eps_X
				endif
				if(i.ne.nx-1) then
					JoC(i+1,j,k) = JoC(i+1,j,k) 
     1					+ coef_oE*eps_X
				endif
				if(i.ne.1.and.i.ne.nx-1) then
					JoE(i,j,k) = JoE(i,j,k) 
     1					+ coef_oE*eps_X
				endif
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+phi(i+1,j,k))/2
c
				d = exp_mu
				der_d = 0
c
				if(info1.ne.1) then
c
				    do 200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val -u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_X = eps2 + val*dm1
c
				if(i.ne.1) then
					JoC(i,j,k) = JoC(i,j,k)
     1					+ coef_oE*eps_X
				endif
				if(i.ne.nx-1) then
					JoC(i+1,j,k) = JoC(i+1,j,k) 
     1					+ coef_oE*eps_X
				endif
				if(i.ne.1.and.i.ne.nx-1) then
					JoE(i,j,k) = JoE(i,j,k) 
     1					+ coef_oE*eps_X
				endif
c
			endif
c
			if(i.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oEh*eps_X
			endif
			if(i.ne.nx-1) then
				resid(i+1,j,k)=resid(i+1,j,k)
     1				-coef_oEh*eps_X
			endif
c
400		continue
500	   continue
c
600	continue
c
c
c ==========================================================================================
c	Define eps for the Y-shifted map
c ==========================================================================================
c
c	Compute electric field on Y-shifted grid
c
	do 1100 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Y(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1000 j = 1,ny-1
c
		hy = y(j+1)-y(j)
c
		do 900 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_oN  = (hxm1+hx)*(hzm1+hz)/(4.d0*hy)
			coef_oNh = coef_oN*(phi(i,j,k)-phi(i,j+1,k))
c
			info1 = ibits(grid_info(i,j,k),4,2)
c
			if(info1.eq.0) then
c
				eps_Y = eps1
c
				if(j.ne.1) then
					JoC(i,j,k) = JoC(i,j,k)
     1					+ coef_oN*eps_Y
				endif
				if(j.ne.ny-1) then
					JoC(i,j+1,k) = JoC(i,j+1,k) 
     1					+ coef_oN*eps_Y
				endif
				if(j.ne.1.and.j.ne.ny-1) then
					JoN(i,j,k) = JoN(i,j,k) 
     1					+ coef_oN*eps_Y
				endif
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c 
				phi_val = (phi(i,j,k)+
     1				phi(i,j+1,k))/2
c
				d = exp_mu
				der_d = 0
c
				if(info1.ne.1) then
c
				    do 700 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
700				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
				eps_Y = eps2 + val*dm1
c
				if(j.ne.1) then
					JoC(i,j,k) = JoC(i,j,k)
     1					+ coef_oN*eps_Y
				endif
				if(j.ne.ny-1) then
					JoC(i,j+1,k) = JoC(i,j+1,k) 
     1					+ coef_oN*eps_Y
				endif
				if(j.ne.1.and.j.ne.ny-1) then
					JoN(i,j,k) = JoN(i,j,k) 
     1					+ coef_oN*eps_Y
				endif
c
			endif
c
			if(j.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oNh*eps_Y
			endif
			if(j.ne.ny-1) then
				resid(i,j+1,k)=resid(i,j+1,k)
     1				-coef_oNh*eps_Y
			endif
c
900		continue
1000	   continue
c
1100	continue
c
c ==========================================================================================
c	Define eps for the Z-shifted map
c ==========================================================================================
c
c	Compute electric field on Z-shifted grid
c
	do 1700 k = 1,nz-1
c
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Z(phi(1,1,k),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1600 j = 2,ny-1
c
		hy = y(j+1)-y(j)
		hym1 = y(j)-y(j-1)
c
		do 1500 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_uC  = (hxm1+hx)*(hym1+hy)/(4.d0*hz)
			coef_uCh = coef_uC*(phi(i,j,k)-phi(i,j,k+1))
c
			info1 = ibits(grid_info(i,j,k),6,2)
c
			if(info1.eq.0) then
c
				eps_Z = eps1
c
				if(k.ne.1) then
					JoC(i,j,k) = JoC(i,j,k)
     1					+ coef_uC*eps_Z
				endif
				if(k.ne.nz-1) then
					JoC(i,j,k+1) = JoC(i,j,k+1) 
     1					+ coef_uC*eps_Z
				endif
				if(k.ne.1.and.k.ne.nz-1) then
					JuC(i,j,k) = JuC(i,j,k) 
     1					+ coef_uC*eps_Z
				endif
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+
     1				phi(i,j,k+1))/2
c
				d = exp_mu
				der_d = 0
c
				if(info1.ne.1) then
c
				    do 1200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d + l_ion(i_ion)*arg
					der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
1200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_Z = eps2 + val*dm1
c
				if(k.ne.1) then
					JoC(i,j,k) = JoC(i,j,k)
     1					+ coef_uC*eps_Z
				endif
				if(k.ne.nz-1) then
					JoC(i,j,k+1) = JoC(i,j,k+1) 
     1					+ coef_uC*eps_Z
				endif
				if(k.ne.1.and.k.ne.nz-1) then
					JuC(i,j,k) = JuC(i,j,k) 
     1					+ coef_uC*eps_Z
				endif
			endif
c
			if(k.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_uCh*eps_Z
			endif
			if(k.ne.nz-1) then
				resid(i,j,k+1)=resid(i,j,k+1)
     1				-coef_uCh*eps_Z
			endif
c
1500		continue
1600	   continue
c
1700	continue
c
        if(iok.eq.1) then
c
	        if(n_ion.eq.0) goto 2200
c
	        do 2100 k = 2,nz-1
c
                    hz = 1.d0/(z(k+1)-z(k-1))
	            call plane_efield(phi(1,1,k),phi(1,1,k-1),
     1			phi(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	            do 2000 j = 2,ny-1
c
	   	        hy = 1.d0/(y(j+1)-y(j-1))
c
		        do 1900 i = 2,nx-1
c
	   		    hx = 1.d0/(x(i+1)-x(i-1))
c
			    if(cc(i,j,k).ne.0) then
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
c
				phi_val = phi(i,j,k)
c
				d = exp_mu
				der_d = 0
				der_g = 0
				g = 0
				do 1800 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
					der_d = der_d-ion_z(i_ion)
     1					*l_ion(i_ion)*arg
					der_g = der_g -ion_z(i_ion)
     1					*k_ion(i_ion)*arg
1800				continue
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*g
     1				*dm1
c
				JoC(i,j,k)=JoC(i,j,k)+cc(i,j,k)*
     1				(der_g*d - g*der_d)*dm1*dm1
c
			endif
c
1900                continue
2000	        continue
c
2100	    continue
c
2200        continue
c
        else
c
                free_edens = 0
                free_charge = 0
                free_iondip = 0
c
	        do 2600 k = 2,nz-1
c
                    hz = 1.d0/(z(k+1)-z(k-1))
	            call plane_efield(phi(1,1,k),phi(1,1,k-1),
     1			phi(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	            do 2500 j = 2,ny-1
c
	   	        hy = 1.d0/(y(j+1)-y(j-1))
c
		        do 2400 i = 2,nx-1
c
	   		    hx = 1.d0/(x(i+1)-x(i-1))
c
                            dvol = hx*hy*hz/8
			    phi_val = phi(i,j,k)
c
                            free_edens = free_edens + Enorm(i,j)
     1                          *Enorm(i,j)*dvol
c
                            free_charge = free_charge+phi_val*fc(i,j,k)
c
			    info1 = ibits(grid_info(i,j,k),0,2)
c
			    if(info1.ne.0) then
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
c
				d = exp_mu
				der_d = 0
				g = 0
				der_g = 0
                                if(info1.ne.1) then
				    do 2300 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
					der_d = der_d-ion_z(i_ion)
     1					*l_ion(i_ion)*arg
					der_g = der_g -ion_z(i_ion)
     1					*k_ion(i_ion)*arg
2300				    continue
                                endif
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*g
     1				*dm1
c
				JoC(i,j,k)=JoC(i,j,k)+cc(i,j,k)*
     1				(der_g*d - g*der_d)*dm1*dm1
c
                                free_iondip = free_iondip -log(d)*dvol
c
			endif
c
2400                continue
2500	        continue
c
2600	    continue
c
            free_edens = -fac_edens*free_edens
            free_charge = free_charge/C_charge
            free_iondip = free_iondip/vol
c
            free_energy = free_edens + free_charge + free_iondip
c
        endif
c
c
	do 2900 k = 1,nz
		do 2800 j = 1,ny
			do 2700 i = 1,nx
				resid(i,j,k) = -resid(i,j,k)+fc(i,j,k)
2700			continue
2800		continue
2900	continue
c
c        stop
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Compute_F1.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the function F1(u) = Langevin(u)*sinhc(u)
c
c	Note:
c			F1(u) = ( cosh(u) - sinh(u)/u ) /u^2
c
c ==========================================================================================
c	Numerical issues:
c ==========================================================================================
c
c	Overflow problems:
c	==================
c
c	When u is large, this can lead to serious overflows (especially at the beginning
c	of the refinement, as u still fluctuates a lot)
c	For that reason, we compute:
c
c	Fp1(u) = exp(-u)*F1(u) = ( (1+exp(-2u))/2 + (1-exp(-2u))/(2u) ) / u^2
c
c	Underflow problems:
c	==================
c
c	Note that:
c			lim  F1(u) = 1/3
c			u->0
c
c
	subroutine compute_F1(u_val,F1_u,sinhc_u,exp_mu)
c
	real*8	u_val,u2,F1_u,sinhc_u
	real*8	tol
	real*8	sinh_u,cosh_u,exp_2mu,exp_mu
c
	parameter	(tol=1.d-04)
c
	if(u_val.le.tol) then
		F1_u = 1.d0/3.d0
		sinhc_u = 1.d0
		exp_mu = 1.d0
        else
		exp_mu = exp(-u_val)
		exp_2mu = exp_mu*exp_mu/2
		sinh_u = 0.5d0 - exp_2mu
		cosh_u = 0.5d0 + exp_2mu
		sinhc_u = sinh_u/u_val
		u2 = u_val*u_val
		F1_u = (cosh_u-sinhc_u)/u2
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	residue_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the routine that
c			computes the residual of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: infor on regular grid and shifted grids
c		fc	 : source term (fixed charges)
c		cc	 : Helmholtz term
c		Space  	 : storage array
c
c	Output:
c		resid	 : residual vector
c
c	This routine considers two cases; z:z electrolytes, and more general
c	systems
c
c	Output:
c		resid	: residual vector
c
	subroutine residue_gpbl(phi,nx,ny,nz,x,y,z,grid_info,
     2			Space,fc,cc,resid)
c
	integer	nx,ny,nz
c
        integer*1 grid_info(*)
c
	real*8	phi(*),Space(*),fc(*),cc(*),resid(*)
        real*8  x(*),y(*),z(*)
c
	call residue_langevin(phi,nx,ny,nz,x,y,z,grid_info,
     1		Space,fc,cc,resid)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	residue_langevin.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the routine that
c			computes the residual of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: infor on regular grid and shifted grids
c		fc	 : source term (fixed charges)
c		cc	 : Helmholtz term
c		Enorm  	 : storage array
c
c	Output:
c		resid	 : residual vector
c
c	This is the generic routine, for any number of ions (up to 10), with
c	any valence
c
	subroutine residue_langevin(phi,nx,ny,nz,x,y,z,grid_info,
     2			Enorm,fc,cc,resid)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer info1
c
	real*8	eps_X,eps_Y,eps_Z
	real*8	coef_oE,coef_oN,coef_uC
	real*8	coef_oEh,coef_oNh,coef_uCh
	real*8	hxm1,hym1,hzm1
	real*8	hx,hy,hz
	real*8	free_edens,free_charge,free_iondip,free_energy,dvol
c
	real*8	phi(nx,ny,nz)
	real*8	fc(nx,ny,nz),resid(nx,ny,nz)
	real*8	cc(nx,ny,nz)
c
	real*8	Enorm(nx,ny)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8	vol,p0_e,coef_l
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion
c
	real*8	u_val,F1_u,sinhc_u,val,exp_mu
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,g,dm1
	real*8	arg,exp_max,exp_min
	real*8	l_dip
        real*8  eps1,eps2
c
	parameter       (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
        common /elec_const/     eps1,eps2
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
c	Make sure all arrays are initialized to 0
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,resid)
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
		k_ion(i) = kappa(i)/(1-phi0)
100	continue
c
c ==========================================================================================
c	Compute eps for the X-shifted map
c ==========================================================================================
c
c	Compute electric field on x-shifted grid
c
	do 600 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_X(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 500 j = 2,ny-1
c
		hym1 = y(j) - y(j-1)
		hy = y(j+1)-y(j)
c
		do 400 i = 1,nx-1
c
			hx = x(i+1)-x(i)
c
                       	coef_oE  = (hym1+hy)*(hzm1+hz)/(4.d0*hx)
			coef_oEh = coef_oE*(phi(i,j,k)-phi(i+1,j,k))
c
			info1 = ibits(grid_info(i,j,k),2,2)
c
			if(info1.eq.0) then
c
                                eps_X = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+phi(i+1,j,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_X = eps2 + val*dm1
c
			endif
c
			if(i.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oEh*eps_X
			endif
			if(i.ne.nx-1) then
				resid(i+1,j,k)=resid(i+1,j,k)
     1				-coef_oEh*eps_X
			endif
c
400		continue
500	   continue
c
600	continue
c
c ==========================================================================================
c	Define eps for the Y-shifted map
c ==========================================================================================
c
c	Compute electric field on Y-shifted grid
c
	do 1100 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Y(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1000 j = 1,ny-1
c
		hy = y(j+1)-y(j)
c
		do 900 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_oN  = (hxm1+hx)*(hzm1+hz)/(4.d0*hy)
			coef_oNh = coef_oN*(phi(i,j,k)-phi(i,j+1,k))
c
			info1 = ibits(grid_info(i,j,k),4,2)
c
			if(info1.eq.0) then
c
				eps_Y = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c 
				phi_val = (phi(i,j,k)+
     1				phi(i,j+1,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 700 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
700				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
				eps_Y = eps2 + val*dm1
c
			endif
c
			if(j.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oNh*eps_Y
			endif
			if(j.ne.ny-1) then
				resid(i,j+1,k)=resid(i,j+1,k)
     1				-coef_oNh*eps_Y
			endif
c
900		continue
1000	   continue
c
1100	continue
c
c ==========================================================================================
c	Define eps for the Z-shifted map
c ==========================================================================================
c
c	Compute electric field on Z-shifted grid
c
	do 1700 k = 1,nz-1
c
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Z(phi(1,1,k),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1600 j = 2,ny-1
c
		hy = y(j+1)-y(j)
		hym1 = y(j)-y(j-1)
c
		do 1500 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_uC  = (hxm1+hx)*(hym1+hy)/(4.d0*hz)
			coef_uCh = coef_uC*(phi(i,j,k)-phi(i,j,k+1))
c
			info1 = ibits(grid_info(i,j,k),6,2)
c
			if(info1.eq.0) then
c
				eps_Z = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+
     1				phi(i,j,k+1))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 1200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d + l_ion(i_ion)*arg
1200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_Z = eps2 + val*dm1
c
			endif
c
			if(k.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_uCh*eps_Z
			endif
			if(k.ne.nz-1) then
				resid(i,j,k+1)=resid(i,j,k+1)
     1				-coef_uCh*eps_Z
			endif
c
1500		continue
1600	   continue
c
1700	continue
c
	if(n_ion.eq.0) goto 2200
c
	do 2100 k = 2,nz-1
c
	   call plane_efield(phi(1,1,k),phi(1,1,k-1),
     1		phi(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	   do 2000 j = 2,ny-1
c
		do 1900 i = 2,nx-1
c
			if(cc(i,j,k).ne.0) then
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
c
				phi_val = phi(i,j,k)
c
				d = exp_mu
                                g = 0
				do 1800 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
1800				continue
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*
     1						g*dm1
c
			endif
c
1900		continue
2000	   continue
c
2100	continue
c
2200	continue
c
	do 2500 k = 1,nz
		do 2400 j = 1,ny
			do 2300 i = 1,nx
				resid(i,j,k) = -resid(i,j,k)+fc(i,j,k)
2300			continue
2400		continue
2500	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	free_energy_langevin_zero.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the residual of the GPBL
c			equation for a zero field
c
c	Input:
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: infor on regular grid and shifted grids
c
c	Output:
c		resid	: residual vector
c
	subroutine free_energy_langevin_zero(nx,ny,nz,
     1			x,y,z,grid_info,cc,free_energy)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
c
	real*8	hx,hy,hz
c
	real*8	cc(nx,ny,nz)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
        real*8  free_edens,free_charge,free_iondip,free_energy
	real*8	vol,p0_e,coef_l,dvol
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion,info1
c
	real*8	sinhc_u,arg
	real*8	d,g,dm1
	real*8	l_dip
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
100	continue
c
        free_edens = 0
        free_charge = 0
        free_iondip = 0
c
	do 500 k = 2,nz-1
c
           hz = z(k+1)-z(k-1)
c
	   do 400 j = 2,ny-1
c
                hy = y(j+1)-y(j-1)
		do 300 i = 2,nx-1
c
                        hx = x(i+1)-x(i-1)
c
                        dvol = hx*hy*hz/8
c
                        info1 = ibits(grid_info(i,j,k),0,2)
c
			if(info1.ne.0) then
c
                                sinhc_u = 1
c
				d = 1
                                if(cc(i,j,k).ne.0) then
				   do 200 i_ion = 1,n_ion
					arg = 1.d0
					d=d+l_ion(i_ion)*arg
200				   continue
                                endif
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
                                free_iondip = free_iondip -log(d)*dvol
c
			endif
c
300		continue
400	   continue
c
500	continue
c
        free_iondip = free_iondip/vol
c
        free_energy = free_edens + free_charge + free_iondip
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield(phik,phikm1,phikp1,nx,ny,hx,hy,hz,
     1                  iplane,Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dz,val
c
	k = iplane
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	Efield_x(i,j,k) = (u(i+1/2,j,k) - u(i-1/2,j,k))/dx
c
c	where dx = (hx(j+1)-hx(j-1))/2
c
c	Using interpolation:
c
c	u(i+1/2,j,k) = (u(i+1,j,k)+u(i,j,k))/2
c	and
c	u(i-1/2,j,k) = (u(i-1,j,k)+u(i,j,k))/2
c
c	Then:
c
c	Efield_x(i,j,k) = (u(i+1,j,k) - u(i-1,j,k))/(2*dx)
c
	do 200 j = 2,ny-1
		do 100 i = 2,nx-1
			dx = 1.d0/(hx(i+1)-hx(i-1))
			val = (phik(i+1,j)-phik(i-1,j))*dx
			Enorm(i,j) = val*val
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	Efield_y(i,j,k) = (u(i,j+1/2,k) - u(i,j-1/2,k))/dy
c
c	where dy = (hy(j+1)-hy(j-1))/2
c
c	Using interpolation:
c
c	u(i,j+1/2,k) = (u(i,j+1,k)+u(i,j,k))/2
c	and
c	u(i,j-1/2,k) = (u(i,j-1,k)+u(i,j,k))/2
c
c	Then:
c
c	Efield_y(i,j,k) = (u(i,j+1,k) - u(i,j-1,k))/(2*dy)
c
	do 500 j = 2,ny-1
		dy = 1.d0/(hy(j+1)-hy(j-1))
		do 400 i = 2,nx-1
			val = (phik(i,j+1)-phik(i,j-1))*dy
			Enorm(i,j)=Enorm(i,j) + val*val
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	Efield_z(i,j,k) = (u(i,j,k+1/2) - u(i,j,k-1/2))/dz
c
c	where dz = (hz(k+1)-hz(k-1))/2
c
c	Using interpolation:
c
c	u(i,j,k+1/2) = (u(i,j,k+1)+u(i,j,k))/2
c	and
c	u(i,j,k-1/2) = (u(i,j,k-1)+u(i,j,k))/2
c
c	Then:
c
c	Efield_z(i,j,k) = (u(i,j,k+1) - u(i,j,k-1))/(2*dz)
c
	dz = 1.d0/(hz(k+1)-hz(k-1))
	do 900 j = 2,ny-1
		do 800 i = 2,nx-1
			val = (phikp1(i,j)-phikm1(i,j))*dz
			Enorm(i,j)=Enorm(i,j) + val*val
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 2,ny-1
		do 1000 i = 2,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield_X.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield_X(phik,phikm1,phikp1,nx,ny,hx,hy,
     1                  hz,iplane,Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dz,val
c
	k = iplane
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	We have shifted the grid in the X-direction: the Electric field in this direction
c	at index i is therefore simply phi(i+1,j,k) - phi(i,j,k), up to i = nx-1 
c
	do 200 j = 2,ny-1
		do 100 i = 1,nx-1
			dx = 1.d0/(hx(i+1)-hx(i))
			val = (phik(i+1,j)-phik(i,j))*dx
			Enorm(i,j) = val*val
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	The grid is now positioned at half interval along the x-axis. To compute the electric
c	field along y at position (i,j,k) (shifted grid), we need to compute:
c
c	j=2, ny-1:
c	Efield_y(i,j,k) = (u(i+1/2,j+1/2,k) - u(i+1/2,j-1/2,k))/dy
c
c	where dy = (hy(j+1)-hy(j-1))/2
c
c	j = 1:
c	Efield_y(i,j,k) = (u(i+1/2,1+1/2,k) - u(i+1/2,1,k))/dy1
c
c	where dy1 = (hy(2)-hy(1))/2
c
c	Note that we do not need this values for j = ny
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j+1/2,k) = (u(i,j,k) + u(i+1,j,k) + u(i,j+1,k) + u(i+1,j+1,k))/4
c	u(i+1/2,j-1/2,k) = (u(i,j,k) + u(i+1,j,k) + u(i,j-1,k) + u(i+1,j-1,k))/4
c	u(i+1/2,j,k)     = (u(i,j,k) + u(i+1,j,k))/2
c
c	Therefore:
c
c	For j = 2, ny-1
c	Efield_y(i,j,k) = (u(i,j+1,k)+u(i+1,j+1,k)-u(i,j-1,k)-u(i+1,j-1,k))/(4*dy)
c
c	For j = 1
c	Efield_y(i,j,k) = (u(i,j+1,k)+u(i+1,j+1,k)-u(i,j,k)-u(i+1,j,k))/(4*dy1)
c
	do 500 j = 2,ny-1
		dy = 1.d0/(2*(hy(j+1)-hy(j-1)))
		do 400 i = 1,nx-1
			val = (phik(i,j+1)+phik(i+1,j+1)-phik(i,j-1)
     1				-phik(i+1,j-1))*dy
			Enorm(i,j)=Enorm(i,j) + val*val
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	The grid is positioned at half interval along the x-axis. To compute the electric
c	field along z at position (i,j,k) (shifted grid), we need to compute:
c
c	For k = 2,nz-1
c	Efield_z(i,j,k) = (u(i+1/2,j,k+1/2) - u(i+1/2,j,k-1/2))/dz
c
c	where dz = (hz(k+1)-hz(k-1))/2
c
c	and for k = 1
c	Efield_z(i,j,k) = (u(i+1/2,j,1+1/2) - u(i+1/2,j,1))/dz1
c
c	where dz1 = (hz(2)-hz(1))/2
c
c	Note that we do not need this values for k = nz
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j,k+1/2) = (u(i,j,k) + u(i+1,j,k) + u(i,j,k+1) + u(i+1,j,k+1))/4
c	u(i+1/2,j,k-1/2) = (u(i,j,k) + u(i+1,j,k) + u(i,j,k-1) + u(i+1,j,k-1))/4
c	u(i+1/2,j,k)     = (u(i,j,k) + u(i+1,j,k))/2
c
c	Therefore:
c
c	For k=iplane = 2, nz-1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i+1,j,k+1)-u(i,j,k-1)-u(i+1,j,k-1))/(4*dz)
c
c	For k = 1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i+1,j,k+1)-u(i,j,k)-u(i+1,j,k))/(4*dz1)
c
	dz = 1.d0/(2.d0*(hz(k+1)-hz(k-1)))
	do 900 j = 2,ny-1
		do 800 i = 1,nx-1
			val = (phikp1(i,j)+phikp1(i+1,j)-
     1			phikm1(i,j)-phikm1(i+1,j))*dz
			Enorm(i,j)=Enorm(i,j) + val*val
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 2,ny-1
		do 1000 i = 1,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield_Y.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield_Y(phik,phikm1,phikp1,nx,ny,hx,hy,hz,
     1                  iplane,Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dz,val
c
	k = iplane
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	We have shifted the grid in the Y-direction: the Electric field in this direction
c	at index j is therefore simply phi(i,j+1,k) - phi(i,j,k), up to j = ny-1 
c
	do 200 j = 1,ny-1
		dy = 1.d0/(hy(j+1)-hy(j))
		do 100 i = 2,nx-1
			val = (phik(i,j+1)-phik(i,j))*dy
			Enorm(i,j) = val*val
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	The grid is now positioned at half interval along the y-axis. To compute the electric
c	field along x at position (i,j,k) (shifted grid), we need to compute:
c
c	For x = 2, nx-1:
c	Efield_x(i,j,k) = (u(i+1/2,j+1/2,k) - u(i-1/2,j+1/2,k))/dx
c
c	where dx = (hx(i+1)-hx(i-1))/2
c
c	For x = 1:
c	Efield_x(1,j,k) = (u(1+1/2,j+1/2,k) - u(1,j+1/2,k))/dx1
c
c	where dx1 = (hx(2)-hx(1))/2
c
c	Note that we do not need this values for i = nx
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j+1/2,k) = (u(i,j,k) + u(i+1,j,k) + u(i,j+1,k) + u(i+1,j+1,k))/4
c	u(i-1/2,j+1/2,k) = (u(i-1,j,k) + u(i,j+1,k) + u(i,j,k) + u(i-1,j+1,k))/4
c	u(i,j+1/2,k)     = (u(i,j,k) + u(i,j+1,k))/2
c
c	Therefore:
c
c	For i = 2, nx-1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j+1,k)-u(i-1,j,k)-u(i-1,j+1,k))/(4*dx)
c
c	For i = 1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j+1,k)-u(i,j,k)-u(i,j+1,k))/(4*dx1)
c
	do 500 j = 1,ny-1
		do 400 i = 2,nx-1
			dx = 1.d0/(2*(hx(i+1)-hx(i-1)))
			val = (phik(i+1,j)+phik(i+1,j+1)-phik(i-1,j)
     1				-phik(i-1,j+1))*dx
			Enorm(i,j)=Enorm(i,j) + val*val
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	The grid is positioned at half interval along the y-axis. To compute the electric
c	field along z at position (i,j,k) (shifted grid), we need to compute:
c
c	For k = 2,nz-1:
c	Efield_z(i,j,k) = (u(i,j+1/2,k+1/2) - u(i,j+1/2,k-1/2))/dz
c
c	with dz = (hz(k+1)-hz(k-1))/2
c
c	and for k = 1:
c	Efield_z(i,j,1) = (u(i,j+1/2,1+1/2) - u(i,j+1/2,1))/dz1
c
c	with dz1 = (hz(2)-hz(1))/2
c
c	Note that we do not need this values for k = nz
c
c	Using bilinear interpolation:
c
c	u(i,j+1/2,k+1/2) = (u(i,j,k) + u(i,j+1,k) + u(i,j,k+1) + u(i,j+1,k+1))/4
c	u(i,j+1/2,k-1/2) = (u(i,j,k) + u(i,j+1,k) + u(i,j,k-1) + u(i,j+1,k-1))/4
c	u(i,j+1/2,k)     = (u(i,j,k) + u(i,j+1,k))/2
c
c	Therefore:
c
c	For k=iplane = 2, nz-1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i,j+1,k+1)-u(i,j,k-1)-u(i,j+1,k-1))/(4*dz)
c
c	For k = 1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i,j+1,k+1)-u(i,j,k)-u(i,j+1,k))/(4*dz1)
c
	dz = 1.d0/(2*(hz(k+1)-hz(k-1)))
	do 900 j = 1,ny-1
		do 800 i = 2,nx-1
			val = (phikp1(i,j)+phikp1(i,j+1)-
     1			phikm1(i,j)-phikm1(i,j+1))*dz
			Enorm(i,j)=Enorm(i,j) + val*val
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 1,ny-1
		do 1000 i = 2,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield_Z.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield_Z(phik,phikp1,nx,ny,hx,hy,hz,
     1                  iplane,Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dz,val
c
	k = iplane
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	We have shifted the grid in the Z-direction: the Electric field in this direction
c	at index (i,j) is therefore simply phi(i,j,k+1) - phi(i,j,k), up to k = nz-1 
c
	dz = 1.d0/(hz(k+1)-hz(k))
	do 200 j = 2,ny-1
		do 100 i = 2,nx-1
			val = (phikp1(i,j)-phik(i,j))*dz
			Enorm(i,j) = val*val
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	The grid is now positioned at half interval along the z-axis. To compute the electric
c	field along x at position (i,j,k) (shifted grid), we need to compute:
c
c	x = 2,nx-1:
c	Efield_x(i,j,k) = (u(i+1/2,j,k+1/2) - u(i-1/2,j,k+1/2))/dx
c
c	where dx = (hx(i+1)-hx(i-1))/2
c
c	x = 1:
c	Efield_x(1,j,k) = (u(1+1/2,j,k+1/2) - u(1,j,k+1/2))/dx
c
c	where dx1 = (hx(2)-hx(1))/2
c
c	Note that we do not need this values for i = nx
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j,k+1/2) = (u(i,j,k) + u(i+1,j,k) + u(i,j,k+1) + u(i+1,j,k+1))/4
c	u(i-1/2,j,k+1/2) = (u(i-1,j,k) + u(i,j,k+1) + u(i,j,k) + u(i-1,j,k+1))/4
c	u(i,j,k+1/2)     = (u(i,j,k) + u(i,j,k+1))/2
c
c	Therefore:
c
c	For i = 2, nx-1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j,k+1)-u(i-1,j,k)-u(i-1,j,k+1))/(4*dx)
c
c	For i = 1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j,k+1)-u(i,j,k)-u(i,j,k+1))/(4*dx1)
c
	do 500 j = 2,ny-1
		do 400 i = 2,nx-1
			dx = 1.d0/(2*(hx(i+1)-hx(i-1)))
			val = (phik(i+1,j)+phikp1(i+1,j)-phik(i-1,j)
     1				-phikp1(i-1,j))*dx
			Enorm(i,j)=Enorm(i,j) + val*val
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	The grid is positioned at half interval along the y-axis. To compute the electric
c	field along z at position (i,j,k) (shifted grid), we need to compute:
c
c	For y = 2,ny-1:
c	Efield_y(i,j,k) = (u(i,j+1/2,k+1/2) - u(i,j-1/2,k+1/2))/dy
c
c	where dy = (hy(j+1)-hy(j-1))/2
c
c	For y = 1:
c	Efield_y(i,1,k) = (u(i,1+1/2,k+1/2) - u(i,1,k+1/2))/dy1
c
c	where dy1 = (hy(2)-hy(1))/2
c
c	Note that we do not need this values for k = nz
c
c	Using bilinear interpolation:
c
c	u(i,j+1/2,k+1/2) = (u(i,j,k) + u(i,j+1,k) + u(i,j,k+1) + u(i,j+1,k+1))/4
c	u(i,j-1/2,k+1/2) = (u(i,j,k) + u(i,j,k+1) + u(i,j-1,k) + u(i,j-1,k+1))/4
c	u(i,j,k+1/2)     = (u(i,j,k) + u(i,j,k+1))/2
c
c	Therefore:
c
c	For j= 2, ny-1
c	Efield_z(i,j,k) = (u(i,j+1,k)+u(i,j+1,k+1)-u(i,j-1,k)-u(i,j-1,k+1))/(4*dy)
c
c	For j = 1:
c	Efield_z(i,j,k) = (u(i,j+1,k+1)+u(i,j+1,k+1)-u(i,j,k)-u(i,j,k+1))/(4*dy1)
c
	do 900 j = 2,ny-1
		dy=1.d0/(2*(hy(j+1)-hy(j-1)))
		do 800 i = 2,nx-1
			val = (phik(i,j+1)+phikp1(i,j+1)-
     1			phik(i,j-1)-phikp1(i,j-1))*dy
			Enorm(i,j)=Enorm(i,j) + val*val
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 2,ny-1
		do 1000 i = 2,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
