c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c
c	LineSearch.f
c
c       (Performs a 1D linesearch, using Hackbusch/Reusken algorithm)
c
	subroutine linesearch(nx,ny,nz,alpha,ipc,rpc,ac,cc,fc,p,x,r,ap,
     1			zk,zkp1)
c
c	This routine searches for a real value alpha such that:
c
c	J(x + alpha p) < J(x)
c
c ==========================================================================================
c	Input/Output:
c ==========================================================================================
c
c	Input:
c
c		- nx,ny,nz	: mesh dimensions
c		- ipc,rpc	: level information
c		- ac,cc,fc	: level entries
c		- p		: the direction
c		- x		: the previous estimate of the solution
c		- r		: the previous non linear residual (r=r(x))
c
c	Temporary vectors:
c
c		- ap,zk,zkp1
c
c	Output:
c
c		- alpha
c
c ==========================================================================================
c	Declarations
c ==========================================================================================
c
	integer	nx,ny,nz,ipkey
	integer	niters,nitmax,ifail_tol
c
	integer	ipc(*)
c
	real*8	alpha
	real*8	ALP,pAp,pr,val,ALP_old
	real*8	zNN,zntol,DzNN
        real*8  dotvect,dotvect3
c
	real*8	rpc(*),ac(*),cc(*),fc(*)
	real*8	p(*),x(*),r(*),ap(*),zk(*),zkp1(*)
c
c ==========================================================================================
c	Setup: non-linear iteration tolerance and itmax
c ==========================================================================================
c
	nitmax = 20
	zntol  = 1.d-5
	ifail_tol = 0
	ipkey = ipc(10)
c
c ==========================================================================================
c	Firt guess for alpha using linear case
c ==========================================================================================
c
        call matvec_band(nx,ny,nz,ipc,rpc,ac,cc,p,ap)
        pr = dotvect(nx,ny,nz,p,r)
        pAp = dotvect(nx,ny,nz,p,ap)
        alpha = pr/pAp
c
	ALP = alpha
c
	call vectcopy(nx,ny,nz,x,r)
	call vectsmult(nx,ny,nz,ALP,p,r)
c
	call c_vec(cc,x,zk,nx,ny,nz,ipkey)
	call c_vec(cc,r,zkp1,nx,ny,nz,ipkey)
c
c	Evaluate residual of 1D system
c
	call vectcopy(nx,ny,nz,zkp1,r)
	call vectsmult(nx,ny,nz,(-1.d0),zk,r)
	val = dotvect(nx,ny,nz,p,r)
	zNN = ALP*pAp - pr + val
c
c ==========================================================================================
c	Finding alpha using 1D Newton method
c ==========================================================================================
c
	niters = 0
c
	if(dabs(zNN).lt.zntol) goto 300
	if(niters.gt.nitmax) goto 200
c
	ifail_tol = 0
	ALP_old = ALP
c
100	continue
c
		niters = niters + 1
c
c		Construct derivatives of NN
c
		call dc_vec(cc,r,zkp1,nx,ny,nz,ipkey)
		val = dotvect3(nx,ny,nz,zkp1,p)
		DzNN = pAp + val
c
c		Update solution
c
		ALP_old = ALP
		ALP = ALP - zNN / DzNN
c
c		Evaluate residual of 1D system
c
		call vectcopy(nx,ny,nz,x,r)
		call vectsmult(nx,ny,nz,ALP,p,r)
		call c_vec(cc,r,zkp1,nx,ny,nz,ipkey)
		call vectcopy(nx,ny,nz,zkp1,r)
		call vectsmult(nx,ny,nz,(-1.d0),zk,r)
		val = dotvect(nx,ny,nz,p,r)
		zNN = ALP*pAp - pr + val
c
c		Tolerance and itmax checks
c
		if(dabs(zNN).lt.zntol.and.
     1		(dabs((ALP-ALP_old)/ALP).lt.zntol)) goto 300
		if(niters.gt.nitmax) goto 200
c
	goto 100
c
c	Tolerance not reached
c
200	continue
c
	ifail_tol = ifail_tol + 1
c
c	Tolerance reached
c
300	continue
c
c ==========================================================================================
c	Newton search complete: update solution value
c ==========================================================================================
c
	alpha = max(ALP,-0.8d0)
c
c	print message
c
c	if(ifail_tol.gt.0) then
c		write(6,*) 'LineSearch: exceeded 1D Newton itmax: ',
c     1			nitmax
c	endif
c
c ==========================================================================================
c	Return and end
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c
c	dotvect3.f
c
c	This small function computes the sum of x(i,j,k)*y(i,j,k)^2, for all
c	position in the mesh
c
c ==========================================================================================
c ==========================================================================================
c
	function dotvect3(nx,ny,nz,x,y)
c
c ==========================================================================================
c	Declarations
c ==========================================================================================
c
	integer	i,j,k,nx,ny,nz
c
	real*8	dotvect3
	real*8	x(nx,ny,nz),y(nx,ny,nz),z(nx,ny,nz)
c
	dotvect3 = 0
c
	do 300 k = 2,nz-1
		do 200 j = 2,ny-1
			do 100 i = 2,nx-1
				dotvect3=dotvect3 + x(i,j,k)*y(i,j,k)
     1				*y(i,j,k)
100			continue
200		continue
300	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c
c	LineSearch_gpbl.f
c
c       (Performs a 1D linesearch, using Hackbusch/Reusken algorithm)
c
	subroutine linesearch_gpbl(nx,ny,nz,alpha,ipc,rpc,cc,fc,p,x,
     1          r,ap,grid_info,xpos,ypos,zpos,vect)
c
c	This routine searches for a real value alpha such that:
c
c	J(x + alpha p) < J(x)
c
c ==========================================================================================
c	Input/Output:
c ==========================================================================================
c
c	Input:
c
c		- nx,ny,nz	: mesh dimensions
c		- ipc,rpc	: level information
c		- ac,cc,fc	: level entries
c		- p		: the direction
c		- x		: the previous estimate of the solution
c		- r		: the previous non linear residual (r=r(x))
c
c	Temporary vectors:
c
c		- ap
c
c	Output:
c
c		- alpha
c
c ==========================================================================================
c	Declarations
c ==========================================================================================
c
	integer	nx,ny,nz,ipkey
	integer	niters,nitmax,ifail_tol
c
	integer	ipc(*)
c
        integer*1       grid_info(*)
c
	real*8	alpha
	real*8	ALP,pAp,pr,val,ALP_old
	real*8	zNN,zntol,eps,DzNN
        real*8  val2
        real*8  dotvect,dotvect3
c
	real*8	rpc(*),cc(*),fc(*)
	real*8	p(*),x(*),r(*),ap(*)
        real*8  xpos(*),ypos(*),zpos(*),vect(*)
c
c ==========================================================================================
c	Setup: non-linear iteration tolerance and itmax
c ==========================================================================================
c
	nitmax = 20
	zntol  = 1.d-5
	eps    = 1.d-4
	ifail_tol = 0
	ipkey = ipc(10)
c
c ==========================================================================================
c	Firt guess for alpha using linear case
c ==========================================================================================
c
        call residue_gpbl(x,nx,ny,nz,xpos,ypos,zpos,grid_info,
     1                  vect,fc,cc,ap)
        val = dotvect(nx,ny,nz,p,ap)
c
        call residue_gpbl(p,nx,ny,nz,xpos,ypos,zpos,grid_info,
     1                  vect,fc,cc,ap)
        pr = dotvect(nx,ny,nz,p,r)
        pAp = -dotvect(nx,ny,nz,p,ap) + dotvect(nx,ny,nz,p,fc)
        alpha = pr/pAp
c
	ALP = alpha
c
	call vectcopy(nx,ny,nz,x,r)
	call vectsmult(nx,ny,nz,ALP,p,r)
c
c	Evaluate residual of 1D system
c
        call residue_gpbl(r,nx,ny,nz,xpos,ypos,zpos,grid_info,
     1                  vect,fc,cc,ap)
c
	zNN = -dotvect(nx,ny,nz,p,ap) + val -pr
c
c ==========================================================================================
c	Finding alpha using 1D Newton method
c ==========================================================================================
c
	niters = 0
c
	if(niters.gt.nitmax) goto 200
c
	ifail_tol = 0
	ALP_old = ALP
c
100	continue
c
		niters = niters + 1
c
c		Construct derivatives of NN
c
		call vectsmult(nx,ny,nz,eps,p,r)
        	call residue_gpbl(r,nx,ny,nz,xpos,ypos,zpos,
     1			grid_info,vect,fc,cc,ap)
		val2 = -dotvect(nx,ny,nz,ap,p) + val -pr
		DzNN = (val2 - zNN)/eps
c
c		Update solution
c
		ALP_old = ALP
		ALP = ALP - zNN / DzNN
c
c		Evaluate residual of 1D system
c
		call vectcopy(nx,ny,nz,x,r)
		call vectsmult(nx,ny,nz,ALP,p,r)
c
        	call residue_gpbl(r,nx,ny,nz,xpos,ypos,zpos,
     1			grid_info,vect,fc,cc,ap)
c
		zNN = -dotvect(nx,ny,nz,p,ap) + val -pr
c
c		Tolerance and itmax checks
c
		if(dabs(zNN).lt.zntol.and.
     1		(dabs((ALP-ALP_old)/ALP).lt.zntol)) goto 300
		if(niters.gt.nitmax) goto 200
c
	goto 100
c
c	Tolerance not reached
c
200	continue
c
	ifail_tol = ifail_tol + 1
c
c	Tolerance reached
c
300	continue
c
c ==========================================================================================
c	Newton search complete: update solution value
c ==========================================================================================
c
	alpha = max(ALP,-0.8d0)
c
c	print message
c
c	if(ifail_tol.gt.0) then
c		write(6,*) 'LineSearch: exceeded 1D Newton itmax: ',
c     1			nitmax
c	endif
c
c ==========================================================================================
c	Return and end
c ==========================================================================================
c
	return
	end
