c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Newton_Solver_gpbl.f
c       (Newton-based PDE solver, specialized for GPBL equation, based on Michael Holst MG package)
c
c	Version 1:      7/4/07
c
c	Author: Patrice Koehl (in collaboration with Marc Delarue)
c
c	This file contains a series of subroutines for solving modified Poisson-Boltzmann
c	equations. It is inspired form Michael Holst MG code, which is available at:
c	http://www.fetk.org/codes/pmg/index.html
c
c ==========================================================================================
c ==========================================================================================
c 	Newton_Solver_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c	Newton_Solver implements the inexact Newton Multilevel solver applied by Michael
c	Holst for solving the Poisson-Boltzmann equation, and currently implemented in
c	APBS.
c
	subroutine Newton_Solver_gpbl(ijac,iparam,rparam,iwork,rwork,u,
     1				xpos,ypos,zpos,bndx,bndy,bndz,
     2				eps_X,eps_Y,eps_Z,helm,rhs,
     3				grid_info)
c
c	Input:
c               ijac:           flag for Jacobian considered:
c                               0: truncated Jacobian with stencil of size 7
c                               1: full Jacobian with stencil of size 27
c		iparam:		integer array: gives flags to Solver
c		rparam:		real array: constants for solver
c		iwork		integer work array
c		rwork		real work array
c
c	One dimensional:
c		xpos		positions of grid points along X axis
c		ypos		positions of grid points along Y axis
c		zpos		positions of grid points along Z axis
c
c	Two dimensional:
c		bndx		boundary conditions (X=1 and X=Nx planes)
c		bndy		boundary conditions (Y=1 and Y=Ny planes)
c		bndz		boundary conditions (Z=1 and Z=Nz planes)
c
c	Three dimensional:
c		eps_X		epsilon map, defined at positions (i+0.5,j,k)
c		eps_Y		epsilon map, defined at positions (i,j+0.5,k)
c		eps_Z		epsilon map, defined at positions (i,j,k+0.5)
c
c		helm		Helmholtz term in the equation
c		rhs		right hand side of the equation
c
c		grid_info	Contains info on:
c		(bits 0 and 1)	original grid
c		(bits 3 and 4)	projection of molecule on grid, defined at positions (i+0.5,j,k)
c		(bits 5 and 6)  projection of molecule on grid, defined at positions (i,j+0.5,k)
c		(bits 7 and 8)	projection of molecule on grid, defined at positions (i,j,k+0.5)
c
c	Output:
c		u		the dimensionless electrostatic potential map
c
c	Dimensions of all arrays have been set in the calling program
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
        integer         ijac
	integer		iparam(*),iwork(*)
	real*8		rparam(*),rwork(*),u(*)
	real*8		xpos(*),ypos(*),zpos(*),bndx(*),bndy(*),bndz(*)
	real*8		eps_X(*),eps_Y(*),eps_Z(*),helm(*),rhs(*)
c
	integer*1	grid_info(*)
c
c	Local variables:
c ==========================================================================================
c
	integer	i,nx,ny,nz,niwk,nrwk,nlevel,ierror,maxlevel,mxlevel
	integer	mgcoar,mgdisc,mgsolv
	integer k_iz,k_w1,k_w2
	integer k_ipc,k_rpc,k_ac,k_cc,k_fc,k_jac,k_vect
c
c	Variables from multigrid size
c ==========================================================================================
c
	integer	nxc,nyc,nzc,nf,nc,narr,narrc
	integer	n_rpc,n_ipc,n_iz,ireal_tot,iint_tot
c
c ==========================================================================================
c 	Decode input array iparam, and do some sanity check
c ==========================================================================================
c
	nrwk	= iparam(1)
	niwk	= iparam(2)
	nx	= iparam(3)
	ny	= iparam(4)
	nz	= iparam(5)
	nlevel	= iparam(6)
	mgcoar	= iparam(18)
	mgdisc	= iparam(19)
	mgsolv	= iparam(21)
c
	if(nlevel.le.0.or.nx.le.0.or.ny.le.0.or.nz.le.0) then
		write(6,*) 'nx,ny,nz,nlevel must be strictly positive!!'
		ierror = -1
		iparam(51) = ierror
		return
	endif
c
c	Check that the number of levels proposed is compatible with the grid size
c
	mxlevel = maxlevel(nx,ny,nz)
	if(nlevel.gt.mxlevel) then
		write(6,*) 'The max. # of levels for your grid is: ',
     1			mxlevel
                nlevel=mxlevel
c		ierror = -2
c		iparam(51) = ierror
c		return
	endif
c
c ==========================================================================================
c	Get grid size info for all grid levels: this will be used to partition the arrays
c	iwork and rwork so that they can store the whole hierarchy of grids
c ==========================================================================================
c
	call multigrid_size(ijac,mgcoar,mgdisc,mgsolv,nx,ny,nz,nlevel,
     1		nxc,nyc,nzc,nf,nc,narr,narrc,n_rpc,n_iz,n_ipc,
     2		ireal_tot,iint_tot)
c
c ==========================================================================================
c	Allocate space for one extra small vector:
c               - if a truncated Jacobian is used, then only need space
c                 for one plane (perpendicular to Z) of the grid (see
c                 Langevin7.f)
c               - if the full Jacobian is computed, needs space for
c                 10 extra planes (for derivatives; see Langevin27.f)
c ==========================================================================================
c
	ireal_tot = ireal_tot + nx*ny +10*ijac*nx*ny
c
c	Check that work arrays (both real and integer) have enough space
c ==========================================================================================
c
	if(nrwk.lt.ireal_tot) then
		write(6,*)'Real work space must be at least: ',ireal_tot
		ierror = -3
		iparam(51) = ierror
		return
	endif
c
	if(niwk.lt.iint_tot) then
		write(6,*) 'Integer work space must be at least: ',
     1			iint_tot
		ierror = -3
		iparam(51) = ierror
		return
	endif
c
c       Initialise the real work array
c ==========================================================================================
c
	do 100 i = 1,ireal_tot
		rwork(i) = 0
100	continue
c
c	Split the integer work array
c ==========================================================================================
c
	k_iz = 1
	k_ipc = k_iz + n_iz
c
c	Split the real work array
c ==========================================================================================
c
	k_rpc = 1
	k_cc  = k_rpc + n_rpc
	k_fc  = k_cc + narr
	k_w1  = k_fc + narr
	k_w2  = k_w1 + nf
	k_jac  = k_w2 + nf
        if(ijac.eq.0) then
	        k_vect = k_jac + 4*nf + 14*narrc
        else
	        k_vect = k_jac + 14*nf + 14*narrc
        endif
c
c ==========================================================================================
c	Start the multigrid solver
c ==========================================================================================
c
	call newton_mg_gpbl(ijac,iparam,rparam,nx,ny,nz,u,iwork(k_iz),
     1		rwork(k_w1),rwork(k_w2),iwork(k_ipc),rwork(k_rpc),
     2		rwork(k_jac),rwork(k_vect),rwork(k_cc),
     3          rwork(k_fc),xpos,ypos,zpos,bndx,bndy,bndz,eps_X,eps_Y,
     4          eps_Z,helm,rhs,grid_info)
c
c ==========================================================================================
c	The program should have found the solution...return to the
c	main program
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	Newton_mg.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine uses a Newton method, combined with a linear multigrid
c	solver, to solve modified Poisson-Boltzmann equations.
c
	subroutine newton_mg_gpbl(ijac,iparam,rparam,nx,ny,nz,u,iz,w1,
     1		w2,ipc,rpc,jac,vect,cc,fc,xpos,ypos,zpos,gxcf,
     2		gycf,gzcf,a1cf,a2cf,a3cf,ccf,fcf,grid_info)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer		iparam(*),ipc(*),iz(*)
	real*8		rparam(*),rpc(*),cc(*),fc(*),jac(*),vect(*)
	real*8		u(*),w1(*),w2(*)
	real*8		xpos(*),ypos(*),zpos(*),gxcf(*),gycf(*),gzcf(*)
	real*8		a1cf(*),a2cf(*),a3cf(*),ccf(*),fcf(*)
c
	integer*1	grid_info(*)
c
c	Local variables:
c ==========================================================================================
c
	integer		mgkey,nlevel,itmax,iok,itype,iinfo,istop,ipkey
	integer		nu1,nu2,ijac
	integer		nx,ny,nz,ilev,ido,iters,ierror,nlevel_real
	integer		mgprol,mgcoar,mgsolv,mgdisc,mgsmoo
c
c	real*8		epsmach
	real*8		epsilon,errtol,omegal,omegan
c
c ==========================================================================================
c 	Decode the iparam and rparam arrays
c ==========================================================================================
c
	nlevel = iparam(6)	! Number of mesh level
	nu1    = iparam(7)	! number of pre-smoothings
	nu2    = iparam(8)	! number of post-smoothings
	mgkey  = iparam(9)	! multigrid method: 0=variable, v-cycle, 1=nest operation
	itmax  = iparam(10)	! maximum number of iterations
	istop  = iparam(11)	! stopping criteria
	iinfo  = iparam(12)	! runtime status messages :
				! 0=none, 1=some, 2=many, 3=more
	ipkey  = iparam(14)	! Defines type of equation considered
				! 0: Laplace, 1: linear PB, 2: non linear PB
				! 3: Size modified PB, 4: Generalized PB Langevin
	mgprol = iparam(17)	! type of prolongation (0=trilinear)
	mgcoar = iparam(18)	! coarse equations (0=stan, 1=harm, 2=Galerkin)
	mgdisc = iparam(19)	! discretization (0=fv, 1=fe)
	mgsmoo = iparam(20)	! type of smoothing (0=wjac, 1=gsrb, 2=sor, 3=rich, 4=cghs)
	mgsolv = iparam(21)	! coarse equation solver (0=cghs, 1=nanded Linpack)
c
	errtol = rparam(1)	! tolerance factor
	omegal = rparam(9)	! linear relaxation parameter
	omegan = rparam(10)	! non linear relaxation parameter
c
c ==========================================================================================
c 	Build the multigrid data structure in array iz
c ==========================================================================================
c
	call build_structure(nx,ny,nz,nlevel,iz)
c
c ==========================================================================================
c 	Build right-hand-size on fine grid
c ==========================================================================================
c
	ido = 0
	call build_gpbl(nx,ny,nz,nlevel,ipkey,ido,ijac,iz,
     1	ipc,rpc,cc,fc,xpos,ypos,zpos,
     2	gxcf,gycf,gzcf,ccf,fcf,grid_info)
c
c ==========================================================================================
c 	Build right-hand-size and grid description on all coarse grids
c ==========================================================================================
c
	ido = 1
	call build_gpbl(nx,ny,nz,nlevel,ipkey,ido,ijac,iz,
     1	ipc,rpc,cc,fc,xpos,ypos,zpos,
     2	gxcf,gycf,gzcf,ccf,fcf,grid_info)
c
c ==========================================================================================
c 	Prepare for Newton solver and solve
c ==========================================================================================
c
c	Determine machine epsilon (hard coded)
c ==========================================================================================
c
c	epsilon = epsmach(0)
	epsilon = 1.d-9
c
c	Impose zero Dirichlet boundary conditions
c ==========================================================================================
c
	call set_bound_zero(nx,ny,nz,u)
c
c	Call multigrid solver
c ==========================================================================================
c
	nlevel_real = nlevel
c
	iok = 1
	ilev = 1
c
	if(mgkey.eq.0) then
		call newton_vcycle_gpbl(nx,ny,nz,u,ijac,iz,ccf,fcf,w1,
     1		w2,istop,itmax,iters,ierror,nlevel,ilev,nlevel_real,
     2		mgsolv,iok,iinfo,epsilon,errtol,omegan,nu1,nu2,
     3		mgsmoo,a1cf,a2cf,a3cf,ipc,rpc,cc,fc,jac,
     4		mgprol,mgcoar,mgdisc,xpos,ypos,zpos,grid_info,vect)
	elseif(mgkey.eq.1) then
		call newton_nested_gpbl(nx,ny,nz,u,ijac,iz,ccf,fcf,w1,
     1		w2,istop,itmax,iters,ierror,nlevel,ilev,nlevel_real,
     2		mgsolv,iok,iinfo,epsilon,errtol,omegan,nu1,nu2,
     3		mgsmoo,a1cf,a2cf,a3cf,ipc,rpc,cc,fc,jac,
     4		mgprol,mgcoar,mgdisc,xpos,ypos,zpos,grid_info,vect)
	else
		write(6,*) 'Bad mgkey!'
	endif
c
c	Restore boundary conditions
c ==========================================================================================
c
	call set_bound(nx,ny,nz,u,gxcf,gycf,gzcf)
c
c ==========================================================================================
c 	Return and end
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	newton_nested_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c	Newton_nested uses nested iteration for an inexact Newton multilevel method
c
	subroutine newton_nested_gpbl(nx,ny,nz,u,ijac,iz,w0,w1,w2,w3,
     1          istop,itmax,iters,ierror,nlevel,ilev,nlevel_real,mgsolv,
     2		iok,iinfo,epsilon,errtol,omega,nu1,nu2,mgsmoo,
     3		cprime,rhs,utemp,ipc,rpc,cc,fc,jac,mgprol,
     4		mgcoar,mgdisc,xpos,ypos,zpos,grid_info,vect)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer	nx,ny,nz
	integer	istop,itmax,iters,ierror,level,nlevel,ilev,nlevel_real
	integer	mgsolv,mgsmoo,iok,iinfo,nu1,nu2,ijac
	integer	mgprol,mgcoar,mgdisc
c
	integer	ipc(*),iz(50,*)
c
	real*8	epsilon,errtol,omega
	real*8	u(*),w0(*),w1(*),w2(*),w3(*)
	real*8	rpc(*),cc(*),fc(*),jac(*),vect(*)
	real*8	cprime(*),rhs(*),utemp(*)
	real*8	xpos(*),ypos(*),zpos(*)
c
	integer*1	grid_info(*)
c
c	Local variables:
c ==========================================================================================
c
	integer	nxf,nyf,nzf,nxc,nyc,nzc
	integer	itmxd,nleveld,iterd,iokd,istpd
c
	real*8	errd
c
c	Recover grid size
c
	nxf = nx
	nyf = ny
	nzf = nz
c
	call make_coarse(nlevel-1,nxf,nyf,nzf,nxc,nyc,nzc)
c
c	Move up grids: interpolate solution, and use Newton_vcycle
c
	do 100 level=nlevel_real,ilev+1,-1
c
c		errd = errtol
		itmxd = nlevel_real
		nleveld = nlevel_real - level + 1
		iterd = 0
		iokd = iok
		istpd = istop
		call newton_vcycle_gpbl(nxc,nyc,nzc,u,ijac,iz,w0,w1,w2,
     1          w3,istpd,itmxd,iterd,ierror,nleveld,level,nlevel_real,
     2          mgsolv,iokd,iinfo,epsilon,errd,omega,nu1,nu2,mgsmoo,
     3		cprime,rhs,utemp,ipc,rpc,cc,fc,jac,
     4		mgprol,mgcoar,mgdisc,xpos,ypos,zpos,grid_info,vect)
c
		call make_fine(1,nxc,nyc,nzc,nxf,nyf,nzf)
c
		call interpolate(nxc,nyc,nzc,nxf,nyf,nzf,u(iz(1,level)),
     1		u(iz(1,level-1)))
c
		nxc = nxf
		nyc = nyf
		nzc = nzf
c
100	continue
c
	level = ilev
	call newton_vcycle_gpbl(nxc,nyc,nzc,u,ijac,iz,w0,w1,w2,w3,istop,
     1	itmax,iters,ierror,nlevel,level,nlevel_real,mgsolv,
     2	iok,iinfo,epsilon,errtol,omega,nu1,nu2,mgsmoo,
     3	cprime,rhs,utemp,ipc,rpc,cc,fc,jac,
     4	mgprol,mgcoar,mgdisc,xpos,ypos,zpos,grid_info,vect)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	newton_vcycle.f
c ==========================================================================================
c ==========================================================================================
c
c	Newton_vcycle is the inexact Newton multilevel solver
c
	subroutine newton_vcycle_gpbl(nx,ny,nz,x,ijac,iz,w0,w1,w2,w3,
     1          istop,itmax,iters,ierror,nlev,ilev,nlevel_real,mgsolv,
     2		iok,iinfo,epsilon,errtol,omega,nu1,nu2,mgsmoo,
     3		cprime,rhs,xtemp,ipc,rpc,cc,fc,jac,
     4		mgprol,mgcoar,mgdisc,xpos,ypos,zpos,
     5		grid_info,vect)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables:
c ==========================================================================================
c
	integer	nx,ny,nz
	integer	istop,itmax,iters,ierror,nlev,ilev,nlevel_real,ijac
	integer	mgsolv,iok,itype,ichange,iinfo,nu1,nu2,mgsmoo
	integer	mgprol,mgcoar,mgdisc
c
	integer	ipc(*),iz(50,*)
c
	real*8	epsilon,errtol,omega
	real*8	x(*),w0(*),w1(*),w2(*),w3(*)
	real*8	rpc(*),cc(*),fc(*),jac(*),vect(*)
	real*8	cprime(*),rhs(*),xtemp(*)
	real*8	xpos(*),ypos(*),zpos(*)
c
	integer*1	grid_info(*)
c
c	Local variables
c ==========================================================================================
c
	integer	level,lev,i
	integer	itmax_s, iters_s, ierror_s, iok_s, iinfo_s,istop_s
        integer mgsmoo_s
	integer	iter_d,itmax_d,mode,idamp,ipkey
c
	real*8	xnorm1
	real*8	bigc,ord
	real*8	rsden,rsnrm,orsnrm,free_energy
	real*8	xnorm_old,xnorm_new,damp,xnorm_med
	real*8	errtol_s
c
c ==========================================================================================
c	Initialization
c ==========================================================================================
c
        iok = 2
	itype = 1
        ichange = 1
c
c	Recover level information
c ==========================================================================================
c
	level = 1
	lev  = (ilev-1) + level
c
c	Start clock and printout
c ==========================================================================================
c	
	if(iok.eq.1) then
		call print_ini
c		call print_residual(iok,-1,0.0d0,0.0d0,0.0d0)
	elseif(iok.eq.2) then
		call print_ini_free
c		call print_residual_free(iok,-1,0.0d0,0.0d0,0.0d0,0.0d0)
	endif
c
c	Compute denominator for stopping criteria
c ==========================================================================================
c	
c	We only implement the relative residual
c
	if(istop.eq.1) then
c
c		Compute initial residual with zero initial guess
c		This is analogous to simply taking the norm of the rhs of the
c		equation
c
		call set_all_zeros(nx,ny,nz,w1)
                call free_energy_langevin_zero(nx,ny,nz,xpos,ypos,zpos,
     1          grid_info,cc,free_energy)
c
		rsden = xnorm1(nx,ny,nz,fc(iz(1,lev)))
c
	else
		write(6,*) 'Bad istop value ....'
		stop
	endif
c
	if(rsden.eq.0.0d0) then
		rsden = 1.d0
		write(6,*) 'Warning: rhs is zero ....'
	endif
c
	rsnrm = rsden
	orsnrm = rsnrm
c
	if(iok.eq.1) then
		call print_residual(iok,0,rsnrm,rsden,orsnrm)
	elseif(iok.eq.2) then
		call print_residual_free(iok,0,rsnrm,rsden,orsnrm,
     1			free_energy)
	endif
c
c ==========================================================================================
c	Newton iteration
c ==========================================================================================
c
c	Begin: compute residual with the initial guess
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,x(iz(1,lev)))
c
	call operatorJ_gpbl(nx,ny,nz,nlevel_real,ijac,iz,ipc,
     1		rpc,x,jac,xpos(iz(8,lev)),ypos(iz(9,lev)),
     2          zpos(iz(10,lev)),fc(iz(1,lev)),cc(iz(1,lev)),w0,
     3		grid_info(iz(1,lev)),free_energy,iok,itype,vect)
c
	xnorm_old = xnorm1(nx,ny,nz,w0)
c
c	Loop over all levels
c ==========================================================================================
c
	idamp = 1
	iters = 0
c
100	continue
c
		iters = iters + 1
c
c		Define rhs for this iteration
c
		ipkey = ipc(10)
c
		call set_rhs(nx,ny,nz,nlevel_real,iz,ilev,ipkey,
     1		x,w0,cprime,rhs)
c
c		Define tolerance, to get global-superlinear convergence
c               ========================================================
c
		bigc = 1.0d0
		ord  = 2.0d0
		errtol_s = dmin1((0.9d0*xnorm_old),
     1				 (bigc*xnorm_old**ord))
c
c		Solve the Newton equations using linear multigrid
c               ========================================================
c
		call set_all_zeros(nx,ny,nz,xtemp(iz(1,lev)))
c
c		Set parameters for multigrid
c		
		itmax_s  = 10
		istop_s  = 0
		iters_s  = 0
		ierror_s = 0
c
		iok_s    = 2
		iinfo_s  = iinfo
                mgsmoo_s = 1
c
		call lin_mg_vcycle(nx,ny,nz,xtemp,iz,w0,w1,w2,w3,
     1		istop_s,itmax_s,iters_s,ierror_s,
     2		nlev,ilev,nlevel_real,mgsolv,
     3		iok_s,iinfo_s,epsilon,errtol_s,omega,nu1,nu2,mgsmoo,
     4		ipc,rpc,jac,cprime,rhs)
c
                xnorm_new = xnorm1(nx,ny,nz,xtemp(iz(1,lev)))
c
c
c		(note that now rhs and cprime are available as storage arrays)
c
c		Apply damping, if still enabled
c               ========================================================
c
		if(idamp.eq.1) then
c
c		    try the correction
c		    ==================
c
		    call vectcopy(nx,ny,nz,x(iz(1,lev)),w1)
c
		    damp = 1.d0
c
		    call vectsmult(nx,ny,nz,damp,xtemp(iz(1,lev)),w1)
c
		    call residue_gpbl(w1,nx,ny,nz,
     1                  xpos,ypos,zpos,grid_info,vect,fc,cc,w0)
c
		    xnorm_new = xnorm1(nx,ny,nz,w0)
c
		    damp    = 1.0d0
		    iter_d  = 0
		    itmax_d = 5
		    mode    = 0
c
200		    continue
c
			if(iter_d.ge.itmax_d) goto 300
c
			if(mode.eq.0) then
				if(xnorm_new.lt.xnorm_old) mode = 1
			else
				if(xnorm_new.gt.xnorm_med) goto 300
			endif
c
c			Store old solution and residual
c               	===============================================
c
			call vectcopy(nx,ny,nz,w1,w2)
c			call vectcopy(nx,ny,nz,w0,w3)
c
			xnorm_med = xnorm_new
c
c			New damped correction, residual, and its norm
c               	===============================================
c
			call vectcopy(nx,ny,nz,x(iz(1,lev)),w1)
			damp = damp / 2.0d0
			call vectsmult(nx,ny,nz,damp,xtemp(iz(1,lev)),
     1			w1)
c
		    	call residue_gpbl(w1,nx,ny,nz,
     1                  xpos,ypos,zpos,grid_info,vect,fc,cc,w0)
c
			xnorm_new = xnorm1(nx,ny,nz,w0)
c
			iter_d = iter_d + 1
c
		    goto 200
c
300		    continue
c
c
c		Accept the previous Newton step
c               ========================================================
c
		     call vectcopy(nx,ny,nz,w2,x(iz(1,lev)))
c                    if(xnorm_new.lt.xnorm_old) then
c		        call vectcopy(nx,ny,nz,w2,x(iz(1,lev)))
c                    else
c                        if(ichange.eq.1) then
c		                call vectcopy(nx,ny,nz,w2,x(iz(1,lev)))
c                        endif
c                        itype = 0
c                        ichange = 1
c                    endif
c
		    call operatorJ_gpbl(nx,ny,nz,nlevel_real,ijac,iz,
     1			ipc,rpc,x,jac,xpos,ypos,zpos,fc(iz(1,lev)),
     2			cc(iz(1,lev)),w0,grid_info,
     3			free_energy,iok,itype,vect)
c
		    xnorm_med = xnorm1(nx,ny,nz,w0)
c
		    xnorm_new = xnorm_med
		    xnorm_old = xnorm_new
c
c		Determine whether or not to disable damping
c               ========================================================
c
c		    if((iter_d-1).eq.0) then
c			idamp = 0
c		    endif
c
		else
c
c		    Damping is disabled. Accept the Newton step
c                   ========================================================
c
		    damp = 1.d0
		    call vectsmult(nx,ny,nz,damp,xtemp(iz(1,lev)),
     1		    x(iz(1,lev)))
c
		    call operatorJ_gpbl(nx,ny,nz,nlevel_real,ijac,iz,
     1			ipc,rpc,x,jac,xpos,ypos,zpos,fc(iz(1,lev)),
     2			cc(iz(1,lev)),w0,grid_info,
     3			free_energy,iok,itype,vect)
c
		    xnorm_new = xnorm1(nx,ny,nz,w0)
		    xnorm_old = xnorm_new
c
		endif
c
c		Check stopping criteria
c
		if(iok.ne.0) then
			orsnrm = rsnrm
			if(istop.eq.1) then
				rsnrm = xnorm_new
			else
				write(6,*) 'Bad istop value ....'
			endif
			if(iok.eq.1) then
				call print_residual(iok,iters,rsnrm,
     1				rsden,orsnrm)
			elseif(iok.eq.2) then
				call print_residual_free(iok,iters,
     1				rsnrm,rsden,orsnrm,free_energy)
			endif
c                        if((iters.gt.10.or.rsnrm/rsden.lt.1.d-3).
c     1                  and.ichange.eq.0) itype=1
			if((rsnrm/rsden).le.errtol) goto 400
		endif
c
		if(iters.ge.itmax) goto 400
c
	goto 100
c
400	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	set_rhs.f
c ==========================================================================================
c ==========================================================================================
c
c	Define the residual to be passed to the linear solver
c
	subroutine set_rhs(nx,ny,nz,nlev_real,iz,lev,ipkey,x,r,
     1			cprime,rhs)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input and local variables:
c ==========================================================================================
c
	integer	nx,ny,nz,nlev_real,lev,level
	integer	nxx,nyy,nzz,nxold,nyold,nzold,ipkey
c
	integer	iz(50,*)
c
	real*8	x(*),r(*),cprime(*),rhs(*)
c
c	Setup
c ==========================================================================================
c
	nxx = nx
	nyy = ny
	nzz = nz
c
c	Form the rhs of the Newton system: just current residual
c
	call vectcopy(nx,ny,nz,r,rhs(iz(1,lev)))
c
c	Set vector cprime to 0
c
	call set_all_zeros(nx,ny,nz,cprime(iz(1,lev)))
c
c	Build the nlev-1 level operators
c
	do 100 level = lev+1, nlev_real
c
		nxold = nxx
		nyold = nyy
		nzold = nzz
c
		call make_coarse(1,nxold,nyold,nzold,nxx,nyy,nzz)
c
		call restrict(nxold,nyold,nzold,nxx,nyy,nzz,
     1		rhs(iz(1,level-1)),rhs(iz(1,level)))
c
		call restrict(nxold,nyold,nzold,nxx,nyy,nzz,
     1		cprime(iz(1,level-1)),cprime(iz(1,level)))
c
100	continue
c
	return
	end
c
