c ==========================================================================================
c ==========================================================================================
c	Define_Grid.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	6/27/07
c	Author:		Patrice Koehl (based on Cyril Azuaras program)
c			in collaboration with Marc Delarue
c
c	Purpose:	Defines grid and check if the molecule fits in it
c
c			If box size was not preset, set it such that
c			distance from molecule to side of the box
c			is 2* Bjerrum length
c			(The Bjerrum length lB is the separation at which the
c			electrostatic interaction between two elementary
c			charges is equal to kb T, i.e.:
c			kb T = e^2/(4 pi eps lB)
c			under normal conditions, lB = 7 angstroms)
c
c
c	Input:
c		Grid:	
c			grid_type		 :  flag:
c							0: uniform grid (same spacing
c							   in X, Y and Z)
c							1: non uniform grid
c			Npointx, Npointy, Npointz:  number of points in each dimension
c			sizex,sizey,sizez	 :  grid size in each dimension
c			bnd_type		 : type of boundary condition
c						   considered
c
c		Charges:
c			atom_coord, atom_radius, atom_charge: position, radius and
c							      charge of each atom
c
c		l_Bjerrum : Bjerrum length (see main for info)
c
c	(both informations provided in common blocks)
c
c	Output:
c		Grid properly dimensioned
c		Charges centered on the center of the box
c
	subroutine Define_Grid(grid_type,l_log,bnd_type,hx,hy,hz,
     1                          sizex,sizey,sizez,nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
	integer	i,j,l_log,bnd_type
        integer nx,ny,nz
c
c	Information on grid
c ==========================================================================================
c
	integer	grid_type
	real*8	sizex,sizey,sizez,size,size_min
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	shiftx,shifty,shiftz
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Constants
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
c	Local arrays
c ==========================================================================================
c
	real*8	x,old_size,old_sizex,old_sizey,old_sizez
	real*8	grid_max,molecule_max
	real*8	grid_x,grid_y,grid_z
	real*8	VolProt,VolBox,Filling
	real*8	mindistx,mindisty,mindistz
        real*8  diffxy,diffxz,diffx,diffy,diffz
	real*8	mol_size(3)
	real*8	mol_max(3),mol_min(3)
	real*8	Center(3),c_mol(3)
c
c       Pointers
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c	Common blocks
c ==========================================================================================
c
        common /charges/ ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                   ptr_label2,iform,natom,ncharge
	common /centers/ c_mol,Center
        common/mol_size/ mol_min,mol_max
c
	common /constants/ l_Bjerrum,pi,Temp,KB,electron,Ang,Avogadro,
     1			   eps0,beta,C_charge,C_prop
	common /molshift/  shiftx,shifty,shiftz
c
1	format(70('='),/,2('='),'Information on grid       ',40x,
     1	2('='),/,70('='),/)  
2	format('Number of grid points in X, Y and Z    : ',i4, 
     1	2(' x ',i4))
3	format('Grid spacing in X, Y and Z             : ',
     1	3(f8.3,4x))
11	format('Total grid size                        : ',
     1	3(f8.3,4x))
4	format('Total volume of the grid               : ',f15.3)
c5	format('Approximate volume of the protein      : ',f15.3)
6	format('Percent filling (VolProt/VolBox)       : ',f8.3)
7	format('Distances mol / box (X, Y and Z)       : ',
     1	2(f8.3,',',2x),f8.3)
8	format(2('	'),'Warning: grid ',a5,' has been reajusted:',/,
     1	3('	'),'Old grid ',a5,':',f8.3,/,
     2	3('	'),'New grid ',a5,':',f8.3)
9	format('Warning: grid ',a5,' has been reajusted:',/,
     1	'Old grid ',a5,':',f8.3,'	New grid ',a5,':',f8.3)
10	format(2('	'),'Percent filling (VolProt/VolBox)       : ',
     1	f8.3)
12	format('Shift of molecule center            : ',3(f8.3,2x))
c
c ==========================================================================================
c = Write header in log file   =============================================================
c ==========================================================================================
c
	write(l_log,1)
c
c ==========================================================================================
c = Compute molecule size      =============================================================
c ==========================================================================================
c
	do 100 i = 1,3
		mol_min(i) = atom_coord(i)-atom_radius(1)
		mol_max(i) = atom_coord(i)+atom_radius(1)
100	continue
c
	do 300 i = 2,natom
c
		do 200 j = 1,3
			x = atom_coord(3*(i-1)+j)
			if(x+atom_radius(i).gt.mol_max(j)) then
				mol_max(j) = x + atom_radius(i)
			endif
			if(x-atom_radius(i).lt.mol_min(j)) then
				mol_min(j) = x - atom_radius(i)
			endif
200		continue
c
300	continue
c
	do 400 i = 1,3
c		molmaxtmp = mol_max(i)
c		molmintmp = mol_min(i)
c		mol_max(i) = max(mol_max(i),abs(mol_min(i)))
c		mol_min(i) = -mol_max(i)
		mol_size(i) = mol_max(i)-mol_min(i)
400	continue
c
c ==========================================================================================
c = Recenter molecule, such that it fits in box     ========================================
c ==========================================================================================
c
	Center(1) = 0
	Center(2) = 0
	Center(3) = 0
c
c	We also shift the molecule by shiftx, shifty, shiftz : this is useful when studying
c	for example a complex: each monomer can be studied individually, in the same position
c	as in the complex
c
	Center(1) = Center(1) + shiftx
	Center(2) = Center(2) + shifty
	Center(3) = Center(3) + shifty
c
	do 600 i = 1,ncharge
c
		do 500 j = 1,3
			atom_coord(3*(i-1)+j) = atom_coord(3*(i-1)+j)
     1				+Center(j)
500		continue
c
600	continue
c
	do 700 i = 1,3
		mol_min(i) = mol_min(i) + Center(i)
		mol_max(i) = mol_max(i) + Center(i)
700	continue
c
c ==========================================================================================
c = Adjust grid size, if needed   ==========================================================
c ==========================================================================================
c
        diffx = mol_max(1) - mol_min(1)
        diffy = mol_max(2) - mol_min(2)
        diffz = mol_max(3) - mol_min(3)
c
        diffxy = diffy-diffx
        diffxz = diffz-diffx
        if(diffxy.gt.0.2*diffx.or.diffxz.gt.0.2*diffx) then
                if(grid_type.eq.0) grid_type = 1
        endif
c
c
c	Consider two cases: symmetric grid (same size in all 3 dimensions), 
c	or non symmetric grid
c
	old_sizex = sizex
	old_sizey = sizey
	old_sizez = sizez
c
	if(grid_type.eq.0) then
c
		molecule_max = max(mol_size(1),mol_size(2),mol_size(3))
c
c		Make sure that box is at least 1.5 time dimension of
c		molecule
c
		size_min = 1.5d0 * molecule_max
c
c		If 0 boundary condition is used, make sure that molecule
c		is at least 2*L_Bjerrum_water away from grid boundary
c		for water, use eps=80
c
		if(bnd_type.eq.0) size_min = molecule_max + 4*l_Bjerrum
     1					/80.d0
c
		grid_max = sizex
		if(sizey.gt.grid_max) grid_max = sizey
		if(sizez.gt.grid_max) grid_max = sizez
		old_size = grid_max
		grid_x = sizex*dble(nx-1)
		grid_y = sizey*dble(ny-1)
		grid_z = sizez*dble(nz-1)
		grid_max = max(grid_x,grid_y,grid_z)
c
		if(grid_max.lt.size_min) then
c
			write(6,*) ' '
			write(l_log,*) ' '
c
			grid_max = size_min
c
			sizex = grid_max/dble(nx-1)
			sizey = grid_max/dble(ny-1)
			sizez = grid_max/dble(nz-1)
			size = max(sizex,sizey,sizez)
			sizex = size
			sizey = size
			sizez = size
			write(6,8) 'size ','size ',old_size,'size ',size
			write(l_log,9) 'size ','size ',old_size,'size ',
     1			size
			write(6,*) ' '
			write(l_log,*) ' '
c
		endif
c
	else
c
		size_min = 1.5d0 * mol_size(1)
		if(bnd_type.eq.0) size_min = mol_size(1) + 4*l_Bjerrum
     1						/80.d0
		grid_x = sizex*dble(nx-1)
c
		if(grid_x.lt.size_min) then
			write(6,*) ' '
			write(l_log,*) ' '
			grid_max = size_min
			sizex = grid_max/dble(nx-1)
			write(6,8) 'sizex','sizex',old_sizex,'sizex',
     1			sizex
			write(l_log,9)'sizex','sizex',old_sizex,'sizex',
     1			sizex
		endif
c
		size_min = 1.5d0 * mol_size(2)
		if(bnd_type.eq.0) size_min = mol_size(2) + 4*l_Bjerrum
     1						/80.d0
		grid_y = sizey*dble(ny-1)
c
		if(grid_y.lt.size_min) then
			write(6,*) ' '
			write(l_log,*) ' '
			grid_max = size_min
			sizey = grid_max/dble(ny-1)
			write(6,8) 'sizey','sizey',old_sizey,'sizey',
     1			sizey
			write(l_log,9)'sizey','sizey',old_sizey,'sizey',
     1			sizey
		endif
c
		size_min = 1.5d0 * mol_size(3)
		if(bnd_type.eq.0) size_min = mol_size(3) + 4*l_Bjerrum
     1						/80.d0
		grid_z = sizez*dble(nz-1)
c
		if(grid_z.lt.size_min) then
			write(6,*) ' '
			write(l_log,*) ' '
			grid_max = size_min
			sizez = grid_max/dble(nz-1)
			write(6,8) 'sizez','sizez',old_sizez,'sizez',
     1			sizez
			write(l_log,9)'sizez','sizez',old_sizez,'sizez',
     1			sizez
		endif
c
	endif
c
c ==========================================================================================
c = Set coordinates of grid points                  ========================================
c = (Grid is centered on (0,0,0)                    ========================================
c ==========================================================================================
c
	do 800 i = 1,nx
		hx(i) = (i-1)*sizex - sizex*dble(nx/2)
800	continue
c
	do 900 i = 1,ny
		hy(i) = (i-1)*sizey - sizey*dble(ny/2)
900	continue
c
	do 1000 i = 1,nz
		hz(i) = (i-1)*sizez - sizez*dble(nz/2)
1000	continue
c
        mindistx = min(mol_min(1)-hx(1),hx(nx)-mol_max(1))
        mindisty = min(mol_min(2)-hy(1),hy(ny)-mol_max(2))
        mindistz = min(mol_min(3)-hz(1),hz(nz)-mol_max(3))
c
c ==========================================================================================
c = Compute filling ratio         ==========================================================
c ==========================================================================================
c
c	Computer "filling", i.e. space occupied by the protein compared
c	to full space available in the grid
c
	VolBox = (nx-1)*(ny-1)*(nz-1)*sizex*sizey*sizez
	VolProt = mol_size(1)*mol_size(2)*mol_size(3)
c
	Filling = 100.d0*VolProt/VolBox
	write(6,10) filling
	write(6,*) ' '
	write(6,*) "VolProt = ",VolProt
	write(6,*) "VolBox  = ",VolBox
	write(6,*) "mol_mins  ",mol_min(1),mol_min(2),mol_min(3)
	write(6,*) "mol_maxs  ",mol_max(1),mol_max(2),mol_max(3)
	write(6,*) "mol_sizes ",mol_size(1),mol_size(2),mol_size(3)
c
c ==========================================================================================
c = Report info in log file                         ========================================
c ==========================================================================================
c
	write(l_log,2) nx,ny,nz
	write(l_log,3) sizex,sizey,sizez
	write(l_log,11) (nx-1)*sizex,(ny-1)*sizey,
     1		(nz-1)*sizez
	if(shiftx.ne.0.and.shifty.ne.0.and.shiftz.ne.0) then
		write(l_log,12) shiftx,shifty,shiftz
	endif
	write(l_log,4) VolBox
c	write(l_log,5) VolProt
	write(l_log,6) Filling
	write(l_log,7) mindistx,mindisty,mindistz
	write(l_log,*) ' '
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	GenSurf.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Version 2:	7/13/07	: Added molecular surface
c	Version 3:	8/27/07 : routine can be used to define epsilon maps
c
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	Partition the grid: define grid point inside the molecule(s)
c	as 1, and other grid points as 0
c	At this stage, the surface of the molecule is simple: it is either the
c	VdW surface, or the solvent accessible surface (SAS), for which each atom has been
c	inflated by R_probe (1.4 angstroms for the classical SAS).
c
c	Information are stored as bytes
c
c	Input:
c		surf_type:	type of surface considered:
c					0 : Accessible surface area
c					1 : Molecular surface area
c		R_probe:	radius of probe to compute accessible surface area
c		iflag:		flag:
c					- 0 : do not compute surf,vol
c					- 1 : compute surf, volume
c		l_log:		logical unit number for log file
c
c	Available through common blocks:
c		NpointSurf	: number of points sampling the accessible surface
c				  area
c		Pointsurf	: coordinates of these points
c
	subroutine GenSurf(surf_type,R_probe,grid_partition,hx,hy,hz,
     1                  nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
c	Information on grid
c ==========================================================================================
c
        integer nx,ny,nz
c
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Partition of the grid
c
	integer*1	grid_partition(nx,ny,nz)
c
c	Points on the accessible surface area
c ==========================================================================================
c
	integer	NpointSurf
	real*8	PointSurf(*)
c
c
c	Information on charges
c ==========================================================================================
c
	integer	natom,ncharge,iform
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	ipoint,istepx,istepy,istepz
	integer	ix,iy,iz,kx,ky,kz
	integer surf_type
c
	real*8	x,y,z
	real*8	valx,valy,valz,val,rad,rad2
	real*8	R_probe,R_probe2
	real*8	vol,surf
c
c       Pointers
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        pointer (ptr_point,PointSurf)
c
c	Common blocks
c ==========================================================================================
c
        common /charges/ ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                   ptr_label2,iform,natom,ncharge
c
	common /space/   ptr_point,NpointSurf
c
c ==========================================================================================
c = Initialise                 =============================================================
c ==========================================================================================
c
c	Set all grid point to 0 (i.e. solvent)
c
	do 300 k=1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
				grid_partition(i,j,k) = 0
100			continue
200		continue
300	continue
c
c ==========================================================================================
c = Surface type 1: solvent accessible surface =============================================
c ==========================================================================================
c
c
c	To define accessible surface:
c	loop over all atoms: find grid cell to which it belongs, and loop
c	over its neighbour vertices to define points that are in the molecule
c
c	Start loop over all atoms
c
	do 700 ipoint = 1,natom
c
c	   Find grid cell to which it belongs (coordinate of the Bottom,
c	   Left, Front vertex)
c
	   x = atom_coord(3*(ipoint-1) +1)
	   y = atom_coord(3*(ipoint-1) +2)
	   z = atom_coord(3*(ipoint-1) +3)
c
	   call locate_cell(x,hx,ix,nx)
	   call locate_cell(y,hy,iy,ny)
	   call locate_cell(z,hz,iz,nz)
c
	   rad = atom_radius(ipoint)
	   rad = rad + R_probe
	   rad2 = rad*rad
c
c	   Number of vertices to check in each dimension:
c
	   call defstep(rad,hx,ix,istepx,nx)
	   call defstep(rad,hy,iy,istepy,ny)
	   call defstep(rad,hz,iz,istepz,nz)
c
c	   Now loop over all these vertices
c
	   do 600 k = 1,2*istepz + 1
c
		kz = iz - istepz + k -1
		kz=max(1,kz)
		kz=min(kz,nz)
		valz = (z - hz(kz))**2
c
	   	do 500 j = 1,2*istepy + 1
c
		   ky = iy - istepy + j -1 
		   ky=max(1,ky)
		   ky=min(ky,ny)
		   valy=(y-hy(ky))**2
c
		   do 400 i = 1,2*istepx + 1
c
			   kx = ix - istepx + i -1
			   kx=max(1,kx)
			   kx=min(kx,nx)
			   if(grid_partition(kx,ky,kz).eq.0) then
				valx = (x - hx(kx))**2
				val=valx+valy+valz
c
c				if vertex inside solvated ball ipoint,
c				set bit zero to 1
c				(all vertices with bit 0 set to 1
c				belongs to the solvated volume of the molecule)
c
				if(val.le.rad2) then
				    grid_partition(kx,ky,kz) = 1
				endif
c
			   endif
c
400		   continue
500		continue
600	   continue
c
700	continue
c
c	We are done if we wanted to define the accessible surface area
c
	if(surf_type.eq.0) return
c
c ==========================================================================================
c = Surface type 2: Molecular surface        ===============================================
c ==========================================================================================
c
c
c	Now that we have identified the surface, need to go inward, and remove all 
c	points that are within R_probe from points sampling the accessible
c	surface area: this will provide	the molecular surface area
c
c	Number of vertices to check in each dimension:
c
	R_probe2 = R_probe
c
	rad2 = R_probe2*R_probe2
c
	do 1100 ipoint = 1,NpointSurf
c
c	   Find grid cell to which it belongs (coordinate of the Bottom,
c	   Left, Front vertex)
c
	   x = PointSurf(3*(ipoint-1)+1)
	   y = PointSurf(3*(ipoint-1)+2)
	   z = PointSurf(3*(ipoint-1)+3)
c
	   call locate_cell(x,hx,ix,nx)
	   call locate_cell(y,hy,iy,ny)
	   call locate_cell(z,hz,iz,nz)
c
	   call defstep(R_probe2,hx,ix,istepx,nx)
	   call defstep(R_probe2,hy,iy,istepy,ny)
	   call defstep(R_probe2,hz,iz,istepz,nz)
c
c	   Loop over all nearby vertices
c
	   do 1000 k = 1,2*istepz + 1
c
		kz = iz - istepz + k -1
		kz=max(1,kz)
		kz=min(kz,nz)
		valz = (z - hz(kz))**2
c
	   	do 900 j = 1,2*istepy + 1
c
		   ky = iy - istepy + j -1 
		   ky=max(1,ky)
		   ky=min(ky,ny)
		   valy = (y -hy(ky))**2
c
		   do 800 i = 1,2*istepx + 1
c
			   kx = ix - istepx + i -1
			   kx=max(1,kx)
			   kx=min(kx,nx)
			   if(grid_partition(kx,ky,kz).eq.1) then
				valx = (x -hx(kx))**2
				val=valx+valy+valz
c
c				if vertex inside probe centered on the
c				point on the accessible surface,
c				if should be reset to be in the solvent
c
				if(val.le.rad2) then
				    grid_partition(kx,ky,kz)=0
				endif
c
			   endif
c
800		   continue
900		continue
1000	   continue
c
1100	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	GenIonZone.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	Partition the grid: define ions excluded zone (set grid_partition for the
c	vertices in this zone to 2)
c	We use the method for computing the accessible surface area, and check
c	all points inside that surface, but not inside the molecular surface,
c	to 2.
c	The probe for this ion-excluded zone is R_ion
c
c	Information are stored as bytes
c
c	Input:
c		R_ion:	radius of probe to compute ion excluded volume
c
	subroutine GenIonZone(R_ion,grid_partition,hx,hy,hz,nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
c	Information on grid
c
        integer nx,ny,nz
c
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Partition of the grid
c
	integer*1	grid_partition(nx,ny,nz)
c
c	Information on charges
c
	integer	natom,ncharge,iform
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Local variables
c
	integer	i,j,k
	integer	ipoint,istepx,istepy,istepz
	integer	ix,iy,iz,kx,ky,kz
c
	real*8	valx,valy,valz,rad,rad2
	real*8	R_ion
c
c       Pointers
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c	Common blocks
c ==========================================================================================
c
        common /charges/ ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                   ptr_label2,iform,natom,ncharge
c
c ==========================================================================================
c = Ion excluded volume                      ===============================================
c ==========================================================================================
c
c	To define ion excluded volume:
c		loop over all atoms: find grid cell to which it belongs, and loop
c		over its neighbour vertices to define points that are in the ion
c		excluded volume. Tag all of these points that are not inside the
c		molecular surface to 2
c
c	Start loop over all atoms
c
	do 700 ipoint = 1,natom
c
c		Find grid cell to which it belongs (coordinate of the Bottom,
c		Left, Front vertex)
c
		call locate_cell(atom_coord(3*(ipoint-1)+1),hx,ix,
     1				 nx)
		call locate_cell(atom_coord(3*(ipoint-1)+2),hy,iy,
     1				 ny)
		call locate_cell(atom_coord(3*(ipoint-1)+3),hz,iz,
     1				 nz)
c
		rad = atom_radius(ipoint) + R_ion
		rad2 = rad*rad
c
c		Number of vertices to check in each dimension:
c
		call defstep(rad,hx,ix,istepx,nx)
		call defstep(rad,hy,iy,istepy,ny)
		call defstep(rad,hz,iz,istepz,nz)
c
c		Now loop over all these vertices
c
		do 600 k = 1,2*istepz + 1
c
		   kz = iz - istepz + k -1
		   valz=(atom_coord(3*ipoint)-hz(kz))**2
c
		  do 500 j = 1,2*istepy + 1
c
			ky = iy - istepy + j -1 
		   	valy=(atom_coord(3*ipoint-1)-hy(ky))**2
c
			do 400 i = 1,2*istepx + 1
c
			   kx = ix - istepx + i -1
			   valx = (atom_coord(3*ipoint-2)-hx(kx))**2
c
c			   if vertex inside ball ipoint,
c			   set to 2
c
			   if(valx+valy+valz.le.rad2) then
				if(grid_partition(kx,ky,kz).eq.0) then
				   grid_partition(kx,ky,kz)=2
				endif
			   endif
c
400			continue
500		   continue
600	  	continue
c
700	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Compute_Vol_Surf.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine computes the approximate surface area and volume of the
c	molecule, based on the current status of the grid
c
	subroutine compute_vol_surf(l_log,surf_type,grid_partition,
     1			hx,hy,hz,nx,ny,nz)
c
	include 'poisson.h'
c
	integer	i,j,k,nvol,nsurfxy,nsurfxz,nsurfyz
	integer l_log,surf_type
        integer nx,ny,nz
c
	integer*1	grid_partition(nx,ny,nz)
c
	real*8	sizex,sizey,sizez,dx,dy,dz
	real*8	vol,surf
c
	real*8	hx(nx),hy(ny),hz(nz)
c
1	format(2('	'),'Approx. ASA of the molecule        :',f10.3)
2	format(2('	'),'Approx. Solv. Exc. Vol. of mol.    :',f10.3)
3	format(2('	'),'Approx. Mol. Surf. of the molecule :',f10.3)
4	format(2('	'),'Approx. Mol. Vol. of the molecule  :',f10.3)
5	format('Approx. ASA of the molecule            :',f10.3)
6	format('Approx. Solv. Exc. Vol. of mol.        :',f10.3)
7	format('Approx. Mol. Surf. of the molecule     :',f10.3)
8	format('Approx. Mol. Vol. of the molecule      :',f10.3)
c
c ==========================================================================================
c	Initialize
c ==========================================================================================
c
	nvol = 0
	nsurfxy = 0
	nsurfxz = 0
	nsurfyz = 0
c
c ==========================================================================================
c	Compute surface area and volume : sum over all grid cells labeled
c	inside for volume, and over all grid cell face that touches solvent
c	for surface
c ==========================================================================================
c
	vol = 0
	surf = 0
c
	do 300 k=1,nz-1
	   dz = hz(k+1) - hz(k)
	   do 200 j = 1,ny-1
		dy = hy(j+1)-hy(j)
		do 100 i = 1,nx-1
c
		   dx = hx(i+1)-hx(i)
c
c		For each grid point tagged "interior", if it has at least one neighbour
c		in the solvent, it is considered on the surface
c
		   if(grid_partition(i,j,k).eq.1) then
c
c			nvol = nvol + 1
			vol = vol + dx*dy*dz
c
c		   Check points with i+1, and i-1, j+1, j-1, and k+1, k-1
c
			if(grid_partition(i+1,j,k).eq.0) 
     1				surf = surf + dy*dz
			if(grid_partition(i-1,j,k).eq.0)
     1				surf = surf + dy*dz
			if(grid_partition(i,j+1,k).eq.0)
     1				surf = surf + dx*dz
			if(grid_partition(i,j-1,k).eq.0)
     1				surf = surf + dx*dz
			if(grid_partition(i,j,k-1).eq.0)  
     1				surf = surf + dx*dy
			if(grid_partition(i,j,k+1).eq.0)
     1				surf = surf + dx*dy

		   endif
c
100		continue
200	   continue
300	continue
c
	if(surf_type.eq.0) then
c
		write(6,*) ' '
		write(6,1) surf 
		write(6,2) vol 
c
		write(l_log,*) ' '
		write(l_log,5) surf 
		write(l_log,6) vol 
		write(l_log,*) ' '
c
	else
c
		write(6,*) ' '
		write(6,3) surf
		write(6,4) vol
		write(6,*) ' '
		write(l_log,*) ' '
		write(l_log,7) surf 
		write(l_log,8) vol
		write(l_log,*) ' '
c
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	locate_cell.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine uses bisection to search an ordered array
c
	subroutine locate_cell(x,array,idx,n)
c
c ==========================================================================================
c = Define variables                          
c ==========================================================================================
c
	integer	i_low,i_high,i_mid
	integer	n,idx
c
	real*8	x
	real*8	array(n)
c
c ==========================================================================================
c = Perform bisection search                  
c ==========================================================================================
c
	i_low = 0
	i_high = n+1
c
100	if(i_high-i_low.gt.1) then
c
		i_mid = (i_high+i_low)/2
		if(x.ge.array(i_mid)) then
			i_low = i_mid
		else
			i_high = i_mid
		endif
c
		goto 100
c
	endif
c
	if(x.eq.array(1)) then
		idx = 1
	elseif(x.eq.array(n)) then
		idx = n-1
	else
		idx = i_low
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Defstep.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine finds the number of positions that need to be tested around a
c	given position to cover a given range
c
	subroutine defstep(val,array,ix,irange,n)
c
c ==========================================================================================
c = Define variables                          
c ==========================================================================================
c
	integer	ix,irange,i,n,iup,ilow
c
	real*8	val,array(n)
c
	iup = 0
	ilow = 0
c
c ==========================================================================================
c = Find upper range                          
c ==========================================================================================
c
	do 100 i = ix+1,n
		if(array(i)-array(ix).gt.val) then
			iup = i-ix
			goto 200
		endif
100	continue
200	continue
c
c ==========================================================================================
c = Find lower range                          
c ==========================================================================================
c
	do 300 i = ix-1,1,-1
		if(array(ix)-array(i).gt.val) then
			ilow = ix-i
			goto 400
		endif
300	continue
400	continue
c
	irange=max(ilow,iup)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Shifted_Grids.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine precomputes the 3 shifted grid projections of the molecule
c
	subroutine shifted_grids(nx,ny,nz,hx,hy,hz,grid_partition,
     1			grid_shift)
c
c ==========================================================================================
c = Define variables                          
c ==========================================================================================
c
	include 'poisson.h'
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	hx_save(*),hy_save(*),hz_save(*)
c
c	Information on grid
c ==========================================================================================
c
	integer	nx,ny,nz
	real*8	sizex,sizey,sizez
	real*8	hx(nx),hy(ny),hz(nz)
c
	integer*1 grid_partition(nx,ny,nz)
c
c	Information on shifted grids
c ==========================================================================================
c
	integer*1 grid_shift(nx,ny,nz)
c
c       Information on slab (if any)
c ==========================================================================================
c
        integer flag_slab
c
        real*8  z0,zh,eps_slab
c
c	Input parameters
c ==========================================================================================
c
        integer n_ion,n_zz
	integer	surf_type,flag_pymol,flag_box,flag_distrib,flag_eps
	integer	grid_type,bnd_type
	integer	ion_zone
c
	real*8	R_probe,R_ion
c
	real*8	ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8	kappa(max_ion)
c
c	Pointers for local arrays
c ==========================================================================================
c
        pointer (ptr_hx_save,hx_save)
        pointer (ptr_hy_save,hy_save)
        pointer (ptr_hz_save,hz_save)
c
c	Common blocks
c ==========================================================================================
c
	common /input_par/	R_probe,R_ion,flag_pymol,surf_type,
     1				grid_type,ion_zone,flag_box,
     2				flag_distrib,flag_eps,bnd_type
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
        common /slab/           z0,zh,eps_slab,flag_slab
c
c ==========================================================================================
c	Create space for local arrays
c ==========================================================================================
c
        ptr_hx_save = malloc(8*nx)
        ptr_hy_save = malloc(8*ny)
        ptr_hz_save = malloc(8*nz)
c
c ==========================================================================================
c	Initialise grid_shift
c ==========================================================================================
c
	do 300 k = 1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
				grid_shift(i,j,k) = 0
100			continue
200		continue
300	continue
c
c ==========================================================================================
c	Set information for non-shifted map
c ==========================================================================================
c
	do 600 k = 1,nz
		do 500 j = 1,ny
			do 400 i = 1,nx
				if(grid_partition(i,j,k).ne.1) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),0)
				endif
				if(n_ion.ne.0.and.
     1                          grid_partition(i,j,k).eq.0) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),1)
				endif
400			continue
500		continue
600	continue
c	
c ==========================================================================================
c	Define molecular interface using GenSurf, for the X-shifted map
c ==========================================================================================
c
	sizex = hx(nx)-hx(nx-1)
	do 700 i = 1,nx
		hx_save(i) = hx(i)
		if(i.lt.nx) then
			hx(i) = (hx(i)+hx(i+1))/2
		else
			hx(i) = hx(i) + sizex/2
		endif
700	continue
c
	call GenSurf(surf_type,R_probe,grid_partition,hx,hy,hz,nx,ny,nz)
c
	if(ion_zone.eq.1) call GenIonZone(R_ion,grid_partition,hx,hy,hz,
     1			nx,ny,nz)
        if(flag_slab.eq.1) 
     1          call IonSlab(z0,zh,grid_partition,hz,nx,ny,nz)
c
	do 800 i = 1,nx
		hx(i) = hx_save(i)
800	continue
c
c	Now store X-shifted map 
c
	do 1100 k = 1,nz
		do 1000 j = 1,ny
			do 900 i = 1,nx
				if(grid_partition(i,j,k).ne.1) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),2)
				endif
				if(n_ion.ne.0.and.
     1                          grid_partition(i,j,k).eq.0) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),3)
				endif
900			continue
1000		continue
1100	continue
c	
c ==========================================================================================
c	Define molecular interface using GenSurf, for the Y-shifted map
c ==========================================================================================
c
	sizey = hy(nx)-hy(nx-1)
	do 1200 j = 1,ny
		hy_save(j) = hy(j)
		if(j.lt.ny) then
			hy(j) = (hy(j)+hy(j+1))/2
		else
			hy(j) = hy(j) + sizey/2
		endif
1200	continue
c
	call GenSurf(surf_type,R_probe,grid_partition,hx,hy,hz,nx,ny,nz)
c
	if(ion_zone.eq.1) call GenIonZone(R_ion,grid_partition,
     1				hx,hy,hz,nx,ny,nz)
        if(flag_slab.eq.1) 
     1          call IonSlab(z0,zh,grid_partition,hz,nx,ny,nz)
c
	do 1300 j = 1,ny
		hy(j) = hy_save(j)
1300	continue
c
c	Now define Y-shifted map accordingly
c
	do 1600 k = 1,nz
		do 1500 j = 1,ny
			do 1400 i = 1,nx
				if(grid_partition(i,j,k).ne.1) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),4)
				endif
				if(n_ion.ne.0.and.
     1                          grid_partition(i,j,k).eq.0) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),5)
				endif
1400			continue
1500		continue
1600	continue
c	
c ==========================================================================================
c	Define molecular interface using GenSurf, for the Z-shifted map
c ==========================================================================================
c
	sizez = hz(nz) - hz(nz-1)
	do 1700 k = 1,nz
		hz_save(k) = hz(k)
		if(k.lt.nz) then
			hz(k) = (hz(k)+hz(k+1))/2
		else
			hz(k) = hz(k) + sizez/2
		endif
1700	continue
c
	call GenSurf(surf_type,R_probe,grid_partition,hx,hy,hz,nx,ny,nz)
c
	if(ion_zone.eq.1) call GenIonZone(R_ion,grid_partition,hx,hy,hz,
     1				nx,ny,nz)
        if(flag_slab.eq.1) 
     1          call IonSlab(z0,zh,grid_partition,hz,nx,ny,nz)
c
	do 1800 k = 1,nz
		hz(k) = hz_save(k)
1800	continue
c
c	Now define Z-shifted map accordingly
c
	do 2100 k = 1,nz
		do 2000 j = 1,ny
			do 1900 i = 1,nx
				if(grid_partition(i,j,k).ne.1) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),6)
				endif
				if(n_ion.ne.0.and.
     1                          grid_partition(i,j,k).eq.0) then
					grid_shift(i,j,k) = ibset(
     1					grid_shift(i,j,k),7)
				endif
1900			continue
2000		continue
2100	continue
c	
c ==========================================================================================
c	Clear local space
c ==========================================================================================
c
        call free(ptr_hx_save)
        call free(ptr_hy_save)
        call free(ptr_hz_save)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Restricted_Grids.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine restricts the 3 shifted grid projections of the molecule
c
	subroutine restricted_grids(nx,ny,nz,hx,hy,hz,
     1                  grid_partition,grid_shift)
c
c ==========================================================================================
c = Define variables                          
c ==========================================================================================
c
	include 'poisson.h'
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
c	Information on grid
c ==========================================================================================
c
	integer	nx,ny,nz
	real*8	hx(nx),hy(ny),hz(nz)
c
	integer*1 grid_partition(nx,ny,nz)
c
c	Information on shifted grids
c ==========================================================================================
c
	integer*1 grid_shift(nx,ny,nz)
c
	integer	surf_type,flag_pymol,flag_box,flag_distrib,flag_eps
	integer	grid_type,bnd_type
	integer	ion_zone
c
	real*8	R_probe,R_ion
c
c	Common blocks
c ==========================================================================================
c
	common /input_par/	R_probe,R_ion,flag_pymol,surf_type,
     1				grid_type,ion_zone,flag_box,
     2				flag_distrib,flag_eps,bnd_type
c
        call GenSurf(surf_type,R_probe,grid_partition,hx,hy,hz,
     1          nx,ny,nz)
c
        call shifted_grids(nx,ny,nz,hx,hy,hz,grid_partition,
     1          grid_shift)
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	Getidx.f
c
c	Version 1:	7/5/09
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	Convert 1D index into 3D index
c
c ==========================================================================================
c ==========================================================================================
c
        subroutine getidx(n,i,j,k,nx,ny,nz)
c
        integer n,m,i,j,k,nx,ny,nz
c
        k = n/(nx*ny) + 1
        if(mod(n,nx*ny).eq.0) k = k - 1
c
        m = n - (k-1)*nx*ny
c
        j = m/ny + 1
        if(mod(m,ny).eq.0) j = j - 1
c
        i = m - (j-1)*ny
c
        return
        end
