c ==========================================================================================
c ==========================================================================================
c	Write_Grid.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine generates a Pymol script for displaying the grid.
c	Edges of the grid are colored red if the two vertices are inside
c	the molecule, and blue outside
c
c	Input:
c
c		l_py		: logical unit number of Pymol file
c		itype:		: define output:
c					0:	only big box edges are shown for solvent
c					1:	whole mesh for solvent is shown
c		flag_surf	: display accessible surface (1), or not (0)
c		ishift		: shift grid, or not (for maps in XPLOR or EZD format)
c
c	Available through common blocks:
c
c		NpointSurf	: number of points on accessible surface
c		PointSurf	: coordinates of points on accessible surface
c
	subroutine Write_Grid(l_py,ion_zone,itype,flag_surf,ishift,
     1			grid_partition,hx,hy,hz,nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
	integer	i,j,k
	integer	l_py,itype,flag_surf
	integer ion_zone,zone_type
	integer	ishift
        integer nx,ny,nz
c
c	Information on grid
c
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Partition of the grid
c
	integer*1	grid_partition(nx,ny,nz)
c
c	Information on points on accessible surface area
c
	integer	NpointSurf
	real*8	PointSurf(*)
c
c	Variable to re-center grid on molecule
c
	real*8	Center_mol(3),Center_grid(3)
	real*8	shiftx,shifty,shiftz,shift(3)
c
c	local variables
c
	real*8	Totsizex,Totsizey,Totsizez
	real*8	Xcoord,Ycoord,Zcoord,Rad
c
        pointer (ptr_point,PointSurf)
c
	common /space/   ptr_point,NpointSurf
c
	common /centers/ Center_mol,Center_grid
	common /molshift/ shiftx,shifty,shiftz
c
1	format('from pymol.cgo import *')
2	format('from pymol import cmd',/)
3	format('Molecule = [',/,'   BEGIN, LINES,',/,
     1  '   COLOR, 1.0, 0.0, 0.0,',/,/)
4	format('Solvent = [',/,'   BEGIN, LINES,',/,
     1  '   COLOR, 0.0, 0.0, 1.0,',/,/)
5	format('   VERTEX,',f8.3,',',f8.3,',',f8.3,',')
6	format(' ')
7	format('   END',/,'   ]',/,/)
8	format('cmd.load_cgo(Molecule,''Molecule'')')
9	format('cmd.load_cgo(Solvent,''Solvent'')')
10	format('IonExcluded = [',/,'   BEGIN, LINES,',/,
     1  '   COLOR, 0.0, 1.0, 0.0,',/,/)
11	format('cmd.load_cgo(IonExcluded,''IonExcluded'')')
12	format('AccSurf = [',/,
     1  '   COLOR, 0.0, 1.0, 0.0,',/,/)
13	format('   SPHERE,',f8.3,',',f8.3,',',f8.3,',',f8.3,',')
14	format('cmd.load_cgo(AccSurf,''AccSurf'')')
c
	if(ishift.eq.1) then
		do 50 i = 1,3
			shift(i) = Center_mol(i) - Center_grid(i)
50		continue
	else
		do 75 i = 1,3
			shift(i) = 0.d0
75		continue
	endif
c ==========================================================================================
c = Write header for CGO pymol file ========================================================
c ==========================================================================================
c
	write(l_py,1)
	write(l_py,2)
c
c	We loop twice over all edges of the grid: first draw "red" ones, i.e.
c	those inside the molecule, and then draw "blue" ones, those inside
c	the solvent
c
c ==========================================================================================
c = Generate "red" object: molecule(s)  ====================================================
c ==========================================================================================
c
c	Define object as red
c
	write(l_py,3)
c
c	Generate object
c
	zone_type = 1
	call Write_CGO_object(zone_type,l_py,shift,grid_partition,
     1			hx,hy,hz,nx,ny,nz)
c
c	close object
c
	write(l_py,7)
c
c ==========================================================================================
c = Generate "green" object: ion-excluded region ===========================================
c ==========================================================================================
c
	if(ion_zone.eq.1) then
c
c		Define object as green
c
		write(l_py,10)
c
c		Generate object
c
		zone_type = 2
		call Write_CGO_object(zone_type,l_py,shift,
     1			grid_partition,hx,hy,hz,nx,ny,nz)
c
c		close object
c
		write(l_py,7)
c
	endif
c
c ==========================================================================================
c = Generate "blue" object: solvent      ====================================================
c ==========================================================================================
c
	write(l_py,4)
c
c	If itype = 1, generate whole mesh
c
	if(itype.eq.1) then
c
		zone_type = 0
		call Write_CGO_object(zone_type,l_py,shift,
     1			grid_partition,hx,hy,hz,nx,ny,nz)
c
	else
c
c	If itype = 0, only generate edges of full grid
c
c
                Totsizex = hx(nx)-hx(1)
                Totsizey = hy(ny)-hy(1)
                Totsizez = hz(nz)-hz(1)
c
		do 300 i = 1,2
c
		Xcoord = dble(2-i)*hx(1)+dble(i-1)*hx(nx)
     1			 +shift(1)
c
		   do 200 j = 1,2
c
			Ycoord=dble(2-j)*hy(1)+dble(j-1)*hy(ny)
     1			+shift(2)
c
			do 100 k = 1,2
c
			   Zcoord=dble(2-k)*hz(1)+dble(k-1)*hz(nz)
     1			  +shift(3)
c
			   if(i.ne.2) then
				write(l_py,5) Xcoord,Ycoord,Zcoord
				write(l_py,5) Xcoord+Totsizex,Ycoord,
     1				Zcoord
				write(l_py,6)
			   endif
c
			   if(j.ne.2) then
				write(l_py,5) Xcoord,Ycoord,Zcoord
				write(l_py,5) Xcoord,Ycoord+Totsizey,
     1				Zcoord
				write(l_py,6)
			   endif
c
			   if(k.ne.2) then
				write(l_py,5) Xcoord,Ycoord,Zcoord
				write(l_py,5) Xcoord,Ycoord,Zcoord
     1				+Totsizez
				write(l_py,6)
			endif
c
100		    continue
c
200		continue
c
300	    continue
c
	endif
c
	write(l_py,7)
c
c ==========================================================================================
c = Generate "Green" object: Accessible surface area  ======================================
c ==========================================================================================
c
	if(flag_surf.eq.1) then
c
		Rad = 0.05d0
c
		write(l_py,12)
c
		do 400 i = 1,NpointSurf
			Xcoord = PointSurf(3*(i-1)+1)+shift(1)
			Ycoord = PointSurf(3*(i-1)+2)+shift(2)
			Zcoord = PointSurf(3*(i-1)+3)+shift(3)
			write(l_py,13) Xcoord,Ycoord,Zcoord,Rad
400		continue
c
		write(l_py,7)
c
	endif
c
c ==========================================================================================
c = Tell pymol to load objects that we have created  ========================================
c ==========================================================================================
c
	write(l_py,8)
	if(ion_zone.eq.1) write(l_py,11)
	write(l_py,9)
	if(flag_surf.eq.1) write(l_py,14)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Write_CGO_object.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/07/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine generates a script for Pymol to display part of the mesh
c	used for solving the Poisson Boltzmann Equation.
c
c	We proceed in 3 steps: lines parallel to Z axis, Y axis, and X axis,
c				respectively.
c
c	For each of these lines, define fraction of interest, and write corresponding
c	segment in pymol file
c
	subroutine Write_CGO_object(zone_type,l_py,shift,
     1			grid_partition,hx,hy,hz,nx,ny,nz)
c
c ==========================================================================================
c	Define all variables
c ==========================================================================================
c
	integer	i,j,k
	integer	zone_type,l_py
	integer	istart,iend
        integer nx,ny,nz
c
c	Information on grid
c
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Partition of the grid
c
	integer*1	grid_partition(nx,ny,nz)
c
	real*8	shift(3)
c
c	Local variables
c
	real*8	Xcoord,Ycoord,Zcoord
	real*8	Xcoord_start,Xcoord_end
	real*8	Ycoord_start,Ycoord_end
	real*8	Zcoord_start,Zcoord_end
c
	integer*1	info1
c
5	format('   VERTEX,',f8.3,',',f8.3,',',f8.3,',')
6	format(' ')
c
c ==========================================================================================
c	Generate Pymol object      
c ==========================================================================================
c
c	Start with lines parallel to Z axis
c
	do 300 i = 1,nx
c
	    Xcoord = hx(i) + shift(1)
c
	    do 200 j = 1,ny
c
		Ycoord=hy(j)+shift(2)
c
		istart = 0
		Zcoord_start = 0
		do 100 k = 1,nz
c
			Zcoord=hz(k)+shift(3)
c
			info1 = grid_partition(i,j,k)
			if(info1.ne.zone_type) then
				if(istart.eq.0) goto 100
				iend = k-1
				Zcoord_end=hz(iend)+shift(3)
				write(l_py,5)Xcoord,Ycoord,Zcoord_start
				write(l_py,5)Xcoord,Ycoord,Zcoord_end
				write(l_py,6)
				istart=0
			else
				if(istart.eq.0) then
					istart = 1
					Zcoord_start = Zcoord
				else
					if(k.eq.nz) then
						Zcoord_end = Zcoord
						write(l_py,5)Xcoord,
     1						Ycoord,Zcoord_start
						write(l_py,5)Xcoord,
     2						Ycoord,Zcoord_end
						write(l_py,6)
					endif
				endif
			endif
100		continue
c
200	    continue
c
300	continue
c
c	Now do lines parallel to Y axis
c
	do 600 i = 1,nx
c
	    Xcoord = hx(i) + shift(1)
c
	    do 500 k = 1,nz
c
		Zcoord=hz(k)+shift(3)
c
		istart = 0
		Ycoord_start = 0
		do 400 j = 1,ny
c
			Ycoord=hy(j)+shift(2)
c
			info1 = grid_partition(i,j,k)
			if(info1.ne.zone_type) then
				if(istart.eq.0) goto 400
				iend = j-1
				Ycoord_end=hy(iend)+shift(2)
				write(l_py,5)Xcoord,Ycoord_start,Zcoord
				write(l_py,5)Xcoord,Ycoord_end,Zcoord
				write(l_py,6)
				istart=0
			else
				if(istart.eq.0) then
					istart = 1
					Ycoord_start = Ycoord
				else
					if(j.eq.ny) then
						Ycoord_end = Ycoord
						write(l_py,5)Xcoord,
     1						Ycoord_start,Zcoord
						write(l_py,5)Xcoord,
     2						Ycoord_end,Zcoord
						write(l_py,6)
					endif
				endif
			endif
400		continue
c
500	    continue
c
600	continue
c
c	Finish with lines parallel to X axis
c
	do 900 j = 1,ny
c
	    Ycoord = hy(j) + shift(2)
c
	    do 800 k = 1,nz
c
		Zcoord=hz(k)+shift(3)
c
		istart = 0
		Xcoord_start = 0
		do 700 i = 1,nx
c
			Xcoord=hx(i)+shift(1)
c
			info1 = grid_partition(i,j,k)
			if(info1.ne.zone_type) then
				if(istart.eq.0) goto 700
				iend = i-1
				Xcoord_end=hx(iend)+shift(1)
				write(l_py,5)Xcoord_start,Ycoord,Zcoord
				write(l_py,5)Xcoord_end,Ycoord,Zcoord
				write(l_py,6)
				istart=0
			else
				if(istart.eq.0) then
					istart = 1
					Xcoord_start = Xcoord
				else
					if(i.eq.nx) then
						Xcoord_end = Xcoord
						write(l_py,5)
     1						Xcoord_start,Ycoord,
     1						Zcoord
						write(l_py,5)
     1						Xcoord_end,Ycoord,
     1						Zcoord
					endif
				endif
			endif
700		continue
c
800	    continue
c
900	continue
c
	return
	end
