c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	GenAccSurf.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/29/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	Generates a set of point on the accessible surface
c			area of a molecule. 
c			This points will then be used to generate the grid-based
c			molecular surface of the molecule
c			(see GenSurf)
c
c	Input:
c		R_probe		: radius of water probe
c
c	Output (in common block AccSurf):
c		Npoint		: number of points generated on accessible surface of molecule
c		PointAccSurf	: coordinates of these points
c
c
c	The procedure follows the scheme proposed
c	by Shrake and Rupley (J. Mol. Biol., 79, 351-371, 1973)
c	but performs the calculation based on LeGrand and Merz algorithm
c	(J. Comp. Chem. 14, 349-352, 1993)
c
	subroutine GenAccSurf(R_probe)
c
	include 'poisson.h'
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	Npoint
	integer	ibuild
c
	real*8	R_probe
c
	real*8	PointAccSurf(*)
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Pointers                 
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        pointer (ptr_point,PointAccSurf)
c
c	Common blocks            
c ==========================================================================================
c
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,
     2                          natom,ncharge
c
	common /space/		ptr_point,Npoint
c
	save ibuild
c
	data ibuild /0/
c
c ==========================================================================================
c = Setup: 1. build lookup tables for Legrand-Merz approach
c ==========================================================================================
c
	if(ibuild.eq.0) then
		call sphere
		call buildlookup
		ibuild = 1
	endif
c
c ==========================================================================================
c = Setup: 2. Project molecule on coarse grid              
c ==========================================================================================
c
c	Project grid on the grid (size of the grid is defined at 5 angstrom)
c
	call defgrid(atom_coord,natom)
c
c ==========================================================================================
c = Build point list
c ==========================================================================================
c
	Npoint = 0
c
	call AccPointList(R_probe)
c
c ==========================================================================================
c = Return to calling program
c ==========================================================================================
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	AccPointList.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/29/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	define accessible points on inflated surface of each atom
c
c	Input of the program :
c
c		R_probe		: radius of water probe
c
c	Output:
c		Npoint		: number of points generated on accessible surface of molecule
c		PointAccSurf	: coordinates of these points
c
	subroutine AccPointList(R_probe)
c
	include 'poisson.h'
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	Npoint,Npoint_max
	integer	i,j,k,l,m,ia,ja,ka,icell
c
	real*8	dist,distsq,rad2,rad,val,valsq,v2,v2sq
	real*8	r1,r2,diffx,diffy,diffz
	real*8	pi,pid
c
	real*8	at1(3),at2(3)
c
c	Variables for coarse grid
c
	integer	ncellx,ncelly,ncellz,nplane
	integer	ncell,listcell(125)
	integer	natincell(*),listatincell(natotc,*)
c
	real*8	cube,xyzmin(3)
c
c	Variable for loopkup tables
c
	integer	npoint_s,ninteg
	integer	iphi,itheta,idiq,kint(512),lbit(512),icenter
	integer	maskdist(512,200,16),closestp(200,200)
	integer	istore(16),istat(16)
	real*8	spheres(3,512)
c
	real*8	xdist,xtheta,xphi,phi,theta,cosine,diq
c
	real*8	R_probe
c
	real*8	PointAccSurf(*)
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Pointers                 
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        pointer (ptr_natincell,natincell)
        pointer (ptr_listatincell,listatincell)
c
        pointer (ptr_point,PointAccSurf)
c
c	Common blocks            
c ==========================================================================================
c
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,
     2                          natom,ncharge
	common /grid/           cube,xyzmin,ncellx,ncelly,ncellz
        common /grid2/          ptr_natincell,ptr_listatincell
	common /lookup/		xdist,xtheta,xphi,kint,lbit,maskdist,
     1				closestp
	common /spherepos/	spheres
	common /space/		ptr_point,Npoint
c
c
	pi = dacos(-1.d0)
	pid = 180.d0/pi
c
	npoint_s = 512
	ninteg   = npoint_s/32
c
	nplane    = ncellx*ncelly
c
	do 100 i = 1,ninteg
		istat(i) = 0
100	continue
c
        Npoint_max = 100000
        ptr_point = malloc(24*Npoint_max)
c
c ==========================================================================================
c =    Generate list of surface points             
c ==========================================================================================
c
c
c	Loop over all atoms of the protein ...
c
	do 1000 i = 1,natom
c
c		Initialise status of i
c
		do 300 j = 1,ninteg
			do 200 k = 1,32
				istat(j)=ibset(istat(j),k-1)
200			continue
300		continue
c
		r1 = atom_radius(i)
c
		val = r1+R_probe
		valsq = val*val
c
c		Define center of the atom
c
		at1(1) = atom_coord(3*(i-1)+1)
		at1(2) = atom_coord(3*(i-1)+2)
		at1(3) = atom_coord(3*(i-1)+3)
c
c		Define which cells in the lattice must be checked
c
		ia = nint((at1(1)-xyzmin(1))/cube) + 1
		ja = nint((at1(2)-xyzmin(2))/cube) + 1
		ka = nint((at1(3)-xyzmin(3))/cube) + 1
		icell = (ka-1)*nplane + (ia-1)*ncelly + ja
c
	   	call neighbour(icell,ncellx,ncelly,ncellz,
     1	    	ncell,listcell)
c
		do 700 k = 1,ncell
c
	    	   do 600 l = 1,natincell(listcell(k))
c
			j=listatincell(l,listcell(k))
			if(j.eq.i) goto 600
c
			at2(1) = atom_coord(3*(j-1)+1)
			at2(2) = atom_coord(3*(j-1)+2)
			at2(3) = atom_coord(3*(j-1)+3)
c
			r2 = atom_radius(j)
			v2 = r2+R_probe
			v2sq = v2*v2
c
c			find distance between atom i and test atom j
c
			distsq = 0
			do 400 m = 1,3
				distsq = distsq + (at1(m)-at2(m))**2
400			continue
c
c			Check distance
c
			rad = r1+r2+2*R_probe
			rad2 = rad*rad
c
			if(distsq.lt.rad2) then
c
				dist = sqrt(distsq)
c
                                if(dist+v2.le.val) goto 600
                                if(dist+val.le.v2) then
                                        do 450 m=1,ninteg
                                                istat(m) = 0
450                                     continue
                                        goto 750
                                endif
c
c				Find point of intersection on sphere
c				for atom 1 
c
				diffx = at2(1) - at1(1)
				diffy = at2(2) - at1(2)
				diffz = at2(3) - at1(3)
c
				phi = dacos(diffz/dist)*pid
				theta = datan2(diffy,diffx)*pid
				if(theta.lt.0) theta = theta + 360d0
				iphi = int(phi/xphi) + 1
				if(iphi.lt.1) iphi = 1
				if(iphi.gt.200) iphi = 200
				itheta = int(theta/xtheta)
				if(itheta.lt.1) itheta = 1
				if(itheta.gt.200) itheta = 200
c
c				Based on lookup tables, find which
c				points on the surface of i are buried
c				by test atoms, amd update istat
c				correspondingly
c
				icenter = closestp(iphi,itheta)
c
				cosine=(valsq+distsq-v2sq)/(2*val*dist)
				diq = sqrt(2.d0-2.d0*cosine)
				idiq = int(diq/xdist) + 1
				if(idiq.lt.1) idiq = 1
				if(idiq.gt.200) idiq = 200
c
				do 500 m = 1,ninteg
				    istore(m) = maskdist(icenter,
     1					idiq,m)
				    istat(m)=iand(istat(m),istore(m))
500				continue
c
			endif
c
c			End loops over neighbours of i
c
600		    continue
c
700	 	continue
c
c		Now check all npoint_s points on the surface of i:
c		if it is accessible, add to list of "accessible" points
c		(for coordinates, take optimized coordinates of point
c		on unit sphere (in array "spheres"), scale to radius
c		R_atom+R_Probe, and translate:
c
750             continue
c
		do 900 m = 1,npoint_s
			if(btest(istat(kint(m)),lbit(m))) then
                                if(Npoint.gt.int(0.95*Npoint_max)) then
                                        call resize_point(Npoint_max)
                                endif
				Npoint = Npoint + 1
				do 800 k = 1,3
					PointAccSurf(3*(Npoint-1)+k)=
     1                                  at1(k)+	val*spheres(k,m)
800				continue
			endif
900		continue
c
c	... End loop over all atoms of the molecule
c
1000	continue
c
c	End subroutine
c
        call free(ptr_natincell)
        call free(ptr_listatincell)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Defgrid.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/29/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose: this subroutine divides the volume containing the molecule
c	in cubic cells to form a lattice.
c
c	Input:
c		atom:	coordinates of all atoms of the molecule
c		natom:  number of atoms
c
c	Output: (in common block grid)
c		natincell contains the number of atoms for each cell
c		listatincell contains the list of atoms for each cell
c		cell	  gives the cell number for each atom
c		xyzmax    is the upper point of the lattice
c		xyzmin    is the lower point of the lattice
c		ncellx    is the number of cells in the x dimension
c		ncelly    is the number of cells in the y dimension
c		ncellz    is the number of cells in the z dimension
c
	subroutine defgrid(atom,natom)
c
	include 'poisson.h'
c
c ==========================================================================================
c =    Setup: declare variables, set common blocks
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer natom,ncellx,ncelly,ncellz,nplane,ncelltot
	integer natincell(*),listatincell(natotc,*)
	integer cell(*)
	integer i,j,ia,ja,ka,icell
c
	real*8	atom(3*natom),cube,xyzmin(3),xyzmax(3)
c
        pointer (ptr_cell,cell)
        pointer (ptr_natincell,natincell)
        pointer (ptr_listatincell,listatincell)
c
c	common blocks
c ==========================================================================================
c
	common /grid/ cube,xyzmin,ncellx,ncelly,ncellz
        common /grid2/ ptr_natincell,ptr_listatincell
c
        ptr_cell = malloc(4*natom)
c
c ==========================================================================================
c =    Generate grid, and pointers
c ==========================================================================================
c
c	1. first position lattice on the molecule : find position of the
c	upper and lower point on the lattice
c
	cube = 3.d0
c
	do 100 i = 1,3
		xyzmin(i) = atom(i)
		xyzmax(i) = atom(i)
100	continue
c
	do 300 i = 2,natom
c
		do 200 j = 1,3
			if(atom(3*(i-1)+j).gt.xyzmax(j)) 
     1			xyzmax(j) = atom(3*(i-1)+j)
			if(atom(3*(i-1)+j).lt.xyzmin(j)) 
     1			xyzmin(j) = atom(3*(i-1)+j)
200		continue
c
300	continue
c
c	Expand size of the lattice on each side by 5 angstrom
c
	do 400 i = 1,3
		xyzmin(i) = xyzmin(i) - 5.d0
		xyzmax(i) = xyzmax(i) + 5.d0
400	continue
c
c	Define number of cells in each dimension, and total number
c	of cells
c
	ncellx = nint((xyzmax(1)-xyzmin(1))/cube) + 1
	ncelly = nint((xyzmax(2)-xyzmin(2))/cube) + 1
	ncellz = nint((xyzmax(3)-xyzmin(3))/cube) + 1
c
	nplane = ncellx*ncelly
	ncelltot = nplane*ncellz
c
        ptr_natincell = malloc(4*ncelltot)
        ptr_listatincell = malloc(4*natotc*ncelltot)
c
c	Give a cell number to each atom
c
	do 500 i = 1,natom
		ia = nint((atom(3*(i-1)+1) -xyzmin(1))/cube)+1
		ja = nint((atom(3*(i-1)+2) -xyzmin(2))/cube)+1
		ka = nint((atom(3*(i-1)+3) -xyzmin(3))/cube)+1
		icell = (ka-1)*nplane + (ia-1)*ncelly + ja
		cell(i) = icell
500	continue
c
c	Now, for each cell, give the list of atoms included
c
	do 600 i = 1,ncelltot
		natincell(i) = 0
600	continue
c
	do 700 i = 1,natom
		natincell(cell(i)) = natincell(cell(i)) + 1
		if(natincell(cell(i)).gt.natotc) then
			write(6,*) 'too many atoms in a cell : ',
     1			natincell(cell(i))
			stop
		endif
		listatincell(natincell(cell(i)),cell(i)) = i
700	continue
c
        call free(ptr_cell)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Neighbour.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/29/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose: This subroutine finds the neighbours of a cell in a cube lattice
c	(two layers are considered : hence 125 neighbours are found)
c
c	Input	:
c			- ilayer	: number of layers
c			- icell		: cell number
c			- ncellx	: number of cells along x
c			- ncelly	: number of cells along y
c			- ncellz	: number of cells along z
c	Output 	:
c			- n		: number of neighbours 
c					  (including icell)
c			- list		: list of neighbours
c
	subroutine neighbour(icell,ncellx,ncelly,ncellz,n,list)
c
	integer	icell,ncellx,ncelly,ncellz,n,list(125)
	integer	nplane,ip,ix,iy,iz
	integer	ilayer,il,ik,ij,ii
	integer	i,j,k
c
	nplane = ncellx*ncelly
	iz = icell/nplane + 1
	if(mod(icell,nplane).eq.0) iz = iz -1
c
	ip = icell - (iz-1)*nplane
	ix = ip/ncelly + 1
	if(mod(ip,ncelly).eq.0) ix = ix -1
c
	iy = ip - (ix-1)*ncelly
c
	ilayer = 2
c
	n = 0
	do 400 il = 0,ilayer
c
		do 300 k = -il,il
			ik = iz + k
			if(ik.le.0) goto 300
			if(ik.gt.ncellz) goto 300
			do 200 i = -il,il
				ii = ix + i
				if(ii.le.0) goto 200
				if(ii.gt.ncellx) goto 200
				if(abs(k).lt.il.and.abs(i).lt.il) then
				    do 100 j = -il,il,2*il
					ij = iy + j
					if(ij.le.0) goto 100
					if(ij.gt.ncelly) goto 100
					n = n + 1
					list(n) = (ik-1)*nplane
     1					+ (ii-1)*ncelly + ij
100				    continue
				else
				    do 150 j = -il,il
					ij = iy + j
					if(ij.le.0) goto 150
					if(ij.gt.ncelly) goto 150
					n = n + 1
					list(n) = (ik-1)*nplane
     1					+ (ii-1)*ncelly + ij
150				    continue
				endif
200			continue
300		continue
c
400	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Sphere.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/29/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose: This small subroutine contains the position of 256 points
c	uniformly distributed on a sphere of radius 1
c
	subroutine sphere
c
	integer	j
c
	real*8	spheres(3,512)
c
	common /spherepos/	spheres
c
	data (spheres(j,1),j=1,3)/ 0.6270259972d0,-0.5837936194d0,
     1    -0.5157745716d0/
	data (spheres(j,2),j=1,3)/ 0.1100389600d0,-0.8929027386d0,
     1     0.4365960681d0/
	data (spheres(j,3),j=1,3)/-0.6772602478d0, 0.7356881658d0,
     1     0.0090265952d0/
	data (spheres(j,4),j=1,3)/ 0.2582718421d0,-0.4792639318d0,
     1     0.8388097158d0/
	data (spheres(j,5),j=1,3)/-0.3081869540d0, 0.4985178501d0,
     1     0.8102473415d0/
	data (spheres(j,6),j=1,3)/-0.5067142058d0, 0.0231306623d0,
     1    -0.8618037399d0/
	data (spheres(j,7),j=1,3)/ 0.4765499128d0,-0.7957335196d0,
     1    -0.3737757970d0/
	data (spheres(j,8),j=1,3)/-0.2460641674d0,-0.1808738155d0,
     1     0.9522274352d0/
	data (spheres(j,9),j=1,3)/ 0.1799240476d0, 0.2363631390d0,
     1     0.9548611436d0/
	data (spheres(j,10),j=1,3)/ 0.6684619631d0,-0.7074175735d0,
     1     0.2296061423d0/
	data (spheres(j,11),j=1,3)/ 0.7071989891d0,-0.6156520000d0,
     1     0.3476236539d0/
	data (spheres(j,12),j=1,3)/-0.8708142906d0,-0.4658948597d0,
     1    -0.1569217989d0/
	data (spheres(j,13),j=1,3)/ 0.9777143678d0,-0.1097227574d0,
     1     0.1789847243d0/
	data (spheres(j,14),j=1,3)/-0.9255301699d0, 0.2784283094d0,
     1    -0.2566545951d0/
	data (spheres(j,15),j=1,3)/ 0.8091992445d0,-0.4660965117d0,
     1    -0.3577018652d0/
	data (spheres(j,16),j=1,3)/-0.8690705024d0,-0.4942631002d0,
     1     0.0205048674d0/
	data (spheres(j,17),j=1,3)/-0.3059698923d0, 0.9379647493d0,
     1     0.1631090250d0/
	data (spheres(j,18),j=1,3)/-0.6127106065d0, 0.7752271278d0,
     1     0.1536509454d0/
	data (spheres(j,19),j=1,3)/-0.5610839295d0, 0.3337202690d0,
     1     0.7575061756d0/
	data (spheres(j,20),j=1,3)/-0.5710843651d0,-0.7811653151d0,
     1    -0.2522764323d0/
	data (spheres(j,21),j=1,3)/-0.2175972771d0,-0.8318270193d0,
     1     0.5106028153d0/
	data (spheres(j,22),j=1,3)/-0.7076285842d0,-0.7059568666d0,
     1    -0.0297773286d0/
	data (spheres(j,23),j=1,3)/ 0.8975833588d0, 0.2152117434d0,
     1    -0.3847440960d0/
	data (spheres(j,24),j=1,3)/ 0.8942688919d0,-0.0381350459d0,
     1     0.4459023069d0/
	data (spheres(j,25),j=1,3)/ 0.0269663361d0, 0.0044135330d0,
     1    -0.9996265990d0/
	data (spheres(j,26),j=1,3)/-0.3185591790d0,-0.3155848446d0,
     1     0.8938267479d0/
	data (spheres(j,27),j=1,3)/ 0.1311190447d0,-0.9498552629d0,
     1     0.2838710545d0/
	data (spheres(j,28),j=1,3)/ 0.6499503340d0, 0.1541424167d0,
     1     0.7441805418d0/
	data (spheres(j,29),j=1,3)/ 0.3272012271d0,-0.9428065081d0,
     1     0.0636808077d0/
	data (spheres(j,30),j=1,3)/ 0.5843417333d0,-0.7752956589d0,
     1    -0.2397110342d0/
	data (spheres(j,31),j=1,3)/-0.7574852945d0, 0.4273776772d0,
     1    -0.4935223903d0/
	data (spheres(j,32),j=1,3)/-0.0343245351d0, 0.9539590543d0,
     1     0.2979663555d0/
	data (spheres(j,33),j=1,3)/-0.8624432152d0, 0.1653126047d0,
     1     0.4783967426d0/
	data (spheres(j,34),j=1,3)/ 0.2797626885d0,-0.9299483622d0,
     1     0.2385977405d0/
	data (spheres(j,35),j=1,3)/ 0.1327566994d0, 0.7766326175d0,
     1    -0.6158063301d0/
	data (spheres(j,36),j=1,3)/-0.1178436526d0, 0.4865416202d0,
     1    -0.8656732209d0/
	data (spheres(j,37),j=1,3)/ 0.3916174801d0, 0.8366887494d0,
     1     0.3828677134d0/
	data (spheres(j,38),j=1,3)/ 0.2996908278d0, 0.9172450299d0,
     1     0.2623870479d0/
	data (spheres(j,39),j=1,3)/-0.6917756901d0, 0.6562278740d0,
     1    -0.3013492525d0/
	data (spheres(j,40),j=1,3)/ 0.2749747679d0, 0.3243376765d0,
     1    -0.9050933370d0/
	data (spheres(j,41),j=1,3)/ 0.1080784959d0, 0.9941013925d0,
     1     0.0090255227d0/
	data (spheres(j,42),j=1,3)/ 0.9215594013d0, 0.3277887894d0,
     1     0.2080451379d0/
	data (spheres(j,43),j=1,3)/-0.4495332123d0, 0.3525576237d0,
     1    -0.8207454009d0/
	data (spheres(j,44),j=1,3)/-0.6468321636d0, 0.1945654209d0,
     1     0.7373957209d0/
	data (spheres(j,45),j=1,3)/ 0.7913350774d0, 0.6091753362d0,
     1     0.0519057325d0/
	data (spheres(j,46),j=1,3)/-0.4979969763d0, 0.5719169350d0,
     1     0.6518512338d0/
	data (spheres(j,47),j=1,3)/ 0.7052653262d0,-0.6718783410d0,
     1    -0.2262306667d0/
	data (spheres(j,48),j=1,3)/ 0.1039315072d0,-0.4752304585d0,
     1     0.8737014668d0/
	data (spheres(j,49),j=1,3)/-0.6774173504d0,-0.2499272633d0,
     1     0.6918396465d0/
	data (spheres(j,50),j=1,3)/-0.3414950297d0, 0.7549457604d0,
     1     0.5598553774d0/
	data (spheres(j,51),j=1,3)/-0.8188836157d0,-0.2168572411d0,
     1    -0.5314156198d0/
	data (spheres(j,52),j=1,3)/ 0.7362529752d0, 0.5625139569d0,
     1     0.3761776239d0/
	data (spheres(j,53),j=1,3)/-0.9216086317d0, 0.3816967480d0,
     1     0.0703215653d0/
	data (spheres(j,54),j=1,3)/-0.2807663248d0, 0.4229081799d0,
     1    -0.8615793302d0/
	data (spheres(j,55),j=1,3)/-0.2118020547d0, 0.9275391687d0,
     1     0.3079139167d0/
	data (spheres(j,56),j=1,3)/ 0.7043311548d0,-0.1381956244d0,
     1     0.6962898777d0/
	data (spheres(j,57),j=1,3)/-0.7799129937d0,-0.6202375927d0,
     1     0.0839109698d0/
	data (spheres(j,58),j=1,3)/ 0.9552906994d0,-0.0821546808d0,
     1    -0.2840251539d0/
	data (spheres(j,59),j=1,3)/ 0.0439074909d0, 0.5423357494d0,
     1    -0.8390137467d0/
	data (spheres(j,60),j=1,3)/ 0.8185896364d0,-0.1730105813d0,
     1    -0.5477027898d0/
	data (spheres(j,61),j=1,3)/-0.3743832056d0, 0.8772632454d0,
     1     0.3004104085d0/
	data (spheres(j,62),j=1,3)/ 0.4217797451d0,-0.8878804404d0,
     1     0.1837666185d0/
	data (spheres(j,63),j=1,3)/-0.7107700190d0, 0.0425760210d0,
     1     0.7021347894d0/
	data (spheres(j,64),j=1,3)/ 0.5922491683d0,-0.5454933415d0,
     1     0.5930243984d0/
	data (spheres(j,65),j=1,3)/ 0.2403170180d0,-0.3446957112d0,
     1    -0.9074318694d0/
	data (spheres(j,66),j=1,3)/ 0.0754194027d0,-0.8140659643d0,
     1     0.5758545993d0/
	data (spheres(j,67),j=1,3)/-0.8542113404d0,-0.4145573775d0,
     1     0.3137915976d0/
	data (spheres(j,68),j=1,3)/-0.7808742194d0,-0.2685142397d0,
     1     0.5640350668d0/
	data (spheres(j,69),j=1,3)/-0.9695072818d0,-0.1515651810d0,
     1     0.1925710947d0/
	data (spheres(j,70),j=1,3)/ 0.0517879764d0,-0.8951033752d0,
     1    -0.4428407764d0/
	data (spheres(j,71),j=1,3)/ 0.6330745830d0,-0.4455487906d0,
     1    -0.6330109380d0/
	data (spheres(j,72),j=1,3)/ 0.6457776313d0, 0.5566705952d0,
     1    -0.5225792757d0/
	data (spheres(j,73),j=1,3)/ 0.3694411821d0, 0.2035199907d0,
     1    -0.9066933475d0/
	data (spheres(j,74),j=1,3)/ 0.8823381658d0,-0.3256917779d0,
     1    -0.3397119766d0/
	data (spheres(j,75),j=1,3)/ 0.6517316696d0, 0.6561335999d0,
     1    -0.3804399163d0/
	data (spheres(j,76),j=1,3)/-0.0643410066d0,-0.1764476485d0,
     1     0.9822048983d0/
	data (spheres(j,77),j=1,3)/-0.0050400685d0, 0.9937918723d0,
     1    -0.1111409563d0/
	data (spheres(j,78),j=1,3)/-0.9720122773d0,-0.2083896473d0,
     1    -0.1084706761d0/
	data (spheres(j,79),j=1,3)/ 0.6190928435d0, 0.4456178793d0,
     1    -0.6466442274d0/
	data (spheres(j,80),j=1,3)/ 0.2118044349d0,-0.8534313183d0,
     1    -0.4762287961d0/
	data (spheres(j,81),j=1,3)/ 0.3050062457d0,-0.6023839178d0,
     1     0.7376346018d0/
	data (spheres(j,82),j=1,3)/ 0.5484947244d0, 0.7681346051d0,
     1    -0.3303373513d0/
	data (spheres(j,83),j=1,3)/ 0.1797496560d0,-0.0451211956d0,
     1     0.9826770267d0/
	data (spheres(j,84),j=1,3)/-0.0840946288d0,-0.2417447630d0,
     1    -0.9666889691d0/
	data (spheres(j,85),j=1,3)/ 0.9878899441d0, 0.0531462518d0,
     1     0.1457701419d0/
	data (spheres(j,86),j=1,3)/-0.1386017077d0, 0.4935464369d0,
     1     0.8586043799d0/
	data (spheres(j,87),j=1,3)/-0.2436059175d0, 0.0151113562d0,
     1    -0.9697565694d0/
	data (spheres(j,88),j=1,3)/-0.5516849185d0, 0.4226251706d0,
     1    -0.7190491749d0/
	data (spheres(j,89),j=1,3)/ 0.4714365153d0,-0.1049495598d0,
     1    -0.8756330293d0/
	data (spheres(j,90),j=1,3)/-0.3489082477d0,-0.3839930917d0,
     1    -0.8548756285d0/
	data (spheres(j,91),j=1,3)/-0.3590946931d0, 0.6177979042d0,
     1     0.6995546805d0/
	data (spheres(j,92),j=1,3)/-0.9315624726d0, 0.1552425613d0,
     1     0.3287721198d0/
	data (spheres(j,93),j=1,3)/ 0.3987864262d0, 0.1130474260d0,
     1     0.9100492656d0/
	data (spheres(j,94),j=1,3)/ 0.0834929188d0, 0.3738800648d0,
     1     0.9237113346d0/
	data (spheres(j,95),j=1,3)/-0.1560302936d0,-0.7489930977d0,
     1    -0.6439440093d0/
	data (spheres(j,96),j=1,3)/ 0.9027309674d0, 0.1309485302d0,
     1     0.4097917556d0/
	data (spheres(j,97),j=1,3)/-0.2100671846d0,-0.4323597192d0,
     1     0.8768904442d0/
	data (spheres(j,98),j=1,3)/ 0.6375375653d0,-0.2911155461d0,
     1     0.7133004919d0/
	data (spheres(j,99),j=1,3)/-0.6977341113d0, 0.3191469534d0,
     1     0.6413363642d0/
	data (spheres(j,100),j=1,3)/ 0.5454850981d0, 0.6061929838d0,
     1     0.5787711760d0/
	data (spheres(j,101),j=1,3)/-0.0119212515d0,-0.5955714234d0,
     1     0.8032138964d0/
	data (spheres(j,102),j=1,3)/ 0.2538171531d0, 0.6918806971d0,
     1    -0.6759274767d0/
	data (spheres(j,103),j=1,3)/ 0.8332396814d0, 0.0831826027d0,
     1     0.5466189605d0/
	data (spheres(j,104),j=1,3)/ 0.7370386148d0, 0.4168195550d0,
     1    -0.5320108447d0/
	data (spheres(j,105),j=1,3)/ 0.2108909253d0,-0.1862682404d0,
     1    -0.9595984370d0/
	data (spheres(j,106),j=1,3)/-0.1418442663d0, 0.9748922582d0,
     1     0.1716545633d0/
	data (spheres(j,107),j=1,3)/-0.6194476690d0, 0.1059216051d0,
     1    -0.7778593697d0/
	data (spheres(j,108),j=1,3)/-0.0978248849d0, 0.0995765499d0,
     1    -0.9902094741d0/
	data (spheres(j,109),j=1,3)/ 0.1680270292d0, 0.9746239639d0,
     1    -0.1479021513d0/
	data (spheres(j,110),j=1,3)/-0.5820465134d0,-0.5296458209d0,
     1     0.6170066131d0/
	data (spheres(j,111),j=1,3)/ 0.5186533139d0,-0.4187671935d0,
     1    -0.7454077929d0/
	data (spheres(j,112),j=1,3)/-0.8512369220d0,-0.1303468706d0,
     1     0.5083359085d0/
	data (spheres(j,113),j=1,3)/ 0.1177249462d0,-0.9769298475d0,
     1    -0.1781822384d0/
	data (spheres(j,114),j=1,3)/ 0.8260926070d0, 0.2534853477d0,
     1     0.5033052584d0/
	data (spheres(j,115),j=1,3)/-0.8784380253d0, 0.4256015689d0,
     1    -0.2172784856d0/
	data (spheres(j,116),j=1,3)/-0.8775908866d0, 0.4759155889d0,
     1    -0.0577805161d0/
	data (spheres(j,117),j=1,3)/ 0.6107273257d0,-0.6438254274d0,
     1     0.4609782562d0/
	data (spheres(j,118),j=1,3)/-0.3906289541d0, 0.9204030408d0,
     1     0.0163481706d0/
	data (spheres(j,119),j=1,3)/-0.8858571826d0, 0.2068225810d0,
     1    -0.4153088875d0/
	data (spheres(j,120),j=1,3)/ 0.3699309971d0,-0.7855097349d0,
     1    -0.4961103847d0/
	data (spheres(j,121),j=1,3)/ 0.6289131132d0,-0.0216924996d0,
     1     0.7771729096d0/
	data (spheres(j,122),j=1,3)/-0.4892490428d0, 0.5630690420d0,
     1    -0.6660244951d0/
	data (spheres(j,123),j=1,3)/-0.6823386612d0,-0.5389111240d0,
     1     0.4939521757d0/
	data (spheres(j,124),j=1,3)/ 0.4925248658d0,-0.8691912596d0,
     1     0.0438840606d0/
	data (spheres(j,125),j=1,3)/-0.0220797098d0,-0.9453230065d0,
     1     0.3253873075d0/
	data (spheres(j,126),j=1,3)/-0.6933776718d0, 0.5698438841d0,
     1    -0.4410276091d0/
	data (spheres(j,127),j=1,3)/-0.7860164288d0, 0.5604580041d0,
     1     0.2608926971d0/
	data (spheres(j,128),j=1,3)/-0.3116685597d0,-0.9276848916d0,
     1    -0.2055807645d0/
	data (spheres(j,129),j=1,3)/ 0.2566810086d0,-0.4960836011d0,
     1    -0.8294672511d0/
	data (spheres(j,130),j=1,3)/-0.3310149503d0,-0.7382138967d0,
     1    -0.5877664038d0/
	data (spheres(j,131),j=1,3)/-0.0519921176d0,-0.4594924160d0,
     1     0.8866586375d0/
	data (spheres(j,132),j=1,3)/ 0.1869145542d0,-0.3381943432d0,
     1     0.9223272389d0/
	data (spheres(j,133),j=1,3)/ 0.4307967128d0, 0.7855185649d0,
     1    -0.4442688110d0/
	data (spheres(j,134),j=1,3)/-0.4981987696d0,-0.7658411772d0,
     1     0.4065529206d0/
	data (spheres(j,135),j=1,3)/-0.4285139230d0, 0.0781997161d0,
     1     0.9001447785d0/
	data (spheres(j,136),j=1,3)/-0.8635700286d0,-0.2786428773d0,
     1     0.4202439203d0/
	data (spheres(j,137),j=1,3)/ 0.3927977172d0,-0.7875707804d0,
     1     0.4748075603d0/
	data (spheres(j,138),j=1,3)/ 0.8774469259d0,-0.4302788522d0,
     1     0.2120070791d0/
	data (spheres(j,139),j=1,3)/ 0.2506628474d0, 0.1279131238d0,
     1    -0.9595865619d0/
	data (spheres(j,140),j=1,3)/ 0.0032058170d0,-0.9555678266d0,
     1    -0.2947538830d0/
	data (spheres(j,141),j=1,3)/ 0.9410887264d0,-0.2728112883d0,
     1     0.1998149393d0/
	data (spheres(j,142),j=1,3)/ 0.0817785543d0, 0.7398619945d0,
     1     0.6677697935d0/
	data (spheres(j,143),j=1,3)/-0.9003086748d0, 0.0123263593d0,
     1     0.4350774079d0/
	data (spheres(j,144),j=1,3)/ 0.5367467609d0, 0.8310297697d0,
     1    -0.1459192807d0/
	data (spheres(j,145),j=1,3)/ 0.8374457118d0, 0.4495369681d0,
     1     0.3108073264d0/
	data (spheres(j,146),j=1,3)/-0.9773725435d0, 0.1627816555d0,
     1    -0.1350742164d0/
	data (spheres(j,147),j=1,3)/-0.4630704862d0, 0.8712979568d0,
     1     0.1624979854d0/
	data (spheres(j,148),j=1,3)/ 0.2352796990d0,-0.8120305756d0,
     1     0.5340878275d0/
	data (spheres(j,149),j=1,3)/-0.9585765689d0, 0.0016351981d0,
     1     0.2848302787d0/
	data (spheres(j,150),j=1,3)/ 0.1088951893d0,-0.4499805546d0,
     1    -0.8863742653d0/
	data (spheres(j,151),j=1,3)/ 0.4487457347d0, 0.8650087918d0,
     1     0.2244706122d0/
	data (spheres(j,152),j=1,3)/-0.0901688252d0,-0.7921725811d0,
     1     0.6035993578d0/
	data (spheres(j,153),j=1,3)/ 0.7531274717d0, 0.5217227954d0,
     1    -0.4007547080d0/
	data (spheres(j,154),j=1,3)/-0.0778964495d0, 0.7349083722d0,
     1     0.6736778367d0/
	data (spheres(j,155),j=1,3)/-0.3519545043d0,-0.9350972920d0,
     1    -0.0414859190d0/
	data (spheres(j,156),j=1,3)/-0.3456951241d0, 0.6416280034d0,
     1    -0.6846958350d0/
	data (spheres(j,157),j=1,3)/-0.5561816546d0,-0.2253514960d0,
     1     0.7999241654d0/
	data (spheres(j,158),j=1,3)/-0.4601826591d0, 0.8775764952d0,
     1    -0.1345043318d0/
	data (spheres(j,159),j=1,3)/ 0.7147551548d0,-0.5066080177d0,
     1     0.4821549388d0/
	data (spheres(j,160),j=1,3)/ 0.9629024389d0, 0.1615850962d0,
     1    -0.2161229969d0/
	data (spheres(j,161),j=1,3)/ 0.9987647146d0,-0.0457995155d0,
     1     0.0192730179d0/
	data (spheres(j,162),j=1,3)/-0.9921127295d0,-0.0120310978d0,
     1     0.1247701272d0/
	data (spheres(j,163),j=1,3)/ 0.2834219581d0,-0.9357251457d0,
     1    -0.2099772495d0/
	data (spheres(j,164),j=1,3)/-0.7892672838d0, 0.6136072585d0,
     1    -0.0233085175d0/
	data (spheres(j,165),j=1,3)/-0.8809130870d0, 0.3027056236d0,
     1     0.3638151161d0/
	data (spheres(j,166),j=1,3)/-0.8195430824d0, 0.1182580142d0,
     1    -0.5606818867d0/
	data (spheres(j,167),j=1,3)/-0.1119691983d0,-0.0810415825d0,
     1    -0.9904015148d0/
	data (spheres(j,168),j=1,3)/ 0.3710218903d0, 0.5866641746d0,
     1    -0.7198388036d0/
	data (spheres(j,169),j=1,3)/ 0.2338871806d0, 0.8811390923d0,
     1     0.4109631210d0/
	data (spheres(j,170),j=1,3)/ 0.8397320877d0, 0.4710496705d0,
     1    -0.2701152140d0/
	data (spheres(j,171),j=1,3)/-0.3130084654d0,-0.5261955032d0,
     1    -0.7906604789d0/
	data (spheres(j,172),j=1,3)/-0.1910114280d0, 0.6084262728d0,
     1     0.7702805365d0/
	data (spheres(j,173),j=1,3)/-0.1101209956d0,-0.9053297387d0,
     1    -0.4101846298d0/
	data (spheres(j,174),j=1,3)/ 0.0194663013d0, 0.9146345679d0,
     1    -0.4038126674d0/
	data (spheres(j,175),j=1,3)/-0.9945911652d0,-0.1038272946d0,
     1    -0.0028821852d0/
	data (spheres(j,176),j=1,3)/ 0.4869174838d0,-0.6848695852d0,
     1     0.5420931795d0/
	data (spheres(j,177),j=1,3)/-0.6644583529d0,-0.6602406105d0,
     1    -0.3501106019d0/
	data (spheres(j,178),j=1,3)/ 0.6213256968d0, 0.7833487766d0,
     1    -0.0178626035d0/
	data (spheres(j,179),j=1,3)/ 0.6213686483d0,-0.2900430620d0,
     1    -0.7278571461d0/
	data (spheres(j,180),j=1,3)/ 0.7297587535d0,-0.4628928895d0,
     1    -0.5031722712d0/
	data (spheres(j,181),j=1,3)/-0.2604427446d0, 0.0867270491d0,
     1     0.9615861874d0/
	data (spheres(j,182),j=1,3)/ 0.7015723479d0, 0.2966271920d0,
     1    -0.6479263458d0/
	data (spheres(j,183),j=1,3)/-0.5056807878d0,-0.0656146624d0,
     1     0.8602218649d0/
	data (spheres(j,184),j=1,3)/ 0.4836363161d0,-0.0477464380d0,
     1     0.8739657839d0/
	data (spheres(j,185),j=1,3)/-0.2703073252d0,-0.6539051634d0,
     1    -0.7066413427d0/
	data (spheres(j,186),j=1,3)/-0.7922889084d0, 0.5099994993d0,
     1    -0.3349310323d0/
	data (spheres(j,187),j=1,3)/-0.8271719949d0,-0.0500483266d0,
     1    -0.5597156921d0/
	data (spheres(j,188),j=1,3)/-0.9324935750d0,-0.3584898983d0,
     1    -0.0440536643d0/
	data (spheres(j,189),j=1,3)/ 0.4009169150d0,-0.5307320590d0,
     1    -0.7467188954d0/
	data (spheres(j,190),j=1,3)/ 0.7833560842d0,-0.6178072807d0,
     1    -0.0683184405d0/
	data (spheres(j,191),j=1,3)/ 0.7175816439d0,-0.5894907296d0,
     1    -0.3709140925d0/
	data (spheres(j,192),j=1,3)/-0.1620544020d0, 0.3383897413d0,
     1    -0.9269470070d0/
	data (spheres(j,193),j=1,3)/-0.1637798647d0, 0.9780667106d0,
     1    -0.1286921343d0/
	data (spheres(j,194),j=1,3)/ 0.4758174894d0, 0.7315921217d0,
     1     0.4882322033d0/
	data (spheres(j,195),j=1,3)/ 0.7993236229d0,-0.3618563387d0,
     1     0.4797308996d0/
	data (spheres(j,196),j=1,3)/ 0.9630002188d0, 0.2547903776d0,
     1    -0.0878205105d0/
	data (spheres(j,197),j=1,3)/-0.5320164546d0, 0.7945749317d0,
     1     0.2925904475d0/
	data (spheres(j,198),j=1,3)/ 0.7435022726d0, 0.6171343207d0,
     1    -0.2575841627d0/
	data (spheres(j,199),j=1,3)/ 0.2118859220d0, 0.5759702772d0,
     1    -0.7895331505d0/
	data (spheres(j,200),j=1,3)/-0.0540652506d0,-0.3950855627d0,
     1    -0.9170519870d0/
	data (spheres(j,201),j=1,3)/-0.8005302780d0,-0.5917330251d0,
     1    -0.0948857258d0/
	data (spheres(j,202),j=1,3)/ 0.7383513528d0,-0.6696672608d0,
     1     0.0798939269d0/
	data (spheres(j,203),j=1,3)/ 0.4129477846d0, 0.9068001499d0,
     1    -0.0847798049d0/
	data (spheres(j,204),j=1,3)/-0.6008981668d0,-0.3879002723d0,
     1    -0.6988953941d0/
	data (spheres(j,205),j=1,3)/ 0.0268848679d0,-0.3272515136d0,
     1     0.9445547367d0/
	data (spheres(j,206),j=1,3)/-0.7619112955d0,-0.5474446723d0,
     1     0.3461148776d0/
	data (spheres(j,207),j=1,3)/-0.5983641060d0,-0.7965490993d0,
     1    -0.0864287511d0/
	data (spheres(j,208),j=1,3)/ 0.6062168423d0,-0.6973105582d0,
     1    -0.3824383944d0/
	data (spheres(j,209),j=1,3)/-0.3765357435d0,-0.2304547849d0,
     1    -0.8972800154d0/
	data (spheres(j,210),j=1,3)/-0.1560705454d0,-0.9546122660d0,
     1    -0.2536876161d0/
	data (spheres(j,211),j=1,3)/-0.9233553078d0,-0.2849153265d0,
     1     0.2573678929d0/
	data (spheres(j,212),j=1,3)/ 0.6037054895d0,-0.1316829435d0,
     1    -0.7862565003d0/
	data (spheres(j,213),j=1,3)/ 0.3618449495d0, 0.2667357105d0,
     1     0.8932638430d0/
	data (spheres(j,214),j=1,3)/-0.5006592880d0,-0.8656209821d0,
     1     0.0063712416d0/
	data (spheres(j,215),j=1,3)/ 0.4180604394d0,-0.1958250918d0,
     1     0.8870614423d0/
	data (spheres(j,216),j=1,3)/-0.9411637497d0, 0.3244254553d0,
     1    -0.0946515727d0/
	data (spheres(j,217),j=1,3)/-0.0375377913d0, 0.2465922600d0,
     1    -0.9683920547d0/
	data (spheres(j,218),j=1,3)/ 0.1364729253d0, 0.6278095411d0,
     1     0.7663095463d0/
	data (spheres(j,219),j=1,3)/ 0.3371530392d0,-0.0465479947d0,
     1     0.9402984166d0/
	data (spheres(j,220),j=1,3)/ 0.1631068848d0, 0.8657036687d0,
     1    -0.4732370465d0/
	data (spheres(j,221),j=1,3)/ 0.0293112893d0,-0.7135016007d0,
     1     0.7000402232d0/
	data (spheres(j,222),j=1,3)/ 0.7745752574d0, 0.5934080285d0,
     1     0.2188608745d0/
	data (spheres(j,223),j=1,3)/ 0.8081422131d0, 0.2662261894d0,
     1    -0.5253853629d0/
	data (spheres(j,224),j=1,3)/ 0.5369960783d0, 0.1011491731d0,
     1     0.8374986905d0/
	data (spheres(j,225),j=1,3)/-0.2173803936d0,-0.3212627089d0,
     1    -0.9217027918d0/
	data (spheres(j,226),j=1,3)/-0.6161730119d0, 0.5788541283d0,
     1     0.5340961689d0/
	data (spheres(j,227),j=1,3)/ 0.0209348338d0,-0.9868844478d0,
     1     0.1600650472d0/
	data (spheres(j,228),j=1,3)/ 0.9380740703d0, 0.0650328678d0,
     1    -0.3402760125d0/
	data (spheres(j,229),j=1,3)/ 0.0706872955d0,-0.9974374602d0,
     1    -0.0110371775d0/
	data (spheres(j,230),j=1,3)/ 0.3596386202d0, 0.4437398532d0,
     1     0.8208258071d0/
	data (spheres(j,231),j=1,3)/-0.3449620468d0,-0.4773289646d0,
     1     0.8081820623d0/
	data (spheres(j,232),j=1,3)/-0.4427016840d0, 0.7016013150d0,
     1    -0.5583644095d0/
	data (spheres(j,233),j=1,3)/-0.0205107925d0,-0.7028931612d0,
     1    -0.7109996564d0/
	data (spheres(j,234),j=1,3)/-0.2709016782d0,-0.8919380531d0,
     1    -0.3620204277d0/
	data (spheres(j,235),j=1,3)/-0.5598896983d0,-0.5371719330d0,
     1    -0.6308485081d0/
	data (spheres(j,236),j=1,3)/-0.4870299946d0, 0.1911939433d0,
     1    -0.8522010681d0/
	data (spheres(j,237),j=1,3)/-0.7344813924d0, 0.0237959303d0,
     1    -0.6782114994d0/
	data (spheres(j,238),j=1,3)/-0.3694217945d0, 0.1072299339d0,
     1    -0.9230543207d0/
	data (spheres(j,239),j=1,3)/-0.3352041766d0,-0.8513560730d0,
     1     0.4035232299d0/
	data (spheres(j,240),j=1,3)/ 0.5530132309d0,-0.8046693792d0,
     1     0.2160637789d0/
	data (spheres(j,241),j=1,3)/ 0.5958516310d0, 0.4643119684d0,
     1     0.6552672965d0/
	data (spheres(j,242),j=1,3)/ 0.8661740108d0, 0.4743843789d0,
     1     0.1571688396d0/
	data (spheres(j,243),j=1,3)/ 0.6872661967d0,-0.4019136113d0,
     1     0.6050872862d0/
	data (spheres(j,244),j=1,3)/ 0.4542718143d0, 0.5424686591d0,
     1     0.7066575356d0/
	data (spheres(j,245),j=1,3)/ 0.2268959060d0,-0.9722043957d0,
     1    -0.0577655674d0/
	data (spheres(j,246),j=1,3)/-0.0815980127d0, 0.3655734733d0,
     1     0.9271988999d0/
	data (spheres(j,247),j=1,3)/-0.2451155437d0,-0.1612397780d0,
     1    -0.9559916863d0/
	data (spheres(j,248),j=1,3)/ 0.1637710437d0, 0.8207425215d0,
     1     0.5473214399d0/
	data (spheres(j,249),j=1,3)/ 0.4948651555d0,-0.3226898393d0,
     1     0.8068331584d0/
	data (spheres(j,250),j=1,3)/ 0.9458715736d0,-0.2335246463d0,
     1    -0.2253734809d0/
	data (spheres(j,251),j=1,3)/ 0.1344498210d0, 0.2870547746d0,
     1    -0.9484317593d0/
	data (spheres(j,252),j=1,3)/ 0.9532829992d0, 0.0077228966d0,
     1     0.3019799336d0/
	data (spheres(j,253),j=1,3)/ 0.8078600331d0,-0.4734272685d0,
     1     0.3510395823d0/
	data (spheres(j,254),j=1,3)/ 0.0033306930d0, 0.8411062475d0,
     1    -0.5408596740d0/
	data (spheres(j,255),j=1,3)/-0.8735459448d0, 0.4256706540d0,
     1     0.2360550289d0/
	data (spheres(j,256),j=1,3)/-0.0504699830d0,-0.8791778573d0,
     1     0.4738133346d0/
	data (spheres(j,257),j=1,3)/-0.5820288087d0,-0.3824728869d0,
     1     0.7176050144d0/
	data (spheres(j,258),j=1,3)/ 0.3226675290d0, 0.7868431507d0,
     1     0.5260831892d0/
	data (spheres(j,259),j=1,3)/ 0.4547266942d0,-0.5841742837d0,
     1     0.6722827083d0/
	data (spheres(j,260),j=1,3)/-0.8315982092d0,-0.5202113467d0,
     1     0.1944854061d0/
	data (spheres(j,261),j=1,3)/ 0.4398971534d0,-0.8677363603d0,
     1    -0.2313527685d0/
	data (spheres(j,262),j=1,3)/-0.1708537097d0, 0.8159651870d0,
     1     0.5522769445d0/
	data (spheres(j,263),j=1,3)/-0.7683107112d0, 0.1755613828d0,
     1     0.6155297327d0/
	data (spheres(j,264),j=1,3)/ 0.0941272281d0, 0.1456856137d0,
     1    -0.9848430164d0/
	data (spheres(j,265),j=1,3)/-0.8453097911d0, 0.5241153738d0,
     1     0.1037035777d0/
	data (spheres(j,266),j=1,3)/ 0.7483936674d0, 0.2043738931d0,
     1     0.6309819573d0/
	data (spheres(j,267),j=1,3)/-0.4381088501d0, 0.7919917307d0,
     1     0.4252172785d0/
	data (spheres(j,268),j=1,3)/ 0.2664000696d0,-0.8808790477d0,
     1     0.3912583625d0/
	data (spheres(j,269),j=1,3)/-0.9231552484d0,-0.1418493243d0,
     1     0.3572998134d0/
	data (spheres(j,270),j=1,3)/ 0.3035178081d0, 0.5875195347d0,
     1     0.7501318128d0/
	data (spheres(j,271),j=1,3)/-0.4972071103d0,-0.3094556301d0,
     1    -0.8105691225d0/
	data (spheres(j,272),j=1,3)/-0.8020063400d0, 0.3119410052d0,
     1     0.5093904591d0/
	data (spheres(j,273),j=1,3)/-0.8833495761d0,-0.2768609606d0,
     1    -0.3782083222d0/
	data (spheres(j,274),j=1,3)/-0.2205220924d0, 0.1889424838d0,
     1    -0.9569068631d0/
	data (spheres(j,275),j=1,3)/-0.0854265006d0,-0.9957388567d0,
     1     0.0347338776d0/
	data (spheres(j,276),j=1,3)/ 0.5639513068d0, 0.0286498818d0,
     1    -0.8253109159d0/
	data (spheres(j,277),j=1,3)/ 0.1848525156d0,-0.0168901472d0,
     1    -0.9826211225d0/
	data (spheres(j,278),j=1,3)/-0.0424159536d0,-0.9901637399d0,
     1    -0.1333291230d0/
	data (spheres(j,279),j=1,3)/ 0.2439636327d0, 0.7124385579d0,
     1     0.6579612809d0/
	data (spheres(j,280),j=1,3)/-0.7991583823d0, 0.2765942419d0,
     1    -0.5337054482d0/
	data (spheres(j,281),j=1,3)/ 0.3973808485d0, 0.6719966547d0,
     1     0.6249071590d0/
	data (spheres(j,282),j=1,3)/-0.6697121414d0, 0.6871040643d0,
     1     0.2817333002d0/
	data (spheres(j,283),j=1,3)/-0.1760636170d0,-0.5683295491d0,
     1     0.8037431968d0/
	data (spheres(j,284),j=1,3)/ 0.9060359420d0, 0.4015649148d0,
     1    -0.1335832742d0/
	data (spheres(j,285),j=1,3)/ 0.9743261628d0, 0.2041944554d0,
     1     0.0948322353d0/
	data (spheres(j,286),j=1,3)/-0.6407963718d0,-0.0922072299d0,
     1     0.7621534207d0/
	data (spheres(j,287),j=1,3)/ 0.9310411718d0,-0.1584059615d0,
     1     0.3287398483d0/
	data (spheres(j,288),j=1,3)/ 0.2730834606d0, 0.9615938163d0,
     1    -0.0276180382d0/
	data (spheres(j,289),j=1,3)/-0.9890659962d0,-0.0217348661d0,
     1    -0.1458631233d0/
	data (spheres(j,290),j=1,3)/-0.0236707429d0,-0.5427133581d0,
     1    -0.8395843655d0/
	data (spheres(j,291),j=1,3)/ 0.7256904762d0, 0.6835261302d0,
     1    -0.0785198205d0/
	data (spheres(j,292),j=1,3)/-0.1756300212d0, 0.2329558960d0,
     1     0.9564965479d0/
	data (spheres(j,293),j=1,3)/-0.6448450144d0,-0.6580936693d0,
     1     0.3886999226d0/
	data (spheres(j,294),j=1,3)/-0.5015290265d0, 0.2081271404d0,
     1     0.8397331296d0/
	data (spheres(j,295),j=1,3)/ 0.0369391570d0, 0.9869721638d0,
     1     0.1565932519d0/
	data (spheres(j,296),j=1,3)/ 0.0260820741d0, 0.5069879002d0,
     1     0.8615584684d0/
	data (spheres(j,297),j=1,3)/-0.8175390963d0, 0.0239130735d0,
     1     0.5753763907d0/
	data (spheres(j,298),j=1,3)/-0.4615621087d0,-0.8733740018d0,
     1    -0.1554936421d0/
	data (spheres(j,299),j=1,3)/-0.9113352562d0,-0.3888278020d0,
     1     0.1352072158d0/
	data (spheres(j,300),j=1,3)/-0.7406612791d0,-0.5239244091d0,
     1    -0.4206234459d0/
	data (spheres(j,301),j=1,3)/-0.4277453821d0,-0.8490965037d0,
     1    -0.3099500210d0/
	data (spheres(j,302),j=1,3)/ 0.7177611782d0, 0.3646042274d0,
     1     0.5931969727d0/
	data (spheres(j,303),j=1,3)/-0.6245111810d0,-0.2274479018d0,
     1    -0.7471634605d0/
	data (spheres(j,304),j=1,3)/-0.9971892291d0, 0.0739562050d0,
     1    -0.0120050472d0/
	data (spheres(j,305),j=1,3)/-0.6304693876d0,-0.0610976110d0,
     1    -0.7738058110d0/
	data (spheres(j,306),j=1,3)/ 0.2570874357d0,-0.7544182641d0,
     1    -0.6039529230d0/
	data (spheres(j,307),j=1,3)/ 0.3592252741d0,-0.2266935273d0,
     1    -0.9052995345d0/
	data (spheres(j,308),j=1,3)/ 0.8736702721d0, 0.4865132095d0,
     1    -0.0022699451d0/
	data (spheres(j,309),j=1,3)/ 0.0096446288d0,-0.0390291845d0,
     1     0.9991915251d0/
	data (spheres(j,310),j=1,3)/ 0.7609467117d0, 0.1390468736d0,
     1    -0.6337397485d0/
	data (spheres(j,311),j=1,3)/-0.4171754032d0, 0.8626132297d0,
     1    -0.2861172819d0/
	data (spheres(j,312),j=1,3)/-0.7942302334d0,-0.3725935347d0,
     1    -0.4799712432d0/
	data (spheres(j,313),j=1,3)/ 0.0030006083d0, 0.2341073677d0,
     1     0.9722061184d0/
	data (spheres(j,314),j=1,3)/ 0.5173618139d0, 0.5804877127d0,
     1    -0.6287851532d0/
	data (spheres(j,315),j=1,3)/-0.4647390377d0,-0.4624900775d0,
     1    -0.7550632788d0/
	data (spheres(j,316),j=1,3)/-0.7182753769d0, 0.1872265727d0,
     1    -0.6700945407d0/
	data (spheres(j,317),j=1,3)/ 0.8903902971d0,-0.4072738269d0,
     1    -0.2033055550d0/
	data (spheres(j,318),j=1,3)/-0.2213654086d0,-0.8321806298d0,
     1    -0.5084021590d0/
	data (spheres(j,319),j=1,3)/ 0.8595823163d0,-0.2037290847d0,
     1     0.4686285326d0/
	data (spheres(j,320),j=1,3)/-0.3032598825d0, 0.7659897662d0,
     1    -0.5668272416d0/
	data (spheres(j,321),j=1,3)/ 0.9234505498d0,-0.3778050805d0,
     1     0.0670999499d0/
	data (spheres(j,322),j=1,3)/-0.2370239100d0, 0.5692137336d0,
     1    -0.7872835522d0/
	data (spheres(j,323),j=1,3)/-0.9762363160d0, 0.1388376394d0,
     1     0.1663934052d0/
	data (spheres(j,324),j=1,3)/ 0.8981513213d0,-0.1808600133d0,
     1    -0.4007665899d0/
	data (spheres(j,325),j=1,3)/-0.4174595825d0,-0.1935438966d0,
     1     0.8878447258d0/
	data (spheres(j,326),j=1,3)/ 0.7873308063d0,-0.5778661873d0,
     1     0.2148973500d0/
	data (spheres(j,327),j=1,3)/-0.4835802150d0,-0.6834486379d0,
     1    -0.5468529374d0/
	data (spheres(j,328),j=1,3)/ 0.2084244324d0, 0.9687355925d0,
     1     0.1345756581d0/
	data (spheres(j,329),j=1,3)/-0.5776845481d0, 0.7070123893d0,
     1    -0.4079387751d0/
	data (spheres(j,330),j=1,3)/-0.5937896874d0, 0.7947413897d0,
     1    -0.1256977746d0/
	data (spheres(j,331),j=1,3)/ 0.8070466002d0,-0.5502227758d0,
     1    -0.2143144467d0/
	data (spheres(j,332),j=1,3)/ 0.4037531960d0, 0.6955591853d0,
     1    -0.5942901450d0/
	data (spheres(j,333),j=1,3)/ 0.3268906145d0,-0.8745206089d0,
     1    -0.3582683781d0/
	data (spheres(j,334),j=1,3)/ 0.6232940978d0,-0.7790372875d0,
     1     0.0678628935d0/
	data (spheres(j,335),j=1,3)/ 0.7265171755d0,-0.3121010449d0,
     1    -0.6121811263d0/
	data (spheres(j,336),j=1,3)/ 0.2570177648d0, 0.1055855863d0,
     1     0.9606214408d0/
	data (spheres(j,337),j=1,3)/-0.6854608664d0,-0.4598899213d0,
     1    -0.5644861920d0/
	data (spheres(j,338),j=1,3)/-0.5293321417d0,-0.8345405289d0,
     1     0.1528057242d0/
	data (spheres(j,339),j=1,3)/-0.9193369403d0,-0.3282542575d0,
     1    -0.2169532961d0/
	data (spheres(j,340),j=1,3)/ 0.3270023181d0, 0.4639351072d0,
     1    -0.8233065652d0/
	data (spheres(j,341),j=1,3)/ 0.3428059660d0,-0.3408707093d0,
     1     0.8753806196d0/
	data (spheres(j,342),j=1,3)/-0.2678437494d0, 0.9195192340d0,
     1    -0.2876527495d0/
	data (spheres(j,343),j=1,3)/ 0.5455019573d0,-0.8323426485d0,
     1    -0.0981495290d0/
	data (spheres(j,344),j=1,3)/ 0.6936884706d0, 0.7117132108d0,
     1     0.1107276448d0/
	data (spheres(j,345),j=1,3)/-0.6073148660d0, 0.4507550713d0,
     1     0.6542083149d0/
	data (spheres(j,346),j=1,3)/-0.0717747262d0, 0.9969307411d0,
     1     0.0312647745d0/
	data (spheres(j,347),j=1,3)/ 0.1756214643d0,-0.9777858318d0,
     1     0.1144201400d0/
	data (spheres(j,348),j=1,3)/-0.8975664145d0,-0.1216536602d0,
     1    -0.4237628093d0/
	data (spheres(j,349),j=1,3)/-0.9500341106d0,-0.1739205293d0,
     1    -0.2592042403d0/
	data (spheres(j,350),j=1,3)/-0.1947959272d0, 0.7017795498d0,
     1    -0.6852444894d0/
	data (spheres(j,351),j=1,3)/ 0.1690619972d0,-0.9285806065d0,
     1    -0.3303877999d0/
	data (spheres(j,352),j=1,3)/-0.0784103148d0, 0.6253100237d0,
     1    -0.7764271999d0/
	data (spheres(j,353),j=1,3)/-0.1839948928d0,-0.4722945359d0,
     1    -0.8620230570d0/
	data (spheres(j,354),j=1,3)/-0.8024465407d0, 0.5674702210d0,
     1    -0.1845456516d0/
	data (spheres(j,355),j=1,3)/ 0.3655305787d0, 0.9256204200d0,
     1     0.0980522013d0/
	data (spheres(j,356),j=1,3)/-0.2346701665d0, 0.9718833203d0,
     1     0.0193060782d0/
	data (spheres(j,357),j=1,3)/-0.4768476492d0, 0.6855592290d0,
     1     0.5501135000d0/
	data (spheres(j,358),j=1,3)/-0.7805456543d0,-0.5669210747d0,
     1    -0.2633419385d0/
	data (spheres(j,359),j=1,3)/-0.3895853459d0,-0.7568693361d0,
     1     0.5247590555d0/
	data (spheres(j,360),j=1,3)/-0.3837599839d0,-0.9151021689d0,
     1     0.1237590205d0/
	data (spheres(j,361),j=1,3)/ 0.5392806479d0, 0.6816393064d0,
     1    -0.4945141442d0/
	data (spheres(j,362),j=1,3)/ 0.6768975210d0, 0.5184865209d0,
     1     0.5224762900d0/
	data (spheres(j,363),j=1,3)/ 0.6900902335d0, 0.0043437989d0,
     1    -0.7237103019d0/
	data (spheres(j,364),j=1,3)/ 0.3836108728d0,-0.3867697262d0,
     1    -0.8386011431d0/
	data (spheres(j,365),j=1,3)/ 0.5686191202d0,-0.1740591851d0,
     1     0.8039749351d0/
	data (spheres(j,366),j=1,3)/ 0.0071612758d0, 0.3954481512d0,
     1    -0.9184603834d0/
	data (spheres(j,367),j=1,3)/ 0.7414713388d0, 0.0325655854d0,
     1     0.6701938051d0/
	data (spheres(j,368),j=1,3)/ 0.5794928442d0, 0.7970407220d0,
     1     0.1700415568d0/
	data (spheres(j,369),j=1,3)/-0.1992291606d0,-0.9759183126d0,
     1    -0.0888323628d0/
	data (spheres(j,370),j=1,3)/-0.3435251041d0,-0.0531102191d0,
     1     0.9376405535d0/
	data (spheres(j,371),j=1,3)/ 0.0890730491d0, 0.0986942748d0,
     1     0.9911233183d0/
	data (spheres(j,372),j=1,3)/ 0.1909889316d0, 0.5033835625d0,
     1     0.8426910567d0/
	data (spheres(j,373),j=1,3)/-0.7189720785d0,-0.6564712524d0,
     1     0.2283082237d0/
	data (spheres(j,374),j=1,3)/ 0.3921418627d0,-0.9160095807d0,
     1    -0.0845648126d0/
	data (spheres(j,375),j=1,3)/-0.6187690426d0,-0.6088232813d0,
     1    -0.4964464564d0/
	data (spheres(j,376),j=1,3)/-0.1373887298d0,-0.6150374878d0,
     1    -0.7764362340d0/
	data (spheres(j,377),j=1,3)/ 0.5543186245d0,-0.4389402695d0,
     1     0.7071508342d0/
	data (spheres(j,378),j=1,3)/-0.6499114581d0,-0.7529561764d0,
     1     0.1033058228d0/
	data (spheres(j,379),j=1,3)/ 0.0548222096d0,-0.1510227512d0,
     1    -0.9870089432d0/
	data (spheres(j,380),j=1,3)/-0.3854665148d0,-0.8029172012d0,
     1    -0.4546861929d0/
	data (spheres(j,381),j=1,3)/ 0.7881221735d0, 0.4120404767d0,
     1     0.4572593194d0/
	data (spheres(j,382),j=1,3)/ 0.1369444219d0, 0.9485559429d0,
     1     0.2854607652d0/
	data (spheres(j,383),j=1,3)/ 0.3954830469d0,-0.6674095247d0,
     1    -0.6309973740d0/
	data (spheres(j,384),j=1,3)/ 0.6648534944d0, 0.6924291108d0,
     1     0.2801994959d0/
	data (spheres(j,385),j=1,3)/-0.4227553359d0,-0.6046383327d0,
     1    -0.6750484521d0/
	data (spheres(j,386),j=1,3)/-0.7830689753d0,-0.4145876414d0,
     1     0.4635947233d0/
	data (spheres(j,387),j=1,3)/-0.8070159728d0, 0.4438489821d0,
     1     0.3895167527d0/
	data (spheres(j,388),j=1,3)/ 0.0682699484d0, 0.9020692105d0,
     1     0.4261576628d0/
	data (spheres(j,389),j=1,3)/-0.5939198001d0, 0.2693144858d0,
     1    -0.7581088172d0/
	data (spheres(j,390),j=1,3)/ 0.9499142823d0, 0.1727720899d0,
     1     0.2604086427d0/
	data (spheres(j,391),j=1,3)/-0.7059474274d0, 0.6913655001d0,
     1    -0.1537919862d0/
	data (spheres(j,392),j=1,3)/-0.3382529690d0, 0.2745566922d0,
     1    -0.9001130772d0/
	data (spheres(j,393),j=1,3)/-0.5820839410d0, 0.6964805067d0,
     1     0.4196345902d0/
	data (spheres(j,394),j=1,3)/ 0.5022798864d0, 0.2547716904d0,
     1     0.8263209434d0/
	data (spheres(j,395),j=1,3)/-0.1345591927d0,-0.9691971623d0,
     1     0.2062781721d0/
	data (spheres(j,396),j=1,3)/ 0.8171380910d0, 0.5613249387d0,
     1    -0.1311474490d0/
	data (spheres(j,397),j=1,3)/-0.5732307470d0, 0.6177768930d0,
     1    -0.5382919479d0/
	data (spheres(j,398),j=1,3)/ 0.3096891431d0, 0.9278161488d0,
     1    -0.2079659315d0/
	data (spheres(j,399),j=1,3)/-0.7562595542d0,-0.1146685311d0,
     1     0.6441448708d0/
	data (spheres(j,400),j=1,3)/ 0.6258594608d0, 0.3087968016d0,
     1     0.7162014177d0/
	data (spheres(j,401),j=1,3)/ 0.2711303066d0,-0.6277868987d0,
     1    -0.7296382437d0/
	data (spheres(j,402),j=1,3)/ 0.1644480157d0, 0.4445106037d0,
     1    -0.8805493588d0/
	data (spheres(j,403),j=1,3)/ 0.4363188033d0, 0.8584710362d0,
     1    -0.2695429129d0/
	data (spheres(j,404),j=1,3)/ 0.4822640526d0, 0.4658806027d0,
     1    -0.7418737410d0/
	data (spheres(j,405),j=1,3)/ 0.4185583238d0, 0.0550189379d0,
     1    -0.9065218398d0/
	data (spheres(j,406),j=1,3)/ 0.7981748314d0,-0.0192373177d0,
     1    -0.6021186462d0/
	data (spheres(j,407),j=1,3)/ 0.5171159175d0, 0.1863515408d0,
     1    -0.8353826854d0/
	data (spheres(j,408),j=1,3)/-0.0038581199d0, 0.8325195268d0,
     1     0.5539822672d0/
	data (spheres(j,409),j=1,3)/-0.7195380116d0, 0.4508396551d0,
     1     0.5282126989d0/
	data (spheres(j,410),j=1,3)/-0.3111809151d0,-0.6140519499d0,
     1     0.7253320900d0/
	data (spheres(j,411),j=1,3)/-0.2860765029d0,-0.9247196496d0,
     1     0.2511051654d0/
	data (spheres(j,412),j=1,3)/-0.5819691649d0, 0.0650553341d0,
     1     0.8106045242d0/
	data (spheres(j,413),j=1,3)/ 0.1035951972d0,-0.1882572317d0,
     1     0.9766407988d0/
	data (spheres(j,414),j=1,3)/-0.7094849534d0, 0.5777537161d0,
     1     0.4035241559d0/
	data (spheres(j,415),j=1,3)/ 0.7201356340d0,-0.1548624907d0,
     1    -0.6763300065d0/
	data (spheres(j,416),j=1,3)/-0.1416518205d0, 0.8935854779d0,
     1    -0.4259574572d0/
	data (spheres(j,417),j=1,3)/ 0.2644389571d0,-0.1956150652d0,
     1     0.9443552214d0/
	data (spheres(j,418),j=1,3)/-0.5110543849d0,-0.1449217048d0,
     1    -0.8472432444d0/
	data (spheres(j,419),j=1,3)/-0.1777763745d0,-0.9120824867d0,
     1     0.3694605501d0/
	data (spheres(j,420),j=1,3)/ 0.8881602525d0,-0.0371610216d0,
     1    -0.4580288467d0/
	data (spheres(j,421),j=1,3)/ 0.1484717766d0,-0.6071017980d0,
     1     0.7806302188d0/
	data (spheres(j,422),j=1,3)/-0.1388094909d0,-0.3064355799d0,
     1     0.9417160722d0/
	data (spheres(j,423),j=1,3)/ 0.9844227919d0,-0.1405306242d0,
     1    -0.1056546754d0/
	data (spheres(j,424),j=1,3)/-0.0273087712d0, 0.6274548825d0,
     1     0.7781738890d0/
	data (spheres(j,425),j=1,3)/-0.3811453234d0,-0.0635790891d0,
     1    -0.9223263749d0/
	data (spheres(j,426),j=1,3)/-0.6893804564d0, 0.3411754450d0,
     1    -0.6390257444d0/
	data (spheres(j,427),j=1,3)/ 0.6470927235d0, 0.1637526480d0,
     1    -0.7446180749d0/
	data (spheres(j,428),j=1,3)/-0.0586162790d0,-0.8283147316d0,
     1    -0.5571883319d0/
	data (spheres(j,429),j=1,3)/-0.3394025095d0, 0.2189914267d0,
     1     0.9147943439d0/
	data (spheres(j,430),j=1,3)/-0.4349414077d0,-0.6426758083d0,
     1     0.6307089481d0/
	data (spheres(j,431),j=1,3)/ 0.0903512677d0, 0.6682669184d0,
     1    -0.7384145002d0/
	data (spheres(j,432),j=1,3)/-0.6410150693d0, 0.4850224400d0,
     1    -0.5948553721d0/
	data (spheres(j,433),j=1,3)/ 0.9869126234d0, 0.0130963349d0,
     1    -0.1607232396d0/
	data (spheres(j,434),j=1,3)/-0.2507751593d0, 0.3690332472d0,
     1     0.8949448485d0/
	data (spheres(j,435),j=1,3)/ 0.8575990869d0, 0.1107382632d0,
     1    -0.5022557548d0/
	data (spheres(j,436),j=1,3)/-0.0336567626d0, 0.7435234116d0,
     1    -0.6678623801d0/
	data (spheres(j,437),j=1,3)/ 0.1290499119d0,-0.7079202445d0,
     1    -0.6944026552d0/
	data (spheres(j,438),j=1,3)/-0.7190436055d0,-0.3060551213d0,
     1    -0.6239443534d0/
	data (spheres(j,439),j=1,3)/ 0.1849892946d0, 0.9289116883d0,
     1    -0.3207834724d0/
	data (spheres(j,440),j=1,3)/-0.5954183829d0,-0.7551266342d0,
     1     0.2743368652d0/
	data (spheres(j,441),j=1,3)/-0.8976751702d0, 0.0464734932d0,
     1    -0.4382003004d0/
	data (spheres(j,442),j=1,3)/-0.5463824369d0,-0.6578591504d0,
     1     0.5183508183d0/
	data (spheres(j,443),j=1,3)/-0.9365853245d0, 0.2845731231d0,
     1     0.2045142234d0/
	data (spheres(j,444),j=1,3)/ 0.5217483733d0,-0.5640878864d0,
     1    -0.6399871026d0/
	data (spheres(j,445),j=1,3)/ 0.6154450589d0, 0.6561974651d0,
     1     0.4366145511d0/
	data (spheres(j,446),j=1,3)/ 0.8839149751d0, 0.2960443894d0,
     1     0.3620110997d0/
	data (spheres(j,447),j=1,3)/ 0.4968314537d0,-0.2653242293d0,
     1    -0.8262938702d0/
	data (spheres(j,448),j=1,3)/-0.7425334681d0, 0.6559988818d0,
     1     0.1353126593d0/
	data (spheres(j,449),j=1,3)/-0.4577273683d0,-0.3542916978d0,
     1     0.8154526652d0/
	data (spheres(j,450),j=1,3)/-0.5376495903d0, 0.8429008966d0,
     1     0.0212366797d0/
	data (spheres(j,451),j=1,3)/-0.8419953280d0,-0.4291755216d0,
     1    -0.3268826077d0/
	data (spheres(j,452),j=1,3)/ 0.8424949113d0,-0.5333953415d0,
     1     0.0754435825d0/
	data (spheres(j,453),j=1,3)/-0.1539797580d0, 0.8124129370d0,
     1    -0.5623837248d0/
	data (spheres(j,454),j=1,3)/-0.2956672820d0, 0.8522155270d0,
     1    -0.4316359042d0/
	data (spheres(j,455),j=1,3)/ 0.5496959823d0,-0.7491664835d0,
     1     0.3695726006d0/
	data (spheres(j,456),j=1,3)/-0.3160491825d0, 0.9389481958d0,
     1    -0.1359749970d0/
	data (spheres(j,457),j=1,3)/-0.9546273229d0,-0.0265991843d0,
     1    -0.2966128078d0/
	data (spheres(j,458),j=1,3)/ 0.5054449173d0,-0.6923077889d0,
     1    -0.5150100592d0/
	data (spheres(j,459),j=1,3)/ 0.4943799353d0, 0.3961689349d0,
     1     0.7737174255d0/
	data (spheres(j,460),j=1,3)/-0.8511951878d0, 0.3618142518d0,
     1    -0.3802067852d0/
	data (spheres(j,461),j=1,3)/-0.1378343058d0,-0.6881137573d0,
     1     0.7123911574d0/
	data (spheres(j,462),j=1,3)/ 0.9758020701d0,-0.2131047156d0,
     1     0.0489561054d0/
	data (spheres(j,463),j=1,3)/ 0.7621212722d0,-0.2516933254d0,
     1     0.5965078678d0/
	data (spheres(j,464),j=1,3)/-0.6954054167d0,-0.6915941028d0,
     1    -0.1952150184d0/
	data (spheres(j,465),j=1,3)/-0.4690549780d0,-0.5089052738d0,
     1     0.7218052715d0/
	data (spheres(j,466),j=1,3)/ 0.8110600951d0,-0.0868570809d0,
     1     0.5784784953d0/
	data (spheres(j,467),j=1,3)/ 0.6455814358d0, 0.7351392387d0,
     1    -0.2068693054d0/
	data (spheres(j,468),j=1,3)/ 0.9936442245d0, 0.1074674701d0,
     1    -0.0334947443d0/
	data (spheres(j,469),j=1,3)/ 0.8362880626d0, 0.3730785569d0,
     1    -0.4017893314d0/
	data (spheres(j,470),j=1,3)/ 0.2878920247d0, 0.7917583509d0,
     1    -0.5387363882d0/
	data (spheres(j,471),j=1,3)/-0.6917237260d0,-0.4007054607d0,
     1     0.6007939919d0/
	data (spheres(j,472),j=1,3)/ 0.4354744680d0, 0.3397082650d0,
     1    -0.8336427786d0/
	data (spheres(j,473),j=1,3)/ 0.9137931149d0, 0.3113224297d0,
     1    -0.2608840506d0/
	data (spheres(j,474),j=1,3)/ 0.5772719042d0, 0.3211384289d0,
     1    -0.7507511293d0/
	data (spheres(j,475),j=1,3)/-0.4418009352d0, 0.7903464440d0,
     1    -0.4244578096d0/
	data (spheres(j,476),j=1,3)/-0.7352746338d0,-0.1429046811d0,
     1    -0.6625326143d0/
	data (spheres(j,477),j=1,3)/-0.2376310390d0,-0.9680764874d0,
     1     0.0797458706d0/
	data (spheres(j,478),j=1,3)/-0.3910695158d0, 0.4981280230d0,
     1    -0.7739076860d0/
	data (spheres(j,479),j=1,3)/-0.5338834866d0,-0.7421145408d0,
     1    -0.4052584744d0/
	data (spheres(j,480),j=1,3)/ 0.9468496277d0,-0.3118695932d0,
     1    -0.0788234697d0/
	data (spheres(j,481),j=1,3)/-0.1147754307d0, 0.9536574506d0,
     1    -0.2781439689d0/
	data (spheres(j,482),j=1,3)/ 0.8133326032d0,-0.3250134121d0,
     1    -0.4825519230d0/
	data (spheres(j,483),j=1,3)/ 0.6736860990d0,-0.7347478107d0,
     1    -0.0793265067d0/
	data (spheres(j,484),j=1,3)/-0.5612991506d0, 0.7807935839d0,
     1    -0.2744169143d0/
	data (spheres(j,485),j=1,3)/ 0.0426952312d0, 0.9655347848d0,
     1    -0.2567483135d0/
	data (spheres(j,486),j=1,3)/-0.4370586154d0,-0.8538227779d0,
     1     0.2827833634d0/
	data (spheres(j,487),j=1,3)/-0.9724122983d0, 0.2310391096d0,
     1     0.0321753316d0/
	data (spheres(j,488),j=1,3)/-0.0884769772d0, 0.0984996001d0,
     1     0.9911960721d0/
	data (spheres(j,489),j=1,3)/ 0.5014969879d0, 0.8641003050d0,
     1     0.0427952568d0/
	data (spheres(j,490),j=1,3)/ 0.1887459007d0,-0.7203061241d0,
     1     0.6674833875d0/
	data (spheres(j,491),j=1,3)/ 0.3149425682d0, 0.8685049787d0,
     1    -0.3827666140d0/
	data (spheres(j,492),j=1,3)/-0.4562259536d0, 0.4653594792d0,
     1     0.7584843007d0/
	data (spheres(j,493),j=1,3)/ 0.4206916662d0,-0.8426370231d0,
     1     0.3361270135d0/
	data (spheres(j,494),j=1,3)/-0.9510463136d0, 0.1234327898d0,
     1    -0.2833288827d0/
	data (spheres(j,495),j=1,3)/ 0.1143789817d0,-0.5927603872d0,
     1    -0.7972155116d0/
	data (spheres(j,496),j=1,3)/-0.9655099655d0,-0.2509399551d0,
     1     0.0694236669d0/
	data (spheres(j,497),j=1,3)/ 0.2477902591d0, 0.3653354096d0,
     1     0.8972903800d0/
	data (spheres(j,498),j=1,3)/-0.2794831911d0, 0.8544975420d0,
     1     0.4378619607d0/
	data (spheres(j,499),j=1,3)/-0.2372863796d0, 0.7103664249d0,
     1     0.6626271322d0/
	data (spheres(j,500),j=1,3)/ 0.4106302720d0,-0.4688750867d0,
     1     0.7820095477d0/
	data (spheres(j,501),j=1,3)/ 0.5370512582d0, 0.7730025589d0,
     1     0.3377025170d0/
	data (spheres(j,502),j=1,3)/ 0.3422137810d0,-0.7081278102d0,
     1     0.6176121215d0/
	data (spheres(j,503),j=1,3)/ 0.0815733435d0,-0.3037441771d0,
     1    -0.9492551103d0/
	data (spheres(j,504),j=1,3)/ 0.8825896461d0,-0.3199145686d0,
     1     0.3445144196d0/
	data (spheres(j,505),j=1,3)/ 0.9366624306d0, 0.3475652452d0,
     1     0.0431496411d0/
	data (spheres(j,506),j=1,3)/-0.2682937502d0,-0.7324278391d0,
     1     0.6257538846d0/
	data (spheres(j,507),j=1,3)/ 0.0885696478d0,-0.8076566568d0,
     1    -0.5829632426d0/
	data (spheres(j,508),j=1,3)/-0.1673846925d0,-0.0446273995d0,
     1     0.9848810892d0/
	data (spheres(j,509),j=1,3)/-0.1041663230d0, 0.8969094702d0,
     1     0.4297706125d0/
	data (spheres(j,510),j=1,3)/ 0.8766786360d0,-0.4766736797d0,
     1    -0.0649366789d0/
	data (spheres(j,511),j=1,3)/-0.4059008313d0, 0.3460273298d0,
     1     0.8458780067d0/
	data (spheres(j,512),j=1,3)/ 0.3249788428d0,-0.0596769588d0,
     1    -0.9438365390d0/
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Buildlookup.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/29/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose: this subroutine builds the look up table for studying sphere interaction
c	See paper by LeGrand and Merz, J. Comp. Chem. 14, 349-352 (1993)
c
	subroutine buildlookup
c
	integer	npointmax,ndmax,nintmax,nthetamax,nphimax
c
	parameter	(npointmax = 512)
	parameter	(nintmax   = npointmax/32)
	parameter	(ndmax     = 200)
	parameter	(nthetamax = 200)
	parameter	(nphimax   = 200)
c
	integer	i,j,k,l
	integer	ndist,ntheta,nphi,npoint
	integer	kint,lbit,idist,ipoint
	integer	kint2(npointmax),lbit2(npointmax)
	integer	maskdist(npointmax,ndmax,nintmax)
	integer closestp(nthetamax,nphimax)
c
	real*8	xdist,xtheta,xphi,pi,dist,norm,distmin,phi,theta
	real*8	tdist(3)
	real*8	spheres(3,512)
c
	common /spherepos/	spheres
	common /lookup/		xdist,xtheta,xphi,kint2,lbit2,maskdist,
     1				closestp
c
	npoint = 512
c
	ndist  = 200
	ntheta = 200
	nphi   = 200
c
	xdist  = 0.01d0
	xtheta = 1.8d0
	xphi   = 0.9d0
c
	pi = acos(-1.d0)
c
c	1. Ensures that all points are on the surface of the sphere
c	(to correct for round-off errors)
c
	do 500 i = 1,npoint
		norm = 0
		do 300 k = 1,3
			norm = norm + spheres(k,i)**2
300		continue
		norm = sqrt(norm)
		do 400 k = 1,3
			spheres(k,i) = spheres(k,i)/norm
400		continue
500	continue
c
c	3. Build look-up table for distances :
c	For each point i and for each distance d (in between 0 and 2),
c	define all points that are within d of point i : store this
c	information in a mask table such that information is store
c	per bit
c
c	3.1 Initialise all bits to 0 : non exposed
c
	do 900 i = 1,npoint
		do 800 j = 1,ndist
			do 700 k = 1,npoint/32
				do 600 l = 0,31
					maskdist(i,j,k) = 
     1					ibset(maskdist(i,j,k),l)
600				continue
700			continue
800		continue
900	continue
c
c	3.2 Now define distance mask
c
	do 1300 i = 1,npoint
c
		do 1200 j = 1,npoint
c
			kint = j/32 + 1
			if(mod(j,32).eq.0) kint = kint - 1
			lbit = j - (kint-1)*32 - 1
c
			dist = 0
			do 1000 k = 1,3
				dist=dist+(spheres(k,i)-spheres(k,j))**2
1000			continue
			dist = sqrt(dist)
c
			idist = int(dist/xdist) + 1
			do 1100 k = idist,ndist
				maskdist(i,k,kint)=
     1				ibclr(maskdist(i,k,kint),lbit)
1100			continue
c
1200		continue
c
1300	continue
c
	do 1350 i = 1,npoint
		kint2(i) = i/32 + 1
		if(mod(i,32).eq.0) kint2(i) = kint2(i) - 1
		lbit2(i) = i - (kint2(i)-1)*32 - 1
1350    continue
c
c	4. Now define position look-up table : for each value of theta
c	and phi, find closest point on the sphere
c
	do 1800 i = 1,nphi
c
		phi = xphi/2 + (i-1)*xphi
		phi = phi*pi/180
c
		do 1700 j = 1,ntheta
c
			theta = xtheta/2 + (j-1)*xtheta
			theta = theta*pi/180
c
			tdist(1) = cos(theta)*sin(phi)
			tdist(2) = sin(theta)*sin(phi)
			tdist(3) = cos(phi)
c
			distmin = 0
			do 1400 k = 1,3
				distmin = distmin + (spheres(k,1)-
     1				tdist(k))**2
1400			continue
			ipoint = 1
			do 1600 k = 2,npoint
				dist = 0
				do 1500 l = 1,3
					dist = dist + (spheres(l,k)-
     1					tdist(l))**2
1500				continue
				if(dist.lt.distmin) then
					distmin = dist
					ipoint  = k
				endif
1600			continue
c
			closestp(i,j) = ipoint
c
1700		continue
c
1800	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Write_Surf.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine generates a Pymol script for displaying the 
c	accessible surface.
c	The surface is described by a set of point: each point is represented
c	as a small sphere
c
c	Input:
c
c		l_py		: logical unit number of Pymol file
c
	subroutine Write_Surf(l_py,Npoint,Point)
c
	include 'poisson.h'
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
	integer	i
	integer Npoint
	integer l_py
c
	real*8	Point(3,Npoint)
c
c	local variables
c
	real*8	Rad
c
1	format('from pymol.cgo import *')
2	format('from pymol import cmd',/)
3	format('Surface = [',/,'   COLOR, 0.0, 1.0, 0.0,',/,/)
5	format('   SPHERE,',f8.3,',',f8.3,',',f8.3,','f8.3,' ,')
7	format('   END',/,'   ]',/,/)
8	format('cmd.load_cgo(Surface,''Surface'')')
c
c ==========================================================================================
c = Write header for CGO pymol file ========================================================
c ==========================================================================================
c
	write(l_py,1)
	write(l_py,2)
c
c ==========================================================================================
c = Generate object for surface         ====================================================
c ==========================================================================================
c
c	Define object as green
c
	write(l_py,3)
c
c	Generate object
c
	Rad = 0.2d0
	do 100 i = 1,Npoint
		write(l_py,5) Point(1,i),Point(2,i),Point(3,i),Rad
100	continue
c
c	close object
c
	write(l_py,7)
c
c ==========================================================================================
c = Tell pymol to load the object that we have created  ========================================
c ==========================================================================================
c
	write(l_py,8)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Resize_point.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c       This subroutine resizes the vector containing the coordinates of the
c       points on the accessible surface
c
        subroutine resize_point(Npointmax)
c
        integer Npoint,Npointmax,i
c
        real*8  point1(*),point2(*)
c
        pointer (ptr_point1,point1)
        pointer (ptr_point2,point2)
c
        common /space/ ptr_point1,Npoint
c
        Npointmax = Npointmax*2
c
        ptr_point2=malloc(24*Npointmax)
c
        do 100 i = 1,3*Npoint
                point2(i) = point1(i)
100     continue
c
        call free(ptr_point1)
c
        ptr_point1 = ptr_point2
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	Clean_Surf.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c       This subroutine cleans the array for the point on the accessible surface
c
        subroutine Clean_Surf
c
        integer Npoint
c
        real*8  point1(*)
c
        pointer(ptr_point1,point1)
c
        common /space/ptr_point1,Npoint
c
        call free(ptr_point1)
c
        return
        end
