c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	AquaSol.f
c	(Generalized Poisson Boltzmann Solver)
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/4/07
c	Author:		Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose: the is a generalized Poisson Boltzmann solver:
c
c	We are solving the Generalized Non linear Poisson Boltzmann equation:
c
c	-Delta (eps(u(r),r) Delta(u(r)))+ kappa(r) f(u(r)) = C Sum_i zi d(r-ri)
c
c	Where:
c		u is the dimensionless electrostatic potential
c		(i.e. u(r) = electron * phi(r)/(kB T) )
c
c		Delta means Laplacian

c		Eps is the permittivity (dielectric constant) :
c			epsilon(r) = eps1 	inside the solute
c			epsilon(r) = eps2	inside the solvent
c
c		For all ions in the solution:
c		kappa is the Debye-Huckel coefficient that depends on the ionic
c		concentration I :
c
c		kappa = Ck *(298/T) * ion_conc * ion_z
c
c			ion_conc is the ionic concentration in mol/l
c			ion_val is the ion valence
c
c		f(u) :
c			f(u(r)) = - exp(-ion_z * u(r))
c
c			(note that for the linearized Poisson Boltzmann equation
c			f(u(r)) = -1 + ion_z*u(r)
c
c		C Sum_i zi d(r-r_i) is the charge source. The sum extends over
c		all charges i, placed at position i, of fractional value zi.
c
c	Note that this solver can solve:
c
c			The Laplace equation (setting source charges and ion concentrations to 0)
c			The Poisson equation (setting ion concentration to 0)
c			The Non linear Poisson Boltzmann equation
c			The size Modified Poisson Boltzmann equation (SMPB)
c			the Generalized Poisson Boltzmann Langevin equation (GPBL)
c
c	At this stage, it solves the equation using a finite volume method.
c	In the future, we will implement a Finite Element Method.
c
c ==========================================================================================
c = poisson.h defines array sizes for AquaSol    ===========================================
c ==========================================================================================
c
	include 'poisson.h'
c
c ==========================================================================================
c = Declares all local/global arrays          ==============================================
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i,j
	integer	n_arg
	integer	ierr,iflag,ishift,idx,idx2
	integer	l_log,l_py,l_pdb
	integer	surf_type,flag_pymol,flag_box,flag_distrib,flag_eps
	integer	flag_surf,flag,flag_dip,flag_psi2
	integer	grid_type,bnd_type
	integer	ion_zone
	integer	nx,ny,nz
	integer	iener
        integer*8 r_map
        integer r_work
        integer nnode,nfudge,nsize,nsize_byte
        integer n_iter,niter_max,itmax
        integer method,equation,ndiscrete,nlevel,nlin,ijac,nsmooth
        integer zero,rwork_size2,imethod,info
	integer	flag_equa
c
        integer*8 rwork_size
        integer   iwork_size
        integer*8       rwork_byte,ival
c
        real*8  res,res_c,resnum,resnum_old,resden,rmem,tol,lambda
	real*8	xnorm1
	real*8	R_probe,R_ion
	real*8	eps1,eps2
        real*8  r_tot
	real*8	shiftx,shifty,shiftz
	real*8	c_mol(3),c_grid(3)
	real*8 nwater
c
        character       filetype*11
c
c	Constants
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
	character	fname_pdb*100,fname_log*100,fname_py*100
	character	fname_input*100,fname_newpdb*100
        character       fname_ene*100
	character	fpdbcenter*100
c
c	Information on molecule
c ==========================================================================================
c
        integer natom,ncharge,iform
c
        real*8  atom_coord(*),atom_radius(*)
        real*8  atom_charge(*)
c
        character  atom_label(*)*30,atom_label2(*)*12
c
c	Information on maps over the grid:
c ==========================================================================================
c
c	eps_X,eps_Y,eps_Z:	dielectric constants
c	crg_map:		charge map
c	gxcf,gycf,gzcf:		boundary conditions
c	helmoltz_map:		region in which helmoltz term is non zero
c
	real*8	eps_X(*)
	real*8	eps_Y(*)
	real*8	eps_Z(*)
	real*8	crg_map(*)
	real*8	helmoltz_map(*)
	real*8	gxcf(*)
	real*8	gycf(*)
	real*8	gzcf(*)
c
        real*8  resid(*)
        real*8  space(*)
        real*8  u_old(*)
c
        real*8  rwork(*)
        real*8  iwork(*)
c
	real*8	psi(*)
	real*8	psi_old(*)
	real*8	sinhu(*)
c
        integer iparam(100)
        real*8  rparam(100)
c
c       solution of the equation and energy values
c ==========================================================================================
c
	real*8	ener,free_ene
	real*8	free_ene1,free_ene2,free_ene3
        real*8  ene_atom(*)
c
	real*8	u(*)
c
c	Information on grid
c ==========================================================================================
c
	real*8	sizex,sizey,sizez
	real*8	hx(*),hy(*),hz(*)
        real*8  xf(*),yf(*),zf(*)
c
	integer*1 grid_partition(*)
c
	integer*1 grid_shift(*)
c
c       Information on slab (if any)
c ==========================================================================================
c
        integer flag_slab
c
        real*8  z0,zh,eps_slab
c
c	Output maps
c ==========================================================================================
c
	integer		l_map,map_type
	integer		flag_phi,flag_edens,flag_nion,flag_dion
	integer		flag_dipole,flag_map_format,flag_psi
c
	real*8		origin(3)
c
	character	fname_out_phi*100,fname_out_edens*100
	character	fname_out_nion*100,fname_out_dion*100
        character       fname_out_dipole*100,fname_out_psi*100
c	
c       Pointers
c ==========================================================================================
c
        pointer (ptr_eps_X,eps_X)
        pointer (ptr_eps_Y,eps_Y)
        pointer (ptr_eps_Z,eps_Z)
        pointer (ptr_crg_map,crg_map)
        pointer (ptr_helmoltz_map,helmoltz_map)
        pointer (ptr_u,u)
        pointer (ptr_grid_partition,grid_partition)
        pointer (ptr_grid_shift,grid_shift)
        pointer (ptr_hx,hx)
        pointer (ptr_hy,hy)
        pointer (ptr_hz,hz)
        pointer (ptr_xf,xf)
        pointer (ptr_yf,yf)
        pointer (ptr_zf,zf)
        pointer (ptr_gxcf,gxcf)
        pointer (ptr_gycf,gycf)
        pointer (ptr_gzcf,gzcf)
        pointer (ptr_resid,resid)
        pointer (ptr_u_old,u_old)
        pointer (ptr_space,space)
        pointer (ptr_rwork,rwork)
        pointer (ptr_iwork,iwork)
c
	pointer	(ptr_psi,psi)
	pointer (ptr_psi_old,psi_old)
	pointer (ptr_sinhu,sinhu)
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        pointer (ptr_ene,ene_atom)
c
c	Common blocks
c ==========================================================================================
c
	common /input_file/	fname_pdb,fname_log,fname_py
	common /input_par/	R_probe,R_ion,flag_pymol,surf_type,
     1				grid_type,ion_zone,flag_box,
     2				flag_distrib,flag_eps,bnd_type
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,
     2                          natom,ncharge
	common /elec_const/	eps1,eps2
c
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1			   	Avogadro,eps0,beta,C_charge,C_prop
	common /molshift/	shiftx,shifty,shiftz
	common /centers/ 	c_mol,c_grid
        common /slab/           z0,zh,eps_slab,flag_slab
	common /log_file/	l_log
        common /pdesolver/      tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
	common /output_map/	fname_out_phi,fname_out_edens,
     1				fname_out_nion,fname_out_dion,
     2                          fname_out_dipole,fname_out_psi,
     3				flag_phi,flag_edens,flag_nion,flag_dion,
     4				flag_dipole,flag_map_format,flag_psi
c
	common /psi_pot/	ptr_psi,flag_psi2
	common /numwater/	nwater
	common /centering/	fpdbcenter
c
        niter_max = 100
        lambda    = 1.0
c
c
c ==========================================================================================
c = Input/Output formats       =============================================================
c ==========================================================================================
c
1	format(a)
2	format(1x,'Input parameter file                          : ',$)
3	format(70('='),/,2('='),'Solution: energetics and maps ',36x,
     1  2('='),/,70('='),/)  
4	format('Grid energy  : ',f15.5,' Kcal/mol   (',f15.5,' kJ )')
5	format(70('='),/,2('='),'Solving the elliptic equation ',36x,
     1  2('='),/,70('='),/)  
6       format('Memory requirement (MBytes) : ',f12.3)
7       format(2x,i5,15x,1pe12.5,8x,1pe12.5,8x,1pe12.5)
8	format('Free energy (density)   : ',f15.5,' Kcal/mol   (',
     1  f15.5,' kJ )')
9	format('Free energy (charges)   : ',f15.5,' Kcal/mol   (',
     1  f15.5,' kJ )')
10	format('Free energy (ions)      : ',f15.5,' Kcal/mol   (',
     1  f15.5,' kJ )')
11	format('Free energy (dip+ions)  : ',f15.5,' Kcal/mol   (',
     1  f15.5,' kJ )')
12	format('Free energy (total)     : ',f15.5,' Kcal/mol   (',
     1  f15.5,' kJ )')
13	format('Number of water molecule: ',f25.5)
c
c ==========================================================================================
c = Read all input information =============================================================
c ==========================================================================================
c
	n_arg = iargc()
c
	if(n_arg.eq.0) then
		write(6,2)
		read(5,1) fname_input
	else
		call getarg(1,fname_input)
	endif
c
	call read_input(fname_input,l_log,ierr,nx,ny,nz,sizex,sizey,
     1                  sizez)
	flag	  = flag_distrib
c
	if(ierr.eq.1) then
		write(6,*) ' '
		write(6,*) 'Problem with Parameter file:',fname_input(1:
     1			index(fname_input,' ')-1)
		write(6,*) ' '
		stop
	endif
c
        if(flag_map_format.eq.5) then
                filetype = 'unformatted'
        else
                filetype = 'formatted'
        endif
c
	flag_equa = equation
c
c ==========================================================================================
c = Allocate memory to all maps   ==========================================================
c ==========================================================================================
c
        nnode = nx*ny*nz
        nfudge = 100
        nsize  = nnode + nnode/6 + nfudge
        nsize_byte = 8*nsize
c
        ptr_eps_X               = malloc(nsize_byte)
        ptr_eps_Y               = malloc(nsize_byte)
        ptr_eps_Z               = malloc(nsize_byte)
        ptr_crg_map             = malloc(nsize_byte)
        ptr_helmoltz_map        = malloc(nsize_byte)
        ptr_u                   = malloc(nsize_byte)
c
        ptr_grid_partition      = malloc(nsize)
        ptr_grid_shift          = malloc(5*nsize)
c
        ptr_hx                  = malloc(8*nx)
        ptr_hy                  = malloc(8*ny)
        ptr_hz                  = malloc(8*nz)
c
        ptr_xf                  = malloc(8*5*nx)
        ptr_yf                  = malloc(8*5*ny)
        ptr_zf                  = malloc(8*5*nz)
c
        ptr_gxcf                = malloc(8*5*ny*nz)
        ptr_gycf                = malloc(8*5*nx*nz)
        ptr_gzcf                = malloc(8*5*nx*ny)
c
        ptr_space               = malloc(8*nx*ny)
        ptr_resid               = malloc(8*nnode)
        ptr_u_old               = malloc(8*nnode)
c
        r_map = nsize_byte
        r_map = 6*r_map
        r_map = r_map + 6*nsize
        r_map = r_map + 8*6*(nx+ny+nz) + 
     1                  8*5*(ny*nz+nx*nz+nx*ny)
c
c ==========================================================================================
c = Set all constants          =============================================================
c ==========================================================================================
c
	call Set_Constants
c
c ==========================================================================================
c = Read charges               =============================================================
c ==========================================================================================
c
	ierr = 0
c
	call Read_Charge(fname_pdb,fpdbcenter,l_log,ierr)
	if(ierr.eq.1) then
		write(6,*) ' '
		write(6,*) 'Problem with PDB file:',fname_pdb(1:
     1			index(fname_pdb,' ')-1)
		write(6,*) 'Please check !!'
		write(6,*) ' '
		stop
	elseif(ierr.eq.2) then
		write(6,*) ' '
		write(6,*) 'Problem with PDB file:',fpdbcenter(1:
     1			index(fpdbcenter,' ')-1)
		write(6,*) 'Please check !!'
		write(6,*) ' '
		stop
	endif
c
c ==========================================================================================
c = Generate Box               =============================================================
c ==========================================================================================
c
	write(6,*) '	Define grid...'
	call Define_Grid(grid_type,l_log,bnd_type,hx,hy,hz,sizex,sizey,
     1                  sizez,nx,ny,nz)
c
c ==========================================================================================
c = If a map is written is XPLOR or EZD format, generate centered PDB file =================
c ==========================================================================================
c
	if(flag_map_format.eq.3.or.flag_map_format.eq.4) then
c
	write(6,*) '	Recenter molecule...'
		ishift = 0
		idx = index(fname_pdb,'.') -1
		idx2 = index(fname_pdb,' ') -1
		fname_newpdb = fname_pdb(1:idx)//'_center'//
     1		fname_pdb(idx+1:idx2)
		write(6,*) 
     1		'		PDB file for centered molecule :',
     2		fname_newpdb(1:index(fname_newpdb,' ')-1)
		l_pdb = 14
		open(unit=l_pdb,file=fname_newpdb,status='unknown')
		call Write_Mol_Center(l_pdb)
		close(unit=l_pdb)
c
	else
c
		ishift = 1
c
	endif
c
c ==========================================================================================
c = Generate accessible surface    ==========================================================
c ==========================================================================================
c
	write(6,*) '	Generate points on accessible surface'
	if(surf_type.eq.1) call GenAccSurf(R_probe)
c
c ==========================================================================================
c = Generate Surface           =============================================================
c ==========================================================================================
c
	write(6,*) '	Generate surface...'
c
	iflag = 1
c
	call GenSurf(surf_type,R_probe,grid_partition,hx,hy,hz,nx,ny,nz)
	call compute_vol_surf(l_log,surf_type,grid_partition,hx,hy,hz,
     1			nx,ny,nz)
c
c
c ==========================================================================================
c = Generate ion-excluded zone, if needed ==================================================
c ==========================================================================================
c
	if(ion_zone.eq.1) then
		write(6,*) '	Generate Stern (ion-excluded) region...'
		call GenIonZone(R_ion,grid_partition,hx,hy,hz,nx,ny,nz)
	endif
        if(flag_slab.eq.1) call IonSlab(z0,zh,grid_partition,hz,
     1                          nx,ny,nz)
c
c ==========================================================================================
c = Write grid in pymol file (CGO) =========================================================
c ==========================================================================================
c
c	flag_pymol = 0
	if(flag_pymol.eq.1) then
	        write(6,*) '	Write pymol script...'
		l_py = 13
		open(unit=l_py,file=fname_py,status='unknown')
		flag_surf = 1
		call Write_Grid(l_py,ion_zone,flag_box,flag_surf,
     1		ishift,grid_partition,hx,hy,hz,nx,ny,nz)
		close(unit=l_py)
	endif
c
c ==========================================================================================
c = Define Shifted projections of molecule on grid =========================================
c ==========================================================================================
c
	write(6,*) '	Shifted projections...'
c
	call shifted_grids(nx,ny,nz,hx,hy,hz,grid_partition,grid_shift)
c
c ==========================================================================================
c = Setup is finished; clean some arrays if needed        ==================================
c ==========================================================================================
c
        if(surf_type.eq.1) call Clean_Surf
c
c ==========================================================================================
c = Now solve the NLPB equation using routines from MG package  ============================
c ==========================================================================================
c
c	Initialize potential to 0
c
	do 100 i = 1,nx*ny*nz
		u(i) = 0
100	continue
	if(fname_out_phi.ne.' ') then
		call readmap_bin(nx,ny,nz,fname_out_phi,u,ierr)
		if(ierr.eq.1) then
			write(6,*) 'Could not read input file for phi'
			write(6,*) 'The program will use phi=0'
		endif
	endif
c
	write(6,*) ' '
	write(6,*) 
     1		'Solve Elliptic equation for electrostatic potential :'
	write(6,*) ' '
c
	write(l_log,*) ' '
	write(l_log,5)
c
c ==========================================================================================
c == Get size of storage arrays and create them =========================================
c ==========================================================================================
c
        nfudge = 100
        imethod = method
        zero = 0
c
        if(imethod.le.3) then
c
                itmax = 100
                info = 1
c
c               Set Helmholtz map
c
                call set_helmoltz(helmoltz_map,grid_shift,nx,ny,nz)
c
c               Define charge distrib
c
                call charge_distrib(crg_map,grid_shift,hx,
     1                hy,hz,nx,ny,nz)
c
c               Set boundaries
c
                call set_boundary(gxcf,gycf,gzcf,hx,hy,hz,nx,ny,nz)
c
c               Define epsilon maps
c
	        call set_epsilon(eps_X,eps_Y,eps_Z,grid_shift,hx,hy,hz,
     1          nx,ny,nz)
c
        else
                itmax = 1
                info = 0
                equation = 2
                method = 1
        endif
c
        call GridSize(equation,method,ijac,ndiscrete,zero,zero,
     1           nx,ny,nz,nlevel,rwork_size,iwork_size)
c
        rwork_size = rwork_size + nfudge
        rwork_size2 = rwork_size
c
        rwork_byte = rwork_size
        ival = 8
        rwork_byte = ival*rwork_byte
c
        ptr_rwork=malloc(rwork_byte)
        ptr_iwork=malloc(4*iwork_size)
c
        if(ptr_rwork.eq.0) then
                write(6,*) 'ptr_rwork :',ptr_rwork
                write(6,*) '# of Mbytes for ptr_rwork :',
     1             rwork_byte/1.d6
                write(6,*)'Could not allocate memory needed by Solver'
                write(6,*) 'You need to resize the grid'
                write(6,*) ' '
                stop
        endif
c
        call PDE_init(iparam,rparam,nx,ny,nz,hx,hy,hz,xf,yf,zf,
     1                 itmax,rwork_size2,iwork_size,info)
c
	if(flag_equa.le.4) then
c
		if(imethod.le.3) then
c
	        	call solve_pde(grid_shift,eps_X,eps_Y,eps_Z,
     1                  helmoltz_map,crg_map,gxcf,gycf,gzcf,
     2                  xf,yf,zf,nx,ny,nz,u,rwork,iwork,
     3                  rparam,iparam)
c
			free_ene1 = 0.d0
			free_ene2 = 0.d0
			free_ene3 = 0.d0
			free_ene  = 0.d0
c
			call charge_distrib(crg_map,grid_shift,
     1			hx,hy,hz,nx,ny,nz) 
	        	call set_epsilon(eps_X,eps_Y,eps_Z,grid_shift,
     1			hx,hy,hz,nx,ny,nz)
c
			call energy_elecdensity(u,hx,hy,hz,nx,ny,nz,
     1			eps_X,eps_Y,eps_Z,free_ene1)
			call energy_fixedcharges(u,crg_map,free_ene2,
     1			hx,hy,hz,nx,ny,nz)
			call energy_mobileions(flag_equa,u,grid_shift,
     1			free_ene3,hx,hy,hz,nx,ny,nz)
c
			free_ene = free_ene1 + free_ene2 + free_ene3
c
        	else
c
	        	call solve_sc(grid_shift,eps_X,eps_Y,eps_Z,
     1                  helmoltz_map,crg_map,gxcf,gycf,gzcf,
     2                  xf,yf,zf,nx,ny,nz,u,rwork,iwork,
     3                  rparam,iparam,hx,hy,hz,resid,u_old,space,
     4                  free_ene1,free_ene2,free_ene3)
c
			free_ene = free_ene1 + free_ene2 + free_ene3
c
			equation = 4
c
		endif
c
	elseif(flag_equa.eq.5) then
c
		ptr_psi = malloc(8*nnode)
		ptr_psi_old = malloc(8*nnode)
		ptr_sinhu   = malloc(8*nnode)
c
		call solve_YULP(grid_shift,eps_X,eps_Y,eps_Z,
     1		helmoltz_map,crg_map,gxcf,gycf,gzcf,xf,yf,zf,
     2		nx,ny,nz,u,rwork,iwork,rparam,iparam,hx,hy,hz,
     3		resid,u_old,psi_old,sinhu,space)
c
		equation = 4
c
        endif
c
c ==========================================================================================
c = Compute solvation    ===================================================================
c ==========================================================================================
c
	write(l_log,*)' '
	write(l_log,3)
c
	write(6,*) 'computing energy'
c
c	Grid energy:
c		$GE = \Frac{1}{2} \Sum_{i,j,k}  crg(i) U(i)$
c
c	I implemented two methods for computing the grid energy:
c		- a "volumic" approach, that loops over all grid points
c		- an "atom-based" approach, that loops over all charges,
c		  project them on the grid (the same way the crg map was derived)
c		  and add its charge contribution to the energy
c	The second method should be faster; keep the first method available...
c
	iener = 1
	if(iener.eq.1) then
c
c		Need to re-initialise crg map, as it was overwritten by the PDE solver
c
c	        call GenSurf(surf_type,R_probe,grid_partition,hx,hy,hz,
c     1                  nx,ny,nz)
		call charge_distrib(crg_map,grid_shift,hx,hy,hz,
     1                  nx,ny,nz) 
		call compute_grid_energy_vol(u,crg_map,ener,hx,hy,hz,
     1                  nx,ny,nz)
c
	else
c
                ptr_ene = malloc(8*natom)
c
		if(flag_distrib.eq.3) flag_distrib = 0
		call compute_grid_energy_atom(u,ener,ene_atom,hx,hy,hz,
     1                  nx,ny,nz)
c		idx = index(fname_pdb,'.') -1
c		fname_ene = fname_pdb(1:idx)//'_energy.res'
c                l_pdb = 14
c                open(unit=l_pdb,file=fname_ene,status='unknown')
c                call Write_Ene_Atom(l_pdb,ene_atom)
c                close(unit=l_pdb)
c
                call free(ptr_ene)
	endif
c
c
	write(6,*) ' '
	write(6,*) 'Grid energy :', ener/4.1868d0,' Kcal/mol (',
     1		ener,' kJ )'
	write(6,*) ' '
	write(6,8) free_ene1,free_ene1*4.1868
	write(6,9) free_ene2,free_ene2*4.1868
        if(flag_equa.le.3) then
		write(6,10) free_ene3,free_ene3*4.1868
	elseif(flag_equa.eq.4) then
		write(6,11) free_ene3,free_ene3*4.1868
        endif
	write(6,12) free_ene,free_ene*4.1868
	write(l_log,*) ' '
	write(l_log,*)' '
	write(l_log,8) free_ene1,free_ene1*4.1868
	write(l_log,9) free_ene2,free_ene2*4.1868
        if(flag_equa.le.3) then
		write(l_log,10) free_ene3,free_ene3*4.1868
	elseif(flag_equa.eq.4) then
		write(l_log,11) free_ene3,free_ene3*4.1868
        endif
	write(l_log,12) free_ene,free_ene*4.1868
	write(l_log,*)' '
c
c ==========================================================================================
c = Total Memory size required for calculation  ============================================
c ==========================================================================================
c
        r_tot = r_map/1.d6 + 8*(rwork_size/1.d6)
	if(flag_distrib.eq.3) then 
		r_tot = r_tot + 8*(rwork_size/1.d6)
	endif
c
        write(6,*) ' '
        write(6,*) 'Total memory requirement (MBytes) :',r_tot
        write(6,*) ' '
c
	write(l_log,*)' '
	write(l_log,6) r_tot
	write(l_log,*)' '
c
c ==========================================================================================
c = Generate maps and write in file          ===============================================
c ==========================================================================================
c
	if(flag_phi.eq.1.or.flag_edens.eq.1.or.flag_nion.eq.1.or.
     1	flag_dion.eq.1.or.flag_dipole.eq.1) then
		write(6,*) 
     1		'Write grid maps  :'
		write(6,*) ' '
	endif
c
	Origin(1) = c_mol(1) - sizex*dble(nx/2) - shiftx
	Origin(2) = c_mol(2) - sizey*dble(ny/2) - shifty
	Origin(3) = c_mol(3) - sizez*dble(nz/2) - shiftz
c
c	We use crg_map as temporary storage of the output map
c       and gzcf as temporary storage for internal calculation
c
	if(flag_phi.eq.1) then
c
c		Electrostatic potential
c
		write(6,*) 
     1		'	Write electrostatics potential map in file  : ',
     2		fname_out_phi(1:30)
		map_type = 1
		l_map = 13
		open(unit=l_map,file=fname_out_phi,status='unknown',
     1          form=filetype)
c
		call Elec_Map(u,hx,hy,hz,nx,ny,nz,map_type,crg_map,
     1			l_map,flag_map_format,Origin,grid_shift,gzcf)
c
		close(unit=l_map)
c
	endif
c
	if(flag_edens.eq.1) then
c
c		Energy density
c
		write(6,*) 
     1		'	Write electrostatics energy density in file : ',
     2		fname_out_edens(1:30)
c
		map_type = 2
		l_map = 13
		open(unit=l_map,file=fname_out_edens,status='unknown',
     1          form=filetype)
c
		call Elec_Map(u,hx,hy,hz,nx,ny,nz,map_type,crg_map,
     1			l_map,flag_map_format,Origin,grid_shift,gzcf)
c
		close(unit=l_map)
c
	endif
c
	if(flag_nion.eq.1) then
c
c		Ion density (number)
c
		write(6,*) 
     1		'	Write ion density (number) in file          : ',
     2		fname_out_nion(1:30)
c
		map_type = 3
		l_map = 13
		open(unit=l_map,file=fname_out_nion,status='unknown',
     1          form=filetype)
c
		call Elec_Map(u,hx,hy,hz,nx,ny,nz,map_type,crg_map,
     1			l_map,flag_map_format,Origin,grid_shift,gzcf)
c
		close(unit=l_map)
c
	endif
c
	if(flag_dion.eq.1) then
c
c		Ion density (charge)
c
		write(6,*) 
     1		'	Write ion density (charge) in file          : ',
     2		fname_out_dion(1:30)
c
		map_type = 4
		l_map = 13
		open(unit=l_map,file=fname_out_dion,status='unknown',
     1          form=filetype)
c
		call Elec_Map(u,hx,hy,hz,nx,ny,nz,map_type,crg_map,
     1			l_map,flag_map_format,Origin,grid_shift,gzcf)
c
		close(unit=l_map)
c
	endif
c
	if(flag_dipole.eq.1) then
c
c		Water density (dipole)
c
		if(flag_equa.le.3) then
			write(6,*) 'No water density map'
			goto 200
		endif
c
		write(6,*) 
     1		'	Write water density (dipole) in file        : ',
     2		fname_out_dipole(1:30)
c
		map_type = 5
		l_map = 13
		open(unit=l_map,file=fname_out_dipole,status='unknown',
     1          form=filetype)
c
		call Elec_Map(u,hx,hy,hz,nx,ny,nz,map_type,crg_map,
     1			l_map,flag_map_format,Origin,grid_shift,gzcf)
c
		close(unit=l_map)
c		
		write(6,13) nwater
		write(l_log,13) nwater
c
	endif
c
200	continue
	write(6,*) ' '
c
c ==========================================================================================
c = Close log file, clear memory and stop program ======================================================
c ==========================================================================================
c
	close(unit=l_log)
c
        call free(ptr_eps_X)
        call free(ptr_eps_Y)
        call free(ptr_eps_Z)
        call free(ptr_crg_map)
        call free(ptr_helmoltz_map)
        call free(ptr_u)
        call free(ptr_grid_partition)
        call free(ptr_grid_shift)
        call free(ptr_hx)
        call free(ptr_hy)
        call free(ptr_hz)
        call free(ptr_xf)
        call free(ptr_yf)
        call free(ptr_zf)
        call free(ptr_gxcf)
        call free(ptr_gycf)
        call free(ptr_gzcf)
c
        call free(ptr_rwork)
        call free(ptr_iwork)
        call free(ptr_resid)
        call free(ptr_u_old)
c
        if(flag_equa.eq.5) then
	        call free(ptr_psi)
	        call free(ptr_psi_old)
	        call free(ptr_sinhu)
        endif
c
        call Clean_Charges
c
	stop
	end
