c ==========================================================================================
c ==========================================================================================
c	Energy_fixedcharges.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	8/17/07
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	Compute the free energy for the fixed charges, using:
c
c			E = Sum_{i,j,k} phi(i,j,k) q(i,j,k)
c
c			where phi is the electrostatic potential
c			      q is the charge
c
c	This version just loops over all grid points
c
c	Input:
c		u : dimensioneless potential
c	Output:
c		ener: grid energy
c
c	Units consideration:
c
c	u is the dimensionless electrostatic potential, i.e.:
c
c	u = ec * phi
c	    --------
c            kB T
c
c	hence phi = kB*T*u/ec, and phi*q = kB * T * u *z
c	where z is the partial charge in electron unit.
c
c	The product u*z gives the energy in kB * T units.
c
c	Note that the charge on the grid are stored in crg_map, and have been
c	multiplied by C_charge; we need to remove this factor
c
	subroutine energy_fixedcharges(u,crg_map,ener,hx,hy,hz,nx,ny,nz)
c
c ==========================================================================================
c = poisson.h defines array sizes             ==============================================
c ==========================================================================================
c
	include 'poisson.h'
c
c ==========================================================================================
c = Declares all local/global arrays          ==============================================
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
c
c	Parameters
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
c	Information on grid
c ==========================================================================================
c
	real*8	dx,dy,dz
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Information on maps over the grid:
c ==========================================================================================
c
c	Fixed charge map
c
	real*8	crg_map(nx,ny,nz)
c
c       solution of the equation and energy values
c ==========================================================================================
c
	real*8	ener
c
	real*8	u(nx,ny,nz)
c
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
c
c ==========================================================================================
c = Compute energy             =============================================================
c ==========================================================================================
c
	ener = 0
c!!!!! replaced next line 30/01/12
c	do 300 k = 1,nx
	do 300 k = 1,nz
		dz = hz(k+1)-hz(k)
		do 200 j = 1,ny
			dy = hy(j+1)-hy(j)
c!!!!! replaced next line 30/01/12
c			do 100 i = 1,nz
			do 100 i = 1,nx
				dx = hx(i+1)-hx(i)
				ener = ener + u(i,j,k)*
     1				crg_map(i,j,k)*dx*dy*dz
100			continue
200		continue
300	continue
c
	ener = ener/C_charge
c
c	ener above is in KB*T unit: converts in kJ:
c
	ener = KB*Temp*Avogadro*(1.d-3)*ener
c
c	Now convert in kcal/mol
c
	ener = ener/4.1868
c
c	write(6,*) 'Fixed charge energy        :',ener,
c     1	' (Kcal/Mol)  (or ',ener*4.1868,' (kJ/Mol)'
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Energy_elecdensity.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine computes the total electrostatic energy of the grid
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c
        subroutine energy_elecdensity(u,x,y,z,nx,ny,nz,eps_X,eps_Y,
     1		eps_Z,ene)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	integer	nx,ny,nz
        integer i,j,k
c
        real*8  eps,eps1,eps2,ene,fact,fact2
	real*8	dx,dy,dz,valx,valy,valz,dvol
c
        real*8  u(nx,ny,nz)
c
c	Information on grids
c ==========================================================================================
c
	real*8 	x(nx),y(ny),z(nz)
	real*8	eps_X(nx,ny,nz),eps_Y(nx,ny,nz),eps_Z(nx,ny,nz)
c
	real*8  l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8  C_charge,C_prop
c
        common /elec_const/     eps1,eps2
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
c
	fact = -1.d0/C_charge
c
c
	fact = fact*KB*Temp*Avogadro*(1.d-3)/(2*4.1868d0)
c
	ene = 0
c
        do 300 k = 1,nz-1
c
		dz = z(k+1)-z(k)
                do 200 j = 1,ny-1
			dy = y(j+1)-y(j)
                        do 100 i = 1,nx-1
				dx = x(i+1)-x(i)
				valx = (u(i+1,j,k)-u(i,j,k))/dx
				valy = (u(i,j+1,k)-u(i,j,k))/dy
				valz = (u(i,j,k+1)-u(i,j,k))/dz
				valx = eps_X(i,j,k)*valx*valx
				valy = eps_Y(i,j,k)*valy*valy
				valz = eps_Z(i,j,k)*valz*valz
				dvol = dx*dy*dz
				ene = ene + (valx+valy+valz)*
     1				dvol
100                     continue
200             continue
300     continue
c
	ene = fact*ene
c
c	write(6,*) 'Electrostatic grid energy  :',ene,
c     1	' (Kcal/Mol)  (or ',ene*4.1868,' (kJ/Mol)'
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	Energy_elecdensity0.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine computes the total electrostatic energy of the grid,
c	assuming constant eps over the whole grid
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c
        subroutine energy_elecdensity0(u,x,y,z,nx,ny,nz,ene)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	integer	nx,ny,nz
        integer i,j,k
c
        real*8  eps,eps1,eps2,ene,fact,fact2
	real*8	dx,dy,dz,valx,valy,valz,dvol
c
        real*8  u(nx,ny,nz)
c
c	Information on grids
c ==========================================================================================
c
	real*8 	x(nx),y(ny),z(nz)
c
	real*8  l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8  C_charge,C_prop
c
        common /elec_const/     eps1,eps2
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
c
	fact = -1.d0/C_charge
c
	fact = fact*KB*Temp*Avogadro*(1.d-3)/(2*4.1868d0)
c
	ene = 0
c
        do 300 k = 1,nz-1
c
		dz = z(k+1)-z(k)
                do 200 j = 1,ny-1
			dy = y(j+1)-y(j)
                        do 100 i = 1,nx-1
				dx = x(i+1)-x(i)
				valx = (u(i+1,j,k)-u(i,j,k))/dx
				valy = (u(i,j+1,k)-u(i,j,k))/dy
				valz = (u(i,j,k+1)-u(i,j,k))/dz
				valx = eps1*valx*valx
				valy = eps1*valy*valy
				valz = eps1*valz*valz
				dvol = dx*dy*dz
				ene = ene + (valx+valy+valz)*
     1				dvol
100                     continue
200             continue
300     continue
c
	ene = fact*ene
c
c	write(6,*) 'Electrostatic grid energy  :',ene,
c     1	' (Kcal/Mol)  (or ',ene*4.1868,' (kJ/Mol)'
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	Energy_mobileions.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	8/17/07
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	Compute the free energy for the mobile ions, using:
c
	subroutine energy_mobileions(ieq,u,solvent,ener,hx,hy,hz,
     1			nx,ny,nz)
c
c ==========================================================================================
c = poisson.h defines array sizes             ==============================================
c ==========================================================================================
c
	include 'poisson.h'
c
c ==========================================================================================
c = Declares all local/global arrays          ==============================================
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,ieq
	integer	nx,ny,nz
c
	real*8	fact,val,sum,ener
c
c	Parameters
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
	real*8  exp_min,exp_max
	parameter (exp_min=-80.d0, exp_max = 80.d0)
c
c	Information on grid
c ==========================================================================================
c
	real*8	dx,dy,dz
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Information on maps over the grid:
c ==========================================================================================
c
	real*8	u(nx,ny,nz)
	integer*1	solvent(nx,ny,nz)
c
c	Information on ions
c ==========================================================================================
c
	integer	n_ion,n_zz
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)

c
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
c
	fact = Avogadro*1.0d-27
c
	ener = 0
c
	if(n_ion.eq.0) return
c
c!!!!! next line replaced 30/01/12
c	do 300 k = 1,nx
	do 300 k = 1,nz
		dz = hz(k+1)-hz(k)
		do 200 j = 1,ny
			dy = hy(j+1)-hy(j)
c!!!!! next line replaced 30/01/12
c			do 100 i = 1,nz
			do 100 i = 1,nx
				dx = hx(i+1)-hx(i)
				if(ibits(solvent(i,j,k),1,1).ne.0) then
					sum = 0.d0
					do 50 l = 1,n_ion
						val = -ion_z(l)*u(i,j,k)
						if(val.gt.exp_max) 
     1							val = exp_max
						if(val.lt.exp_min) 
     1							val = exp_min
						if(ieq.eq.1) then
							val = val*val/2
						else
							val = exp(val)-1
						endif
						sum = sum + ion_conc(l)
     1							*val
50					continue
					ener = ener + sum*dx*dy*dz
				endif
100			continue
200		continue
300	continue
c
	ener = ener*fact
c
c	ener above is in KB*T unit: converts in kJ:
c
	ener = KB*Temp*Avogadro*(1.d-3)*ener
c
c	Now convert in kcal/mol
c
	ener = ener/4.1868
c
c	write(6,*) 'Mobile ion energy          :',ener,
c     1	' (Kcal/Mol)  (or ',ener*4.1868,' (kJ/Mol)'
c	write(6,*) ' '
c
	return
	end
c========================================================================================
c========================================================================================
c
c	Langevin_free_energy.f
c
c	This subroutine computes the derivatives of the solvation free energy with
c	respect to the probabilities of the rotamers of each residue in the
c	protein
c
c========================================================================================
c========================================================================================
c
	subroutine langevin_free_energy(u,ener,solvent,hx,hy,hz,
     1			nx,ny,nz)
c
	include 'poisson.h'
c
	integer	i,j,k,info1
	integer	nx,ny,nz
	integer	jrot
	integer	n_solv,n_zz,i_ion,n_ion
c
	integer*1	solvent(nx,ny,nz)
c
	real*8	l_Bjerrum,pi,Temp,KB,KB_cal,electron,Ang,Avogadro
	real*8	eps0,beta,C_charge,C_prop,ener
	real*8	dx,dy,dz
	real*8	der,val,Ex,Ey,Ez,E2,E
	real*8	vol,p0_e,coef_l
	real*8	sol,psi_coef,exp_mu
	real*8	d,dm1,g,phi_val,u_val,arg,exp_m,F1_u,sinhc_u
	real*8	coef,l_dip,phi0,fact,coef_Z0,ratio
c
	real*8  exp_min,exp_max
	parameter (exp_min=-80.d0, exp_max = 80.d0)
c
	real*8	u(nx,ny,nz)
	real*8	hx(nx),hy(ny),hz(nz)
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8  solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/	solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
	common /constants/	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
c
c ==========================================================================================
c       Define some parameters
c ==========================================================================================
c
	fact= KB*Temp*Avogadro*1.d-3/4.1868d0
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
	if(l_dip.eq.0) phi0 = 0.d0
	do 50 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
		if(l_dip.eq.0) l_ion(i) = 0.d0
		k_ion(i) = kappa(i)/(1-phi0)
50	continue
c
	ener = 0.d0
c
	do 400 k = 2,nz-1
	   dz = hz(k+1)-hz(k-1)
	   do 300 j = 2,ny-1
		dy = hy(j+1)-hy(j-1)
		do 200 i = 2,nx-1
			dx = hx(i+1) - hx(i-1)
c
			info1 = ibits(solvent(i,j,k),0,2)
			if(info1.ne.0) then
c
				Ex = (u(i+1,j,k)-u(i-1,j,k))/dx
				Ey = (u(i,j+1,k)-u(i,j-1,k))/dy
				Ez = (u(i,j,k+1)-u(i,j,k-1))/dz
c
				E2 = Ex*Ex + Ey*Ey + Ez*Ez
				E = sqrt(E2)
c
				u_val = p0_e*E
c!!!!! next line added 30/01/12
				phi_val = u(i,j,k)
c
				call compute_F1(u_val,F1_u,sinhc_u,
     1				exp_mu)
c
				d = 0
				do 100 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
100				continue
                                d = d + l_dip*sinhc_u
				d = exp_mu + d
c!!!!! next line replaced 30/01/12
c				ener = ener + (log(d)+u_val)*dx*dy*dz/8
				ener = ener - (log(d)+u_val)*dx*dy*dz/8
c
			endif
c
200		continue
c
300	   continue
400	continue
c
	ener = ener*fact/vol
c
	return
	end
