c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	solve_YULP.f
c	(Generalized Poisson Boltzmann Solver)
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	2/24/08
c	Author:		Patrice Koehl, in collaboration with Marc Delarue
c
c       This program solves simultaneously two PDEs (one GPBL like,
c       and one PBE like) that provides the electrostatic potential phi
c       and a water-water potential (based on a Yukawa potential)
c
c	At this stage, it solves the equation using a finite volume method.
c	In the future, we will implement a Finite Element Method.
c
	subroutine solve_YULP(grid_shift,eps_X,eps_Y,eps_Z,
     1		helmholtz_map,crg_map,gxcf,gycf,gzcf,xf,yf,zf,
     2		nx,ny,nz,u,rwork,iwork,rparam,iparam,hx,hy,hz,
     3		resid,u_old,psi_old,sinhu,space)
c
c ==========================================================================================
c = poisson.h defines array sizes for AquaSol    ===========================================
c ==========================================================================================
c
	include 'poisson.h'
c
c ==========================================================================================
c = Declares all local/global arrays          ==============================================
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i
	integer	n_iter
	integer	l_log
	integer	nx,ny,nz
        integer nlin,method,nlevel,flag_psi
        integer iparam(100)
	integer	iwork(*)
c
        real*8  tol,tol2,lambda,val,xnorm1
        real*8  norm_psi,xnorm1_diff,psi_bulk,v0,b
	real*8	dx,dy,dz
        real*8  rparam(100)
        real*8  space(*)
	real*8	resid(*)
c
c	Constants
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
c
c	Information on maps over the grid:
c ==========================================================================================
c
c	eps_X,eps_Y,eps_Z:	dielectric constants
c	crg_map:		charge map
c	gxcf,gycf,gzcf:		boundary conditions
c	helmholtz_map:		region in which helmholtz term is non zero
c
	real*8	eps_X(*)
	real*8	eps_Y(*)
	real*8	eps_Z(*)
	real*8	crg_map(*)
        real*8  sinhu(*)
	real*8	helmholtz_map(*)
	real*8	gxcf(*)
	real*8	gycf(*)
	real*8	gzcf(*)
	real*8	rwork(*)
c
c       solutions of the equations and energy values
c ==========================================================================================
c
	real*8	ener
c
	real*8	u(*),u_old(*),psi(*),psi_old(*)
        real*8  l_dip(max_solv),l_ion(max_ion)
c
	pointer (ptr_psi,psi)
c
c	Information on grid
c ==========================================================================================
c
	real*8	hx(*),hy(*),hz(*)
        real*8  xf(*),yf(*),zf(*)
c
	integer*1 grid_shift(*)
c
c	Common blocks
c ==========================================================================================
c
        common /fugacities/     l_dip,l_ion
        common /yukawa/         v0,b,psi_bulk
        common /damping/        lambda
	common /log_file/	l_log
	common /psi_pot/	ptr_psi,flag_psi
c
	flag_psi = 1
	tol2     = 1.d-4
c
        call get_boundary_yukawa(l_log)
c
c ==========================================================================================
c = Now solve the NLPB equation using routines from MG package  ============================
c ==========================================================================================
c
c	Initialize potentials to 0 
c
        call set_all_zeros(nx,ny,nz,u)
        call set_all_zeros(nx,ny,nz,psi)
        call set_psi_bulk(nx,ny,nz,psi_bulk,grid_shift,psi)
        call set_psi_bulk(nx,ny,nz,psi_bulk,grid_shift,psi_old)
c
	write(6,*) ' '
	write(6,*) 
     1		'Solve Elliptic equation for electrostatic potential :'
	write(6,*) ' '
c
        n_iter = 1
	iparam(12) = 0
c
100     continue
c
c               First, we solve the GPBL-like equation. 
c               ====================================================
c
c                - Now solve the equation
c
		iparam(14) = 2
		iparam(10) = 1
c
                write(6,*) ' '
                write(6,*) 'Global Iteration number :',n_iter
                write(6,*) ' '
                write(6,*) 'solve GPBL-like equation'
                write(6,*) ' '
                write(l_log,*) ' '
                write(l_log,*) 
     1          'solve GPBL-like eqn for phi (electrostatic potential)'
                write(l_log,*) ' '
c
		call solve_sc(grid_shift,eps_X,eps_Y,eps_Z,helmholtz_map,
     1		crg_map,gxcf,gycf,gzcf,xf,yf,zf,nx,ny,nz,u,rwork,iwork,
     2		rparam,iparam,hx,hy,hz,resid,u_old,space,n_iter-1)
c
c               Second, we solve the "Yukawa" equation. We need to:
c               ====================================================
c
c               - redefine the epsilon maps (have been erased while
c                 solving previous equations)
c
	        call set_epsilon_yukawa(eps_X,eps_Y,eps_Z,grid_shift,
     1          nx,ny,nz)
c
c               - Set charge maps to 0
c
                call set_all_zeros(nx,ny,nz,crg_map)
c
c               - reset boundaries
c
	        call set_boundary_yukawa(psi_bulk,gxcf,gycf,gzcf,nx,
     1          ny,nz)
c
c               - define the terms depending on phi, that are passed
c                 as if it was an Helmholtz term and charge maps
c
                call compute_ion_phi(nx,ny,nz,grid_shift,u,
     1                  helmholtz_map)
                call compute_sinhu_phi(nx,ny,nz,hx,hy,hz,grid_shift,
     1                  u,sinhu,space)
c
c                - Now solve the equation
c
                write(6,*) ' '
                write(6,*) 'solve SMPB-like equation'
                write(6,*) ' '
                write(l_log,*) ' '
                write(l_log,*) 
     1          'solve SMPB-like eqn for psi (Yukawa potential)'
                write(l_log,*) ' '
c
		iparam(14) = 3
		iparam(10) = 100
c
	        call solve_eq2(grid_shift,eps_X,eps_Y,eps_Z,
     1               helmholtz_map,crg_map,gxcf,gycf,gzcf,xf,yf,zf,
     2               nx,ny,nz,psi,rwork,iwork,rparam,iparam,sinhu)
c
c               Update psi:
c               ===========
c
                do 80 i = 1,nx*ny*nz
                        psi(i)=lambda*psi(i)+ (1-lambda)*psi_old(i)
80              continue
c
c                Now check if we converged ...
c                ==============================
c
                norm_psi = xnorm1_diff(nx,ny,nz,grid_shift,psi,psi_old)
		write(6,*) ' '
                write(6,*) 'n_iter, norm_psi,tol2 :',n_iter,norm_psi,
     1		tol2
                write(l_log,*) ' '
                write(l_log,*) 'n_iter, norm_psi :',n_iter,norm_psi
c
                if(norm_psi.lt.tol2.or.n_iter.gt.100) goto 200
c
                n_iter = n_iter + 1
                call vectcopy_all(nx,ny,nz,psi,psi_old)
                goto 100
c
200     continue
c
c        call printline(nx,ny,nz,psi)
c
	return
	end
